//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef BASE_H
#define BASE_H
#include "ControlledProcess.h"
#include "FilterSocket.h"
#include "MsgStore.h"
#include "ICFile.h"
#include "Exception.h"
#include <string>
#include <iostream>

class Base {
public:
  enum Mode {
    Running,
    Stopped,
    Playback
  };
  enum SocketDesc {
    S_Primary,
    S_OnlineCoach,
    S_OfflineCoach,
    S_None
  };

private:
  FilterSocket fsocket;
  FilterSocket fsocket_ol;
  FilterSocket fsocket_cc;

  /////////////// game in progress
  Mode mode;
  int srv_time;
  int max_run;
  int max_delay;
  // if valid, timestep for which a poll was not answered
  int poll_pending_time;

  ControlledProcess server;
  // my idea who is playing which role
  int clientIDOf[NumberSActors];
  // all clients and what they claim to be
  struct SoccerClient {
    int         clientID;
    SoccerActor actor;
    SocketDesc  port;
  };
  vector<SoccerClient>  clients;

  MsgStore messages;

  /////////////// common
  bool finished;
  bool readonly;

  /////////////// logging flags
  bool log_board;
  bool msg_board;
  bool log_commands;
public:
  Base(const string &server_host, const string &logfile, bool readonly) throw (Exception); 
  ~Base();

  bool act();

  void startServer() throw (Exception);

  bool wind(int time) throw (Exception);
  bool wind_relative(int time) throw (Exception);

  int  getTime() const { return srv_time; }
  Mode getMode() const { return mode; }
  bool termReq() const { return finished; }

  /** make a connection to the soccerserver on the primary socket.
   *  actor gives the actor that is simulated
   *  Can be used to simulate a monitor, i.e. batch mode.
   */
  int connect(SoccerActor actor);

  /** send to the soccerserver.
   *  fmsg should have the sender set to a client from connect().
   */
  void sendFMsg(FilterMessage &fmsg);

  /** check wether all players have connected */
  bool allPlayersCheckedIn(bool ignoreCoaches);

  ///
  void setMaxDelay(int md) { max_delay = md; };
private:
  void stopServer() throw (Exception);
  void contServer() throw (Exception);
  //
  void dropNonFunsselMon() throw (Exception);
  //
  SoccerClient *getMessage(FilterMessage &) throw (Exception);
  //
  void sendMessage(FilterMessage &) throw (Exception);
  // send anything to anyone with mrbproxy header
  void sendFMsg(SoccerActor, const string &) throw (Exception);
  // send a protected soccermessage to monitor
  void sendFMon(const SoccerMessage &) throw (Exception);
  // mapping of sockets to clients
  void identify(SoccerClient &client, SoccerActor actor);
  // return the mapping entry for that socket
  SoccerClient &findClient(int id);
  //
  void parseSpecialCommand(SoccerActor actor, const char *msg);
  //
  void adjustTime(const string &);
  //
  SoccerActor coachSide(const string &);
  //
  SoccerActor playerSide(const string &);

  // strncmp for strings
  inline int strncmp(const string &s, const char *str, int len);

  bool sendCurMsg();
};

// and yes, this is the main module
int main(int argc, char **argv);

#endif
