//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "ControlledProcess.h"
#include <signal.h>
#include <ctype.h>
#include <sys/param.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <iostream>
#include <string.h>
#include <stdlib.h>

typedef char *charP;

ControlledProcess::ControlledProcess()
  : pid(-1)
{
}

ControlledProcess::ControlledProcess(const char *cmd[], const char *wd)
  : pid(-1)
{
  if (!wd) {
    // set workdir to pwd 
    char *tmp = new char[MAXPATHLEN];
    getcwd(tmp, MAXPATHLEN);
    restart(cmd, tmp);
    delete[] tmp;
  }
  restart(cmd, wd);
}

ControlledProcess::~ControlledProcess()
{
  term();
  if (processOk())
    kill();
}

bool ControlledProcess::restart(const char *cmd[], const char *wd)
{
  if (wd) {
    workdir = new char[strlen(wd) + 1];
    strcpy(workdir, wd);
  }
  if (cmd) {
    int cnt = 0;
    while (cmd[cnt] != NULL)
      cnt++;
    command = new charP[cnt + 1];
    for (int i = 0; i < cnt; i++) {
      command[i] = new char[strlen(cmd[i]) + 1];
      strcpy(command[i], cmd[i]);
    }
    command[cnt] = NULL;
  }

  term();
  if (processOk())
    kill();

  pid = fork();
  if (pid == 0)
    start();

  return (pid > 0);
}

void ControlledProcess::start()
{
  chdir(workdir);
  execvp(command[0], command);
  // if I get here, the process could not be started
  exit(-1);
}

bool ControlledProcess::sendSignal(int signal)
{
  if (pid < 0)
    return false;
  return (::kill(pid, signal) == 0);
}

bool ControlledProcess::stop()
{
  return sendSignal(SIGSTOP);
}

bool ControlledProcess::cont()
{
  return sendSignal(SIGCONT);
}

bool ControlledProcess::term()
{
  return sendSignal(SIGTERM);
}

bool ControlledProcess::kill()
{
  return sendSignal(SIGKILL);
}

int ControlledProcess::processOk()
{
  if (pid < 0)
    return -1;

  // check process status
  int stat;
  int err = waitpid(pid, &stat, WNOHANG);
  // no process avaiable, so assume it's ok
  if (err == 0)
    return 1;
  // check if process exited
  if ((err == pid) &&
      (WIFEXITED(stat) || WIFSIGNALED(stat))) {
    pid = -1;
    return -1;
  }
  // some error occured
  return 0;
}
