//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

///-*- c++ -*-//////////////////////////////////////////////////////////
//
// FColorModel.C - contains all colors for GUI
//
// $Id: FColors.C,v 1.17 2002/06/03 14:40:26 cs Exp $
//
////////////////////////////////////////////////////////////////////////

#include "FColors.h"
#include "Config.h"

#include <qcolor.h>
#include <qbrush.h>
#include <qpen.h>
#include <strstream.h>
#include <string>
#include <vector>
#include <stdio.h>

#define ICON_HEIGHT 10
#define ICON_WIDTH  10

#if (QT_VERSION >= 200)
#define QT2
#endif

#ifdef QT2
#define QTNSPC(cmd)  Qt::cmd
#else
#define QTNSPC(cmd)      cmd
#endif
                                                                                
void FColors::init() throw (FColors::ParamError) 
{
  leftTeam    = QBrush( readParam("FColorsLeftTeam"));
  leftGoalie  = QBrush( readParam("FColorsLeftGoalie"));
  rightTeam   = QBrush( readParam("FColorsRightTeam"));
  rightGoalie = QBrush( readParam("FColorsRightGoalie"));
  playerNorm  = QPen( readParam("FColorsPlayerNorm"));
  playerKick  = QPen( readParam("FColorsPlayerKick"));
  playerColl  = QPen( readParam("FColorsPlayerColl"),
		      readParamAsInt("FColorsPlayerCWidth"));
  ballFill    = QBrush( readParam("FColorsBallFill"));
  ballPen     = QPen( readParam("FColorsBallPen"));
  background  = readParam("FColorsBackground");
  field       = QBrush( readParam("FColorsField"));
  lines       = QPen( readParam("FColorsLines"));
  goal        = readParam("FColorsGoal");
  marker      = QPen( readParam("FColorsMarker"));
  dflLogColor = readParam("FColorsDflLogColor");

  viewFeel    = QBrush( readParam("FColorsViewConeFeel"));
  viewNear    = QBrush( readParam("FColorsViewConeNear"));
  viewFar     = QBrush( readParam("FColorsViewConeFar"));

  actLog      = readParamAsBool("FColorsActLog");
  actWM       = readParamAsBool("FColorsActWM");
  actDraw     = readParamAsBool("FColorsActDraw");
  autoAct     = readParamAsBool("FColorsAutoAct");


  // find out number of custom drawing colors for agents
  const char *tmp= Config::getValue("FColorsNum");
  int numCustCols(0);
  
  if (!tmp)
    throw ParamError( "FColorsNum" );
  
  if (sscanf(tmp, "%d", &numCustCols) != 1)
    throw ParamError( "FColorsNum format error" );
  
  agentC = vector<QColor>(numCustCols);
  for (int i=0; i < numCustCols; i++) {
    strstream s;
    s << "FColorsCustCol" << i << (char)0;
    agentC[i]= readParam( s.str() );
  }
  
  agentF = vector<Flags>(numCustCols);
  for (int i=0; i < numCustCols; i++) {
    strstream s;
    s << "FColorsCustDfl" << i << (char)0;
    agentF[i].readParam( s.str() );
  }
}

QColor FColors::readParam(const char *name)
  throw (FColors::ParamError)
{
  const char *v= Config::getValue(name);

  if (!v) 
    throw ParamError( name );
  else
    return QColor( v );
}

int FColors::readParamAsInt(const char *name)
  throw (FColors::ParamError)
{
  const char *v= Config::getValue(name);
  int val(0);

  if (!v || (sscanf(v,"%d",&val) != 1)) 
    throw ParamError( name );
  else
    return val;
}

bool FColors::readParamAsBool(const char *name)
  throw (FColors::ParamError)
{
  const char *v= Config::getValue(name);
  char buf[6];
  
  if (!v || (sscanf(v," %5s", buf) != 1))
    throw ParamError( name );
  else 
    return (strncmp(buf,"true", 4) == 0) 
      || (strncmp(buf,"TRUE", 4) == 0);
}

// definition of class members:

QBrush FColors::leftTeam    = QTNSPC(NoBrush);
QBrush FColors::leftGoalie  = QTNSPC(NoBrush);
QBrush FColors::rightTeam   = QTNSPC(NoBrush);
QBrush FColors::rightGoalie = QTNSPC(NoBrush);
QPen   FColors::playerNorm  = QTNSPC(NoPen);
QPen   FColors::playerKick  = QTNSPC(NoPen);
QPen   FColors::playerColl  = QTNSPC(NoPen);
QBrush FColors::ballFill    = QTNSPC(NoBrush);
QPen   FColors::ballPen     = QTNSPC(NoPen);
QColor FColors::background  = QTNSPC(black);
QBrush FColors::field       = QTNSPC(NoBrush);
QPen   FColors::lines       = QTNSPC(NoPen);
QColor FColors::goal        = QTNSPC(black);
QPen   FColors::marker      = QTNSPC(NoPen);
QColor FColors::dflLogColor = QTNSPC(black);
QBrush FColors::viewFeel    = QTNSPC(NoBrush);
QBrush FColors::viewNear    = QTNSPC(NoBrush);
QBrush FColors::viewFar     = QTNSPC(NoBrush);

vector<QColor> FColors::agentC;
vector<FColors::Flags> FColors::agentF;
bool FColors::actLog  = false;
bool FColors::actWM   = false;
bool FColors::actDraw = false;
bool FColors::autoAct = false;

int FColors::numAgentColors()        
{ 
  return agentC.size();
}

const QColor & FColors::agentColor(int i) 
{ 
  return agentC[i]; 
}

const FColors::Flags & FColors::agentFlags(int i) 
{ 
  return agentF[i]; 
}

QPixmap *FColors::createColorIcon(const int color)
{
  QPixmap *tmp( new QPixmap(ICON_WIDTH, ICON_HEIGHT) );
  QPainter p(tmp);
  p.setPen( QPen(QTNSPC(black)) );
  p.fillRect( tmp->rect(), QBrush( QTNSPC(white) ) );
  if (color == -1)
    p.setBrush( QBrush( FColors::dflLogColor ));
  else
    p.setBrush( QBrush( FColors::agentColor( color )));
  p.drawEllipse( tmp->rect() );
  p.end();
  return tmp;
}

QPixmap *FColors::createTeamIcon(const bool left)
{
  QPixmap *tmp( new QPixmap(ICON_WIDTH, ICON_HEIGHT) );
  QPainter p(tmp);
  p.setPen( QPen(QTNSPC(black)) );
  p.fillRect( tmp->rect(), QBrush( QTNSPC(gray) ) );
  if (left)
    p.setBrush( leftTeam );
  else
    p.setBrush( rightTeam );
  p.drawEllipse( tmp->rect() );
  p.end();
  return tmp;
}

void FColors::Flags::readParam(const char *name)
{
  const char *v= Config::getValue(name);
  int len(0), tmpi(0);
  
  // first set defaults
  drawTraces= FColors::actDraw;
  WM=         FColors::actWM;
  log=        FColors::actLog;
  playerNo=   0;
  leftTeam=   true;
  active=     false;
  
  // first item: 1=active, 0=inactive
  switch(v[0]) {
  case '1':
    active= true;
    v++;
    break;
  case '0':
    active= false;
    v++;
    break;
  }
  
  // find first separator
  if (v[0] != ',')
    return;
  // ignore first separator:
  v++;

  // second item: [l|r][nr]
  // check for side first
  switch(v[0]) {
  case 'l':
  case 'L':
    leftTeam= true;
    v++;
    break;
  case 'r':
  case 'R':
    leftTeam= false;
    v++;
    break;
  }
  // check for player number
  if (sscanf(v, "%d %n", &tmpi, &len) == 1) {
    playerNo= min(max(tmpi,0),11);
  }
  
  v+= len;
  
  // find second separator
  if (v[0] != ',')
    return;
  // ignore second separator:
  v++;

  // third item: [D][W][L]
  while (v[0] != 0) {
    switch(v[0]) {
    case 'D':
    case 'd':
      drawTraces= true;
      break;
    case 'W':
    case 'w':
      WM= true;
      break;
    case 'L':
    case 'l':
      log= true;
      break;
    }
    v++;
  }
}

