//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
// FieldMap.h - displays (parts of) the soccer field
// $Id: FieldMap.h,v 1.28 2001/09/11 10:08:38 goetz Exp $
// Time-stamp: <2001-08-31 14:01:16 goetz>
//////////////////////////////////////////////////////////////////////

#ifndef FIELDMAP_H
#define FIELDMAP_H

#include <list>

#include <qwidget.h>
#include <qpopupmenu.h>
#include <qpixmap.h>

#include "FieldObject.h"
#include "FMagFactorDlg.h"

#include "coordinates.h"


class FieldMap : public QWidget
{
  Q_OBJECT
  
public:
  FieldMap( QWidget *parent = 0, const char *name = 0,
	    const double magnify = DFLT_MAG, 
	    const Vector2d &viewTopLeft = 
	    Vector2d(-FIELD_LENGTH/2 - DFLT_BORDER_SIZE, 
		     -FIELD_WIDTH/2 - DFLT_BORDER_SIZE), 
	    const Vector2d &viewBotRight = 
	    Vector2d(FIELD_LENGTH/2 + DFLT_BORDER_SIZE, 
		     FIELD_WIDTH/2 + DFLT_BORDER_SIZE));

  ~FieldMap();

  // conversions based on widget parameters
  Vector2d  getFC( const int x, const int y ) const;
  Vector2d  getFC( const QPoint &q ) const
  { return getFC( q.x(), q.y() ); }
  
  QPoint    getWP( const double x, const double y ) const;
  QPoint    getWP( const Vector2d &v ) const
  { return getWP( v.getX(), v.getY() ); }

  static const double DFLT_MAG;
  static const double DFLT_BORDER_SIZE;
  static const double DFLT_SPEED_MAG;
  static const double DFLT_PLAYER_MAG;
  static const double DFLT_BALL_MAG;

  double speedmag() const { return mag_speed; }
  double playermag() const { return mag_player; }
  double ballmag() const { return mag_ball; }
  bool   drawKickRange() const 
  { return showMenu->isItemChecked(0); }
  bool   showViewCone( bool leftTeam, int PlayerNo ) const;
public slots:
  void      setView( const Vector2d & topLeft, const Vector2d & botRight );
  void      setTime( int time )
  { this->time = time; }
  // redraw of all list objects forced - there is NO copy made of the
  // pointer, so ensure that the list is still in place!
  void      redrawObjs( const list<FieldObject *> * );
  void      magSteps( double player, double ball, double vel );
  void      dflMagSteps( double player, double ball, double vel );

  // set width or height to match correct aspect ratio
  void      correctAspR();
  // set aspect ratio and leave width untouched!
  void      correctAspRWidth();
  // set aspect ratio and leave height untouced!
  void      correctAspRHeight();
  // switch aspect ratio correction on and off
  void      enableAspRCorr( bool );
  // switch zoom in this window on and off
  void      enableZoomThis( bool );

  // stuff also for menu handling:
protected slots:
  void      setInitialDimensions();
  void      zoomToWholeField();
  void      boxZoomTo();
  void      boxZoomFact1();
  void      boxZoomFact2();
  void      boxZoomFact5();
  void      setMarker();
  void      dispFreekickLeft();
  void      dispFreekickRight();
  void      dispDropball();
  void      setPlayerMenu();
  void      showMagStep();
  void      toggleKickRange();
  void      toggleViewConeLeft(int);
  void      toggleViewConeRight(int);
  
  // this is some nasty stuff but neccessary as long as we want to support
  // QT version 1 !!!
  void      toggleViewConeL1()
  { toggleViewConeLeft(1); }
  void      toggleViewConeL2()
  { toggleViewConeLeft(2); }
  void      toggleViewConeL3()
  { toggleViewConeLeft(3); }
  void      toggleViewConeL4()
  { toggleViewConeLeft(4); }
  void      toggleViewConeL5()
  { toggleViewConeLeft(5); }
  void      toggleViewConeL6()
  { toggleViewConeLeft(6); }
  void      toggleViewConeL7()
  { toggleViewConeLeft(7); }
  void      toggleViewConeL8()
  { toggleViewConeLeft(8); }
  void      toggleViewConeL9()
  { toggleViewConeLeft(9); }
  void      toggleViewConeL10()
  { toggleViewConeLeft(10); }
  void      toggleViewConeL11()
  { toggleViewConeLeft(11); }

  void      toggleViewConeR1()
  { toggleViewConeRight(1); }
  void      toggleViewConeR2()
  { toggleViewConeRight(2); }
  void      toggleViewConeR3()
  { toggleViewConeRight(3); }
  void      toggleViewConeR4()
  { toggleViewConeRight(4); }
  void      toggleViewConeR5()
  { toggleViewConeRight(5); }
  void      toggleViewConeR6()
  { toggleViewConeRight(6); }
  void      toggleViewConeR7()
  { toggleViewConeRight(7); }
  void      toggleViewConeR8()
  { toggleViewConeRight(8); }
  void      toggleViewConeR9()
  { toggleViewConeRight(9); }
  void      toggleViewConeR10()
  { toggleViewConeRight(10); }
  void      toggleViewConeR11()
  { toggleViewConeRight(11); }

  void      clearView();
  void      toggleClearView();

signals:
  void      statusMsg ( const QString & );
  void      statusMsg ( const char * );
  void      markerSet ( Vector2d pos, int time );
  void      zoomWindow ( Vector2d topLeft, Vector2d botRight,
			 double mag );
  void      sigFreekick ( Vector2d pos, bool sideLeft );
  void      sigDropBall ( Vector2d pos );
  void      sigSetPlayer ( Vector2d pos );
  void      changedSize();

protected:
  void      mousePressEvent ( QMouseEvent * );  
  void      mouseMoveEvent ( QMouseEvent * );
  void      mouseReleaseEvent ( QMouseEvent * );  
  void      paintEvent ( QPaintEvent * );
  void      resizeEvent ( QResizeEvent * );
  
private:
  QPopupMenu *contextMenu;
  QPopupMenu *contextZoomMenu;
  QPopupMenu *contextAspRMenu;
  QPopupMenu *boxMenu;
  QPopupMenu *boxZoomMenu;
  QPopupMenu *foulMenu;
  QPopupMenu *showMenu;
  QPopupMenu *showViewMenu;
  QPopupMenu *showViewMenuLeft;
  QPopupMenu *showViewMenuRight;
  FMagFactorDlg *magSetDlg;

  // supporting the constructor
  void      buildMenus();

  int       time;

  // the real painting procedure: redraw the buffer
  void      buildPixmbuffer();
  QPixmap   *pixmbuffer;

  QPoint    mouseDownPos;
  int       mouseDownTime;
  bool      dragging;
  QPoint    mouseDragPos;

  // find out what weird box the user has drawn and make something good
  // out of it
  QRect     getBoxInOrder();
  
  // part of the playing field to display
  Vector2d  viewTopLeft;
  Vector2d  viewBotRight;

  // original viewport dimensions (for undoing zoom function)
  Vector2d  oldViewTopLeft;
  Vector2d  oldViewBotRight;
  double    oldMag;

  // some magnifying factors
  double    mag_speed;
  double    mag_player;
  double    mag_ball;

  // here is the list of all current objects (should start with background!)
  const list<FieldObject *> *objs;

  // buffer for writing status msgs
  char      status[80];

  // pixmaps for teams
  QPixmap   *leftTeamIcon;
  QPixmap   *rightTeamIcon;
};

#endif
