//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-///////////////////////////////////////////////////////////
//
// MainWidget.h - main window of FUNSSEL program
//
// $Id: MainWidget.h,v 1.42 2001/09/11 10:08:38 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#ifndef MAINWIDGET_H
#define MAINWIDGET_H

#include <qwidget.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qkeycode.h>
#include <qbutton.h>
#include <qstatusbar.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qapplication.h>
#include <qmessagebox.h>

#include <list.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "coordinates.h"
#include "Config.h"
#include "mcommand.h"

#include "FieldMap.h"
#include "FieldObject.h"
#include "FColorSelect.h"
#include "FColors.h"
#include "FLogWindow.h"
#include "FPlayerWin.h"
#include "FMarkerWin.h"
#include "SoccerMsgParser.h"

#include "SoccerMessage.h"
#include "FunsselIO.h"


class MainWidget : public QWidget
{
  Q_OBJECT
public:
  MainWidget(int port, const char *host, const char *mfilename,
	     QWidget * parent=0, const char * name=0);
  ~MainWidget();
  
  // some functions telling misc. states of the Widget
  int        curTime()
  { return atoi(timeDisp->text()); }

  // all output to be considered
  bool       showLeft()
  { return dispObjMenu->isItemChecked(0); }
  bool       showRight()
  { return dispObjMenu->isItemChecked(1); }
  bool       showTrace()
  { return dispObjMenu->isItemChecked(2); }
  bool       showLogAudio()
  { return dispObjMenu->isItemChecked(3); }
  bool       showLogPlayerCmds()
  { return dispObjMenu->isItemChecked(4); }
  bool       playerMayControl()
  { return dispObjMenu->isItemChecked(5); }
  bool       displayStamina()
  { return dispObjMenu->isItemChecked(6); }
  bool       displayPlayerType()
  { return dispObjMenu->isItemChecked(7); }

  // evaluate settings of all color selects to find out about 
  // the display of a certain object
  // returns -1, if not found; otherwise returns FIRST activated color
  int        getAColorDraw( int playerNo, bool leftTeam );
  int        getAColorWM( int playerNo, bool leftTeam );
  int        getAColorLog( int playerNo, bool leftTeam );

  bool       isOpen()
  { return windowIsOpen; }

public slots:
  // this function is used to activate the parser once
  void       pollMessages();

  void       markerSet ( Vector2d pos, int time );
  void       zoomWindow ( Vector2d topLeft, Vector2d botRight,
			  double mag );
  void       sigFreekick ( Vector2d pos, bool sideLeft );
  void       sigDropBall ( Vector2d pos );
  void       sigSetPlayer ( Vector2d pos );

  void       setPlayerExec( bool sideLeft, int num, Vector2d pos, Angle ang);
  void       windMarker( int time );

  void       windRew0();
  void       windRew1();
  void       windFwd1();
  void       windFwdMax();

  void       playStop();
  void       playStart();
  void       kickOff();
  void       startServer();
  void       setPlayDelay();
  void       incPlayDelay();
  void       decPlayDelay();

  void       setCurrentTime(int);
  void       setServerTime(int);
  void       setPlayMode(const char *mode)
  { playMode->setText(mode); }
  void       setLeftLabel(const char *s)
  { leftTeam->setText(s); }
  void       setRightLabel(const char *s)
  { rightTeam->setText(s); }

signals:
  void      statusMsg(const QString &);  
  void      statusMsg( const char * );

protected slots:
  void       wind();
  void       playMaxTime();

  void       doRedraw();
  
  void       toggleShowLeft()
  { dispObjMenu->setItemChecked(0, !showLeft() ); }
  void       toggleShowRight()
  { dispObjMenu->setItemChecked(1, !showRight() ); }
  void       toggleShowTrace()
  { dispObjMenu->setItemChecked(2, !showTrace() ); }
  void       toggleAudioMsg()
  { dispObjMenu->setItemChecked(3, !showLogAudio() ); }
  void       togglePlayerCmds()
  { dispObjMenu->setItemChecked(4, !showLogPlayerCmds() ); }
  void       togglePlayerMayControl()
  { dispObjMenu->setItemChecked(5, !playerMayControl() ); }
  void       toggleDisplayStamina()
  { dispObjMenu->setItemChecked(6, !displayStamina() ); }
  void       toggleDisplayPlayerType()
  { dispObjMenu->setItemChecked(7, !displayPlayerType() ); }


  void       showMarkerWin( bool );
  void       showPlayerWin( bool );
  void       showLogWin( bool );

  void       toggleMarkerWin()
  { showMarkerWin( !(markerwin->isVisible()) ); }
  void       togglePlayerWin()
  { showPlayerWin( !(playerwin->isVisible()) ); }
  void       toggleLogWin()
  { showLogWin( !(logwin->isVisible()) ); }

  void       matchFieldMapSize();

  // update the status of menu entries ...
  void       closedMarkerWin()
  { windowMenu->setItemChecked(0, false);}
  void       closedPlayerWin()
  { windowMenu->setItemChecked(1, false);}
  void       closedLogWin()
  { windowMenu->setItemChecked(2, false);}

  void       quitApplication();
  void       aboutBox();
  void       aboutQt();

protected:
  virtual void resizeEvent ( QResizeEvent * );
  virtual void closeEvent ( QCloseEvent * );  

  bool       closeQuery();
  bool       deleteFileQuery();
  bool       killServerQuery();
    
private:
  FieldMap   *mainMap;
  QMenuBar   *mainMenu;
  QPopupMenu *fileMenu;
  QPopupMenu *controlMenu;
  QPopupMenu *dispObjMenu;
  QPopupMenu *windowMenu;
  QPopupMenu *aboutMenu;
  QStatusBar *statusBar;

  QLabel     *leftTeam;
  QLabel     *rightTeam;
  QLabel     *timeDisp;
  QLabel     *serverTime;
  QLabel     *playMode;

  QPushButton *btnKickOff;
  
  QPushButton *rewBeg;
  QPushButton *rewStep;
  QPushButton *btnStop;
  QPushButton *fwdStep;
  QPushButton *fwdPlay;
  QPushButton *fwdMax;

  QLineEdit   *destTime;
  
  QPushButton *playTo;
  QPushButton *jumpTo;

  QPushButton *btnfwddelay;
  QPushButton *btnfwddelayd;
  QPushButton *btnfwddelayi;
  
  // related windows
  FLogWindow *logwin;
  FPlayerWin *playerwin;
  FMarkerWin *markerwin;

  // list of objs
  FColorSelect **cSel;

  // some helpers for the constructor
  void       buildMenu();
  void       setupObjects();

  // connect the signals for FieldMap's
  void       connectFieldMap( const FieldMap*);
  void       setMagSteps( FieldMap *, bool isMainMap= false );

  // all child windows to operate
  list<FieldMap *> zoomWindows;

  // the object list of objects to be displayed;
  list<FieldObject *> *objs;

  // auxiliary function for building the object list
  // creates a small list only including the basic objects
  // adds field background and markers
  list<FieldObject *> *basicList();

  // auxiliary function for deleting the list
  // checks for 0 pointer itself (like delete operator)
  void deleteList( list<FieldObject *> * &);

  // parser to parse SoccerMessages
  SoccerMsgParser *smp;

  // icons for menu
  QPixmap    *leftTeamIcon;
  QPixmap    *rightTeamIcon;

  // the socket communication class
  FunsselIO  fio;

  // this switch is true if we already have a logfile
  bool       haveOldLog;
  bool       serverStarted;
  bool       gameRunning;

  // similar for playback: number of cycles required
  int        playBackDelay;
  // number of actually done cycles
  int        playBackCnt;
  // upper bound for playback in "playMaxTime" mode
  int        playBackMax;
  
  bool       listDirty;

  // this is a small hack because qt does not support any flag that tells
  // the user when a widget is closed
  bool       windowIsOpen;
};

#endif
