//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// SoccerMsgParser.h - A parser for messages coming from various sources
//
// $Id: SoccerMsgParser.C,v 1.29 2001/09/11 10:08:38 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

//#define USEPARSETRACE

#ifdef USEPARSETRACE
#define PARSETRACE(output) cout <<"Parser(t=" << msg.time << "): " << \
 output << endl
#warning: SoccerMsgParser debug trace PARSETRACE is activated!
#else
#define PARSETRACE(output)
#endif


#include "SoccerMsgParser.h"

#include "MainWidget.h"

#include "FunsselStrings.h"

// get all the stuff from soccerserver
namespace SServer {
#include "types.h"
#include "param.h"

// PLAYMODE_STRINGS is a #define from types.h
const char *playModeStrings[]= PLAYMODE_STRINGS;
};

SoccerMsgParser::SoccerMsgParser( MainWidget *mw, FMarkerWin *fmw, 
				  FLogWindow *flw )
  : mw(mw), fmw(fmw), flw(flw)
{}

FieldObject *SoccerMsgParser::parseMessage( const SoccerMessage &msg )
{
  SenderGroup grp(None);
  int plNum(-1);
  bool leftTeam(false);
  
  PARSETRACE( "got msg." );

  if (msg.recver != FunsselMonitor )
    return 0;

  getSender(msg, grp, plNum, leftTeam);

  switch(grp) {
  case Player:
    return playerMessage( msg, plNum, leftTeam );
    break;
  case Server:
    return serverMessage( msg );
    break;
  default:
    return 0;
  }
}

FieldObject *SoccerMsgParser::serverMessage( const SoccerMessage &msg )
{

  PARSETRACE( "parsing server msg." );

  // NB: in this procedure I just kept all of Noda's variable names
  // to make debugging easier

  // magic binary data format conversion: it's ugly, but the monitor
  // does it and so do we:
  SServer::dispinfo_t2 *rinfo2= 
    (SServer::dispinfo_t2 *)(string(msg).data());

  // decide which type of info we just got
  char msgType(ntohs((short)(rinfo2->mode)));

  switch (msgType) {
  case SHOW_MODE: {
    
    PARSETRACE("got showinfo msg.");

    ShowinfoType *s= new ShowinfoType();
    
    double sigmastar= (double) SHOWINFO_SCALE2;

    PARSETRACE( "SHOWINFO_SCALE2 " << sigmastar << " " << 1/sigmastar );

    SServer::showinfo_t2 *show2=&(rinfo2->body.show);

    s->ball= new FieldBall(Vector2d( (long)ntohl(show2->ball.x)/sigmastar,
				    (long)ntohl(show2->ball.y)/sigmastar));
    
    if (mw->showLeft()) 
      for (int i=0; i<MAX_PLAYER; i++) {
	short kind= ntohs(show2->pos[i].mode);
	s->ourTeam[i]= new FieldPlayer( Vector2d((long)ntohl(show2->pos[i].x)
						 /sigmastar,
						 (long)ntohl(show2->pos[i].y)
						 /sigmastar),
					Angle( (long)ntohl(show2->pos[i].body_angle)
					       / sigmastar ),
					Angle( (long)ntohl(show2->pos[i].body_angle)
					       / sigmastar ) +
					Angle( (long)ntohl(show2->pos[i].head_angle)
					       / sigmastar ),
					Angle( (long)ntohl(show2->pos[i].view_width) / sigmastar),
					i + 1, true, 
					kind & (KICK | CATCH),
					kind & GOALIE,
					kind & (PLAYER_COLLIDE | BALL_COLLIDE),                                    
					mw->displayStamina() ? ((long)ntohl(show2->pos[i].stamina) / sigmastar) : -1,
					mw->displayPlayerType() ? ((short)ntohs(show2->pos[i].type)) : -1 );
      }
    if (mw->showRight()) 
      for (int i=MAX_PLAYER; i<2*MAX_PLAYER; i++) {
	short kind= ntohs(show2->pos[i].mode);
	s->otherTeam[i-MAX_PLAYER]= new FieldPlayer( Vector2d((long)ntohl(show2->pos[i].x)
							      /sigmastar,
							    (long)ntohl(show2->pos[i].y)
							      /sigmastar),
						     Angle( (long)ntohl(show2->pos[i].body_angle)
							    / sigmastar ),
						     Angle( (long)ntohl(show2->pos[i].body_angle)
							    / sigmastar ) +
						     Angle( (long)ntohl(show2->pos[i].head_angle)
							    / sigmastar ),
						     Angle( (long)ntohl(show2->pos[i].view_width) / sigmastar),
						     i - MAX_PLAYER + 1, false, 
						     kind & (KICK | CATCH),
						     kind & GOALIE,
						     kind & (PLAYER_COLLIDE | BALL_COLLIDE),
						     mw->displayStamina() ? ((long)ntohl(show2->pos[i].stamina) / sigmastar) : -1,
						     mw->displayPlayerType() ? ((short)ntohs(show2->pos[i].type)) : -1 );
      }
  
  mw->setPlayMode( SServer::playModeStrings[ show2->pmode ] );
  char buf[25]= "";
  sprintf(buf, "%d %s", ntohs(show2->team[0].score), show2->team[0].name);
  mw->setLeftLabel(buf);
  sprintf(buf, "%d %s", ntohs(show2->team[1].score), show2->team[1].name);
  mw->setRightLabel(buf);
  
    return s;
  } break;
  case MSG_MODE:

    PARSETRACE("got msg board message." );

    // check whether message or log entry:
    switch(ntohs(rinfo2->body.msg.board)) {
    case MSG_BOARD:
      if (mw->showLogAudio()) {
	char *s= cleanString( rinfo2->body.msg.message );
	flw->addLine(msg.getTime(), s);
	delete [] s;
      }
      break;
    case LOG_BOARD:
      if (mw->showLogPlayerCmds()) {
	char *s= cleanString( rinfo2->body.msg.message );
	flw->addLine(msg.getTime(), s);
	delete [] s;
      }
      break;
    default:
      ; // nothing to do!
    }
    // no graphical object returned !
    return 0;
    break;
  default:
    
    PARSETRACE("got other msg." << (int)msgType );

    return 0;
  }
}

FieldObject *SoccerMsgParser::playerMessage( const SoccerMessage &msg, 
					     int plNum, bool leftTeam)
{
  const char *m= string(msg).c_str();
  PARSETRACE("Got player msg.");
  if (!(mw->showTrace()))
    return 0;
  if (strncmp(m, FILTER_STRING " ", FILTER_STRING_LEN + 1) == 0){
    // ignore filter string and duration
    m+= FILTER_STRING_LEN + 1;
    int dummy, len;
    sscanf(m, "%d %n", &dummy, &len);
    m+= len;
    double x(0),y(0),ang(0),vx(0),vy(0),bangle(0),hangle(0);
    int dur(0),no(0),idummy(0),si(0),age(0),ival(0); // si = the side

    if (strncmp(m, FMC_STOP, FMC_STOP_LEN) == 0) {
      PARSETRACE("Got player stop cmd.");
      if (mw->playerMayControl()) 
	mw->playStop();
    } else if (sscanf(m, FMC_DROPBALL " %d ( %lf , %lf ) )",
		      &dur, &x, &y) == 3) {
      PARSETRACE("Got player drop ball cmd.");
      if (mw->playerMayControl()) {
	mw->sigDropBall(Vector2d( sws(x, leftTeam),
				  sws(y, leftTeam)));
      }
    } else if (sscanf(m, FMC_FREEKUS " %d ( %lf , %lf ) )",
		      &dur, &x, &y) == 3) {
      PARSETRACE("Got player free kick us cmd.");
      if (mw->playerMayControl()) {
	mw->sigFreekick(Vector2d( sws(x, leftTeam),
				  sws(y, leftTeam)), leftTeam);
      }
    } else if (sscanf(m, FMC_FREEKTHEM " %d ( %lf , %lf ) )",
		      &dur, &x, &y) == 3) {
      PARSETRACE("Got player free kick them cmd.");
      if (mw->playerMayControl()) {
	mw->sigFreekick(Vector2d( sws(x, leftTeam),
				  sws(y, leftTeam)), !leftTeam);
      }
    } else if (sscanf(m, FMC_SETPLYR " %d %d %d ( %lf , %lf ) %lf )",
		      &dur, &no, &si, &x, &y, &ang) == 6) {
      PARSETRACE("Got set player cmd.");
      if (mw->playerMayControl()) {
	mw->setPlayerExec( (si==1)? leftTeam : !leftTeam,
			   no, Vector2d(sws(x, leftTeam), sws(y, leftTeam)),
			   sws(Angle(ang), leftTeam));
      }
    } else if (strncmp(m, FMC_MARK, FMC_MARK_LEN) == 0) {
      PARSETRACE("Got marker cmd from player");
      parseMarker( msg.getTime(), m );
    } else if (strncmp(m, FMC_LOG " ", FMC_LOG_LEN+1) == 0) {
      // found log entry. Find out if player is allowed to shout
      int col= mw->getAColorLog( plNum, leftTeam );
      if (col == -1) 
	PARSETRACE("Got log msg from inactive player " 
		   << plNum << " " << leftTeam);
      else {
	PARSETRACE("Got log msg from active player " 
		   << plNum << " " << leftTeam);
	char *s= cleanString( m+FMC_LOG_LEN+1 );
	s[strlen(s)-1]= 0;
	flw->addLine( msg.getTime(), col, plNum, s );
	delete [] s;
      }
    } else if (strncmp(m, FMC_SNAP " ", FMC_SNAP_LEN+1) == 0) {
      int col= mw->getAColorWM( plNum, leftTeam );
      if (col == -1)
	PARSETRACE("Got snapshot msg for inactive");
      else {
	PARSETRACE("Got snapshot msg for active");
	// this code heavily depends on the implementation of Snapshot
	// in the libMRB
	istrstream ist(m+FMC_SNAP_LEN+1);
	QColor acol= FColors::agentColor(col);
	
	// the object to be sent to monitor
	FieldSnapshot *sn= new FieldSnapshot();
      
	ist >> idummy >> si; // time, side
            
	ist >> x >> y; // my pos
	ist >> idummy; // my.number
	ist >> bangle >> hangle >> age; // my body-, head-angle, age
	ist >> ival; // viewWidth
            
	// get our team
	for (int i=0; i < 11; i++) {
	  ist >> x >> y >> age >> bangle;
	  sn->ourTeam[i]= new WFieldPlayer( Vector2d(sws(x,leftTeam),sws(y,leftTeam)),
					    sws(Angle(bangle), leftTeam),
					    i+1, leftTeam,
					    age, Vector2d(), // vel not known
					    acol); // isGoalie not known
	}

	// get other team
	for (int i=0; i < 11; i++) {
	  ist >> x >> y >> age >> bangle;
	  sn->otherTeam[i]= new WFieldPlayer( Vector2d(sws(x,leftTeam),sws(y,leftTeam)),
					      sws(Angle(bangle), leftTeam),
					      i+1, !leftTeam,
					      age, Vector2d(), // vel not known
					      acol); // isGoalie not known
	}
      
	// get ball data
	ist >> x >> y >> age >> idummy >> idummy >> vx >> vy;
	sn->ball= new WFieldBall( Vector2d( sws(x,leftTeam), sws(y,leftTeam) ),
				  age, 
				  Vector2d( sws(vx,leftTeam), sws(vy,leftTeam)),
				  acol);
	return sn;
      } 
    } else { // now try to handle the player's draw traces
      int col= mw->getAColorDraw( plNum, leftTeam );
      if (col == -1) 
	PARSETRACE("Got draw command for inactive player");
      else {
	PARSETRACE("Got draw command for active player");
	// decide which one
	double x1(0), y1(0), r(0), al(0);
	int ival(0), sty(0);
	QColor acol= FColors::agentColor(col);
	if (sscanf(m, FMC_CIRCLE " %d ( %lf , %lf ) %lf %d %d )",
		   &dur, &x, &y, &r, &ival, &sty) == 6) {
	  PARSETRACE("Found circle.");
	  return new FieldCircle(Vector2d( sws(x, leftTeam),
					   sws(y, leftTeam)),
				 r, (ival != 0), acol, MStyle(sty));
	} else if (sscanf(m, FMC_POS " %d ( %lf , %lf ) )",
			  &dur, &x, &y) == 3) {
	  PARSETRACE("Found position marker");
	  return new FieldPosition(Vector2d( sws(x, leftTeam),
					     sws(y, leftTeam)), acol );
	} else if (sscanf(m, FMC_LINE " %d ( %lf, %lf ) ( %lf, %lf ) %d )",
			  &dur, &x, &y, &x1, &y1, &sty) == 6) {
	  PARSETRACE("Found line");
	  return new FieldLine(Vector2d( sws(x,leftTeam), sws(y,leftTeam)),
			       Vector2d( sws(x1,leftTeam), sws(y1,leftTeam)),
			       acol, MStyle(sty));
	} else if (sscanf(m, FMC_VLINE " %d %lf %d )", &dur, &x, &sty) == 3 ) {
	  PARSETRACE("Found vline");
	  return new FieldVLine( sws(x, leftTeam), acol, MStyle(sty));
	} else if (sscanf(m, FMC_BOX " %d ( %lf , %lf ) ( %lf , %lf ) %d",
			  &dur, &x, &y, &x1, &y1, &sty) == 6 ) {
	  PARSETRACE("Found box");
	  return new FieldBox(Vector2d( sws(x,leftTeam), sws(y,leftTeam)),
			      Vector2d( sws(x1,leftTeam), sws(y1,leftTeam)),
			      acol, MStyle(sty));
	} else if (sscanf(m, FMC_CORR " %d ( %lf , %lf ) ( %lf , %lf ) "
			  " %lf %lf %d )", &dur, &x, &y, &x1, &y1, &r, &al, &sty ) == 8) {
	  PARSETRACE("Found corridor");
	  return new FieldCorridor( Vector2d( sws(x,leftTeam), sws(y,leftTeam)),
				    Vector2d( sws(x1,leftTeam),sws(y1,leftTeam)),
				    r, Angle(al), acol, MStyle(sty));
	} else if (sscanf(m, FMC_LABEL " %d ( %lf , %lf ) %n",
			  &dur, &x, &y, &ival) >= 3) {
	  PARSETRACE("Found label");
	  char *s= cleanString(m+ival);
	  FieldLabel *f= new FieldLabel( Vector2d( sws(x,leftTeam), 
						   sws(y,leftTeam)),
					 string(s, strlen(m+ival) - 2), 
					 acol );
	  delete [] s;
	  return f;
	} 
      }
    } 
  }
  return 0;
}

FieldObject *SoccerMsgParser::monitorMessage( const SoccerMessage &msg )
{
  PARSETRACE("Got monitor msg.");

  const char *m= string(msg).c_str();  
  
  int dummy, len;

  if ((sscanf(m, FILTER_STRING " %d %n", &dummy, &len) >= 1) &&
      strncmp(m+len, FMC_MARK, FMC_MARK_LEN) == 0) {
    PARSETRACE("Got marker cmd from FUNSSELMon");
    parseMarker( msg.getTime(), m+len );
  }
  return 0;
}

void SoccerMsgParser::getSender( const SoccerMessage &msg,
				 SenderGroup &g,
				 int &plNum, bool &leftTeam)
{
  // first find out who sent the msg
  // NB: coach is player 0 !

  // first group ...
  if (msg.sender == SoccerServer)
    g= Server;
  else if (msg.sender == FunsselMonitor)
    g=FMonitor;
  else if (msg.sender > 0 && msg.sender < 26)
    g= Player;
  else
    g= None;

  leftTeam= msg.sender > 0 && msg.sender < 13;
  
  plNum= msg.sender % 12;
}

void SoccerMsgParser::parseMarker( int msgtime, const char *msg )
{
  int durtn;
  double posx, posy;
  char buf[201];
  if (sscanf(msg, FMC_MARK " %d ( %lf , %lf ) \"%200[^\"]\" )", &durtn, &posx, 
	     &posy, buf) == 4) {
    char *s= cleanString(buf);
    fmw->addMarker(Vector2d(posx, posy), msgtime, durtn, s);
    delete [] s;
  }
}

char * SoccerMsgParser::cleanString( const char *s )
{
  // some silly users?
  if (s == 0)
    return 0;
  int len(strlen(s));
  char *s2= new char[len+1];
  strcpy(s2, s);

  for (int i=0; i < len; i++)
    // some weird chars (nonprintables) found?
    if (s2[i] < ' ')
      s2[i]= ' ';
  return s2;
}
