//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// testMain.C - 
//
// $Id: funsselmon.C,v 1.10 2002/04/08 15:52:19 cs Exp $
//
////////////////////////////////////////////////////////////////////////

#include "MainWidget.h"
#include "FColors.h"
#include "Config.h"
#include "Exception.h"

#include <qapplication.h>
#include <qstatusbar.h>

#if (QT_VERSION >= 200)
#define QT2
#endif

#ifdef QT2
#include <qwindowsstyle.h>
#endif

#include <unistd.h>
#include <sys/resource.h>
#include <errno.h>

#define FORCE_CORE
                                                                               
int main (int argc, char **argv)
{
#ifdef FORCE_CORE
  struct rlimit rlim;
  getrlimit(RLIMIT_CORE, &rlim);
  rlim.rlim_cur = rlim.rlim_max;
  cerr << "core dump enabled." << endl;
  if (setrlimit(RLIMIT_CORE, &rlim) == -1)
    cerr << errno << endl;
#endif
                                                                        
  try {
    int port(6000);
    char host[50]= "localhost";
    char *mfilename=0;
    QApplication a(argc, argv);
#ifdef QT2
    QApplication::setStyle( new QWindowsStyle ); 
#else
    QApplication::setStyle( WindowsStyle ); 
#endif

    Config::readFile("funsselrc");
  
    port= Config::getValueAsInt("proxy_socket");
  
    Config::readFile("funsselmonrc");
    try {
      FColors::init();
    } catch (Exception e) {
      cerr << e << endl;
      exit(666);
    }
    
    char c;
    while ((c = getopt( a.argc(), a.argv(), "h:p:m:")) != EOF ) 
      switch(c) {
      case 'h':
	strcpy(host, optarg);
	break;
      case 'p':
	port= atoi(optarg);
	break;
      case 'm':
	mfilename= new char[strlen(optarg)+1];
	strcpy(mfilename, optarg);
	break;
      case '?':
      default:
	cerr << "Usage: " << a.argv()[0] << " [-h hostname] [-p port] " 
	     "[-m mfilename]" << endl
	     << endl
	     << " -h hostname    set hostname of FUNSSEL (default: localhost)"
	     << endl
	     << " -p port        set proxy port of FUNSSEL (default: 6000)"
	     << endl
	     << " -m mfilename   set alternative marker save file name"
	     << endl << endl;
	exit(1);
      }

    Socket::enableInts();
    
    MainWidget *mw= new MainWidget(port, host, mfilename);
    a.setMainWidget( mw );
    mw->show();
    
    while(mw->isOpen()) {
      // give time to qt
      a.processEvents(1);
      // give time to parser
      mw->pollMessages();
    }

    delete mw;
    
    exit(0);
  }
  catch (Exception e) {
    cerr << "an exception occured: " << e.descr << endl;
  }
  catch (...) {
    cerr << "an unexpected exception occured." << endl;
  }
}
