#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "Parser.h"
#include "constants.h"


#ifdef USE_PARSERTRACE
#define PARSERTRACE(str) cout << "Parser: " << str << endl;
#warning *** Parser: PARSERTRACE enabled ***
#else
#define PARSERTRACE(str)
#endif

#ifdef USE_COACHTRACE
#define COACHTRACE(str) cout << "Parser:      " << str << endl;
#warning *** Parser: COACHTRACE enabled ***
#else
#define COACHTRACE(str)
#endif

#ifdef USE_PARAMTRACE
#define PARAMTRACE(token) cout << "   " << ##token##.fmt << ": ";\
        switch(token.dt) {\
        case DT_INT:  cout << *(int*)##token##.ptr; break;\
        case DT_DOUBLE: cout << *(double*)##token##.ptr; break;\
        case DT_BOOL: cout << *(bool*)##token##.ptr; break;\
        case DT_STRING: cout << (char*)##token##.ptr; break;\
        default: ;\
        };\
        cout << endl;
#warning *** Parser: PARAMTRACE enabled ***
#else
#define PARAMTRACE(token)
#endif

#include "ParserStrings.C"

play_mode 
Parser::parsePlaymode( const char * s, const playing_side side ) {

  PARSERTRACE("parsing play mode");

  if (!strncmp(s, "before_kick_off", 15))
    return PLAY_MODE_BEFORE_KICK_OFF;

  else if (!strncmp(s, "time_over", 9))
    return PLAY_MODE_TIME_OVER;
  else if (!strncmp(s, "time_up", 7))
    return PLAY_MODE_TIME_OVER;

  else if (!strncmp(s, "play_on", 7))
    return PLAY_MODE_PLAY_ON;

  else if (!strncmp(s, "kick_off_l", 10))
    return side == SIDE_LEFT ? PLAY_MODE_KICK_OFF_US : PLAY_MODE_KICK_OFF_THEM;
  else if (!strncmp(s, "kick_off_r", 10))
    return side == SIDE_RIGHT ? PLAY_MODE_KICK_OFF_US : PLAY_MODE_KICK_OFF_THEM;

  else if (!strncmp(s, "kick_in_l", 9))
    return side == SIDE_LEFT ? PLAY_MODE_KICK_IN_US : PLAY_MODE_KICK_IN_THEM;
  else if (!strncmp(s, "kick_in_r", 9))
    return side == SIDE_RIGHT ? PLAY_MODE_KICK_IN_US : PLAY_MODE_KICK_IN_THEM;

  else if (!strncmp(s, "free_kick_l", 11))
    return side == SIDE_LEFT ? PLAY_MODE_FREE_KICK_US : PLAY_MODE_FREE_KICK_THEM;
  else if (!strncmp(s, "free_kick_r", 11))
    return side == SIDE_RIGHT ? PLAY_MODE_FREE_KICK_US : PLAY_MODE_FREE_KICK_THEM;
  else if (!strncmp(s, "foul_r", 6))
    return side == SIDE_LEFT ? PLAY_MODE_FREE_KICK_US : PLAY_MODE_FREE_KICK_THEM;
  else if (!strncmp(s, "foul_l", 6))
    return side == SIDE_RIGHT ? PLAY_MODE_FREE_KICK_US : PLAY_MODE_FREE_KICK_THEM;

  else if (!strncmp(s, "corner_kick_l", 13))
    return side == SIDE_LEFT ? PLAY_MODE_CORNER_KICK_US : PLAY_MODE_CORNER_KICK_THEM;
  else if (!strncmp(s, "corner_kick_r", 13))
    return side == SIDE_RIGHT ? PLAY_MODE_CORNER_KICK_US : PLAY_MODE_CORNER_KICK_THEM;

  else if (!strncmp(s, "goal_kick_l", 11))
    return side == SIDE_LEFT ? PLAY_MODE_GOAL_KICK_US : PLAY_MODE_GOAL_KICK_THEM;
  else if (!strncmp(s, "goal_kick_r", 11))
    return side == SIDE_RIGHT ? PLAY_MODE_GOAL_KICK_US : PLAY_MODE_GOAL_KICK_THEM;

  else if (!strncmp(s, "drop_ball", 9))
    return PLAY_MODE_DROP_BALL;

  else if (!strncmp(s, "offside_l", 9))
    return side == SIDE_LEFT ? PLAY_MODE_OFFSIDE_US : PLAY_MODE_OFFSIDE_THEM;
  else if (!strncmp(s, "offside_r", 9))
    return side == SIDE_RIGHT ? PLAY_MODE_OFFSIDE_US : PLAY_MODE_OFFSIDE_THEM;

  else if (!strncmp(s, "half_time", 9))
    return PLAY_MODE_HALF_TIME;

  else if (!strncmp(s, "first_half_over", 15))
    return PLAY_MODE_HALF_TIME;

  else if (!strncmp(s, "time_extended", 13))
    return PLAY_MODE_TIME_EXTENDED;

  else if (!strncmp(s, "pause", 5))
    return PLAY_MODE_PAUSE;

  else if (!strncmp(s, "human_judge", 11))
    return PLAY_MODE_HUMAN_JUDGE;

  else if (!strncmp(s, "goal_l", 6))
    return side==SIDE_LEFT ? PLAY_MODE_GOAL_US : PLAY_MODE_GOAL_THEM;
  else if (!strncmp(s, "goal_r", 6))
    return side==SIDE_RIGHT ? PLAY_MODE_GOAL_US : PLAY_MODE_GOAL_THEM;

  else if (!strncmp(s, "free_kick_fault_l", 17))
    return side==SIDE_RIGHT ? PLAY_MODE_FREEKICK_FAULT_THEM : PLAY_MODE_FREEKICK_FAULT_US;
  else if (!strncmp(s, "free_kick_fault_r", 17))
    return side==SIDE_LEFT ? PLAY_MODE_FREEKICK_FAULT_THEM : PLAY_MODE_FREEKICK_FAULT_US;

  else if (!strncmp(s, "goalie_catch_ball_l", 19))
    return side==SIDE_RIGHT ? PLAY_MODE_GOALIE_CATCH_THEM : PLAY_MODE_GOALIE_CATCH_US;
  else if (!strncmp(s, "goalie_catch_ball_r", 19))
    return side==SIDE_LEFT ? PLAY_MODE_GOALIE_CATCH_THEM : PLAY_MODE_GOALIE_CATCH_US;

  else if (!strncmp(s, "penalty_kick_l", 14))
    return side==SIDE_RIGHT ? PLAY_MODE_PENALTY_KICK_THEM : PLAY_MODE_PENALTY_KICK_US;
  else if (!strncmp(s, "penalty_kick_r", 14))
    return side==SIDE_LEFT ? PLAY_MODE_PENALTY_KICK_THEM : PLAY_MODE_PENALTY_KICK_US;

  else if (!strncmp(s, "foul_charge_l", 13))
    return side==SIDE_RIGHT ? PLAY_MODE_FOUL_CHARGE_THEM : PLAY_MODE_FOUL_CHARGE_US;
  else if (!strncmp(s, "foul_charge_r", 13))
    return side==SIDE_LEFT ? PLAY_MODE_FOUL_CHARGE_THEM : PLAY_MODE_FOUL_CHARGE_US;

  else if (!strncmp(s, "foul_push_l", 11))
    return side==SIDE_RIGHT ? PLAY_MODE_FOUL_PUSH_THEM : PLAY_MODE_FOUL_PUSH_US;
  else if (!strncmp(s, "foul_push_r", 11))
    return side==SIDE_LEFT ? PLAY_MODE_FOUL_PUSH_THEM : PLAY_MODE_FOUL_PUSH_US;

  else if (!strncmp(s, "foul_multiple_attack_l", 22))
    return side==SIDE_RIGHT ? PLAY_MODE_FOUL_MULTIPLE_ATTACK_THEM : PLAY_MODE_FOUL_MULTIPLE_ATTACK_US;
  else if (!strncmp(s, "foul_multiple_attack_r", 22))
    return side==SIDE_LEFT ? PLAY_MODE_FOUL_MULTIPLE_ATTACK_THEM : PLAY_MODE_FOUL_MULTIPLE_ATTACK_US;

  else if (!strncmp(s, "foul_ballout_l", 14))
    return side==SIDE_RIGHT ? PLAY_MODE_FOUL_BALLOUT_THEM : PLAY_MODE_FOUL_BALLOUT_US;
  else if (!strncmp(s, "foul_ballout_r", 14))
    return side==SIDE_LEFT ? PLAY_MODE_FOUL_BALLOUT_THEM : PLAY_MODE_FOUL_BALLOUT_US;

  else if (!strncmp(s, "back_pass_l", 11))
    return side==SIDE_RIGHT ? PLAY_MODE_BACKPASS_THEM : PLAY_MODE_BACKPASS_US;
  else if (!strncmp(s, "back_pass_r", 11))
    return side==SIDE_LEFT ? PLAY_MODE_BACKPASS_THEM : PLAY_MODE_BACKPASS_US;

  else {
    cerr << "Unknown Playmode: " << s << endl;
    return PLAY_MODE_UNKNOWN;
  }
}

Parser::Parser( const char * _team, const client_type ctype,
		int clMinVer, int clMaxVer)
  : team(NULL),
    side(SIDE_LEFT),
    isCoach(ctype == CLIENT_ONLINECOACH || ctype == CLIENT_TRAINCOACH),
    data(NULL),
    msg(NULL),
    clangMinVer(clMinVer),
    clangMaxVer(clMaxVer)
{
  PARSERTRACE("entering constructor");
  team = new char[strlen(_team)+1];
  strcpy(team, _team);
  PARSERTRACE("leaving constructor");
}

Parser::~Parser( void ) {
  PARSERTRACE("entering destructor");
  delete [] team;
  PARSERTRACE("entering destructor");
}

bool
Parser::parsePlayerParam( const char * _msg ) {
  int i(0);

  PARSERTRACE("parsing player_param");

  if (strncmp(_msg, "(player_param ", 14))
    return false;

  msg = (char*)(_msg + 14);
  
  while(playerParamTokens[i].fmt != NULL) {
    switch (readToken(playerParamTokens[i++])) {
    case -1:
      return false;
    case 0:
      PARAMTRACE(playerParamTokens[i-1]);
      break;
    case 1: {
      char errmsg[MAX_SERVER_MSG_SIZE];
      int l(0);
      sscanf(msg, " (%[^)]) %n", errmsg, &l);
      msg += l;
      cerr << "Parser::parsePlayerParam: skipping unknown token at " 
	   << i << " (" << errmsg << ")" << endl;
      --i;
      break;
    }
    default:;
    }
  }

  PARSERTRACE("parsing player_param finished");

  return true;
}

bool
Parser::parseServerParam( const char * _msg ) {
  int i(0);

  PARSERTRACE("parsing server_param");
  
  if (strncmp(_msg, "(server_param ", 14))
    return false;
  
  msg = (char*)(_msg + 14);

  while(serverParamTokens[i].fmt != NULL) {
    switch (readToken(serverParamTokens[i++])) {
    case -1:
      return false;
    case 0:
      PARAMTRACE(serverParamTokens[i-1]);
      break;
    case 1: {
      //      return false;
      char errmsg[MAX_SERVER_MSG_SIZE];
      int l(0);
      sscanf(msg, " (%[^)]) %n", errmsg, &l);
      cerr << "Parser::parseServerParam: skipping unknown token at " 
	   << i << " (" << errmsg << ")" << endl;
      --i;
      msg += l;
      break;
    }
    default:;
    }
  }

  PARSERTRACE("parsing server_param finished");
  
  return true;
}

bool
Parser::parseServerString( const char * _msg, SensorData &_data ) {
  PARSERTRACE("checking message type of '" << _msg << "'");
  
  int len(0);

  data = &_data;

  data->think = false;

  // Switch on message type
  if (!strncmp(_msg, "(player_type ", 13)) {
    data->time = -1;
    data->type = SENSOR_TYPE_PLAYER_TYPE;
    msg = (char*)(_msg + 13);
    PARSERTRACE("going to parse '" << msg << "'");
    return parsePlayerType();

  } else if (sscanf(_msg, " (see %d ) %n", &data->time, &len) == 1) {
    if (len > 0) {
      data->type = SENSOR_TYPE_STRANGE;
      return true;
    }
    sscanf(_msg, " (see %*d %n", &len);
    data->type = SENSOR_TYPE_SEE;
    msg = (char*)(_msg+len);
    PARSERTRACE("going to parse '" << msg << "'");
    return parseSee();

  } else if (sscanf(_msg, " (see_global %d %n", &data->time, &len) == 1) {
    if (isCoach) { // coach receives different (see) information !
      data->type = SENSOR_TYPE_EXACT;
      msg = (char*)(_msg+len);
      PARSERTRACE("going to parse '" << msg << "'");
      return parseExact();
    }
    else { 
      cerr << "Parser: received see message for coach" << endl;
      return false;
    }

  } else if (sscanf(_msg, " (fullstate %d %n", &data->time, &len) == 1) {
    data->type = SENSOR_TYPE_ERROR;
    msg = (char*)(_msg+len);
    cerr << "Parser: fullstate message received!" << endl;
    return true;

  } else if (sscanf(_msg, " (hear %d %n", &data->time, &len) == 1) {
    // Aural information from referee, players or coach
    data->type = SENSOR_TYPE_HEAR;
    msg = (char*)(_msg+len);
    PARSERTRACE("going to parse '" << msg << "'");
    return parseHear();

  } else if (sscanf(_msg, " (sense_body %d %n", &data->time, &len) == 1) {
    // Sense body information
    data->type = SENSOR_TYPE_SENSE_BODY;
    msg = (char*)(_msg+len);
    PARSERTRACE("going to parse '" << msg << "'");
    return parseSenseBody();

  } else if (sscanf(_msg, " (ok look %d %n", &data->time, &len) == 1) {
    // exact information as requested by coach
    data->type = SENSOR_TYPE_EXACT;
    msg = (char*)(_msg+len);
    PARSERTRACE("going to parse '" << msg << "'");
    return parseExact();

  } else if (!strncmp(_msg, "(ok ", 4)) {
    // other ok replies to coach not parsed any further
    data->time = -1;
    data->type = SENSOR_TYPE_OK;
    return true;

  } else if (!strncmp(_msg, "(change_player_type ", 20)) {
    data->number = -1, data->heteroType = -1;
    sscanf(_msg, "(change_player_type %d, %d", 
	   &data->number, &data->heteroType);
    data->ourTeamChanged = data->heteroType != -1;
    data->time = -1;
    data->type = SENSOR_TYPE_PLAYER_CHANGED;
    return true;

  } else if (!strncmp(_msg, "(clang", 6)) {
    data->number = -1, data->heteroType = -1;
    sscanf(_msg, "(clang (ver ([]) %d, %d))", 
	   &data->number, &data->heteroType);
    data->time = -1;
    data->type = SENSOR_TYPE_CLANG_VER;
    return true;

  } else if (!strncmp(_msg, "(error ", 7)) {
    data->time = -1;
    data->type = SENSOR_TYPE_ERROR;
    return true;

  } else if (!strncmp(_msg, "(think)", 7)) {
    data->time = -1;
    data->type = SENSOR_TYPE_OK;
    data->think = true;
    return true;

  } else if (!strncmp(_msg, "(warning ", 9)) {
    data->time = -1;
    data->type = SENSOR_TYPE_WARNING; // treat warnings like errors
    return true;

  } 
  else {
    PARSERTRACE("could not determine message type");
    return false;
  }
}

bool
Parser::parseSenseBody( void ) {
  PARSERTRACE("parsing sense body data");
  int i(0), len(strlen(msg)+1);
  char quality[len], width[len];
  quality[0] = 0;
  width[0] = 0;
  double dist(INVALID_DIST), dir(INVALID_DIR);
  len = 0;

  Token senseBodyTokens[] = {
    {" (head_angle %lf)%n", (void*)&data->head_angle,        DT_DOUBLE},
    {" (kick %d)%n",        (void*)&data->kick_count,        DT_INT},
    {" (dash %d)%n",        (void*)&data->dash_count,        DT_INT},
    {" (turn %d)%n",        (void*)&data->turn_count,        DT_INT},
    {" (say %d)%n",         (void*)&data->say_count,         DT_INT},
    {" (turn_neck %d)%n",   (void*)&data->turn_neck_count,   DT_INT},
    {" (catch %d)%n",       (void*)&data->catch_count,       DT_INT},
    {" (move %d)%n",        (void*)&data->move_count,        DT_INT},
    {" (change_view %d)%n", (void*)&data->change_view_count, DT_INT},
    {NULL, NULL, DT_NONE}
  };

  sscanf(msg, " (view_mode %[a-z] %[a-z]) (stamina %lf %lf) (speed %lf %lf)%n",
	 quality, width, &data->stamina, &data->effort, &data->speed,
	 &data->speed_angle, &len);
  if (!len)
    return false;

  msg += len;

  while(senseBodyTokens[i].fmt != NULL) {
    switch (readToken(senseBodyTokens[i++])) {
    case -1:
      return false;
    case 0:
      PARAMTRACE(senseBodyTokens[i-1]);
      break;
    case 1: {
      char errmsg[MAX_SERVER_MSG_SIZE];
      int l(0);
      sscanf(msg, " (%[^)]) %n", errmsg, &l);
      msg += l;
      cerr << "Parser::parseSenseBody: skipping unknown token at " 
	   << i-1 << " (" << errmsg << ")" << endl;
      --i;
      break;
    }
    default:;
    }
  }

  // parse arm info
  len = 0;
  sscanf(msg, " (arm (movable %d) (expires %d) (target %lf %lf) (count %d)) %n",
	 &data->arm_movable, &data->arm_expires, &dist, &dir, &data->point_to_count, &len);
  if (!len)
    return false;


  // parse attentionto info
  msg += len;
  len = 0;
  data->attentionTo = -1;
  sscanf(msg, " (focus (target none) (count %d)) %n",
	 &data->attention_to_count, &len);
  if (!len) {
    sscanf(msg, " (focus (target l %d) (count %d)) %n",
	   &data->attentionTo, &data->attention_to_count, &len);
    data->attentionToOurTeam = (side == SIDE_LEFT); 
    if (!len) {
      sscanf(msg, " (focus (target r %d) (count %d)) %n",
	     &data->attentionTo, &data->attention_to_count, &len);
      if (!len) 
	return false;
      data->attentionToOurTeam = (side == SIDE_RIGHT); 
    }
  }

  // parse tackle info
  msg += len;
  len = 0;
  sscanf(msg, " (tackle (expires %d) (count %d)) %n",
	 &data->tackle_expires, &data->tackle_count, &len);
  if (!len)
    return false;

  // Parse view quality
  if (!strncmp(quality, "high", 4)) 
    data->my_view_quality = VIEW_QUALITY_HIGH;
  else if (!strncmp(quality, "low", 3)) 
    data->my_view_quality = VIEW_QUALITY_LOW;
  else
    return false;

  // Parse view width
  if (!strncmp(width, "narrow", 6))
    data->my_view_width = VIEW_WIDTH_NARROW;
  else if (!strncmp(width, "normal", 6))
    data->my_view_width = VIEW_WIDTH_NORMAL;
  else if (!strncmp(width, "wide", 4))
    data->my_view_width = VIEW_WIDTH_WIDE;
  else
    return false;

  // transform angles to radians
  data->head_angle = DEG_TO_RAD(data->head_angle);
  data->speed_angle = DEG_TO_RAD(data->speed_angle);

  // set arm target
  if (dist != INVALID_DIST && dir != INVALID_DIR) {
    dir = DEG_TO_RAD(dir);
    Angle dir_ang(dir);
    data->arm_target = radAng2Vector2d(dist,dir_ang);
  }
  else
    data->arm_target.setXY(0,0);

  if (data->attentionTo < -1 || data->attentionTo > 11)
    return false;

  PARSERTRACE("parsing sense body data finished");

  return true;
}

bool
Parser::parseHear( void ) {

  char sender[strlen(msg)+1];
  int len(0);
  sender[0] = 0;

  PARSERTRACE ("parsing heard message");

  // Invalidate aural information
  data->coachMsgType = CMT_NONE;
  data->message[0] = 0;
  data->mode = PLAY_MODE_UNKNOWN;
  data->sender_dir = INVALID_DIR;
  data->sender_num = -1;
  data->sender_name[0] = 0;

  sscanf(msg, " %s %n ", sender, &len);

  if (!strncmp(sender, "referee", 7)) {
    msg += len;
    if (sscanf(msg, " %[-0-9a-zA-Z .+*/?<>_()] ", data->message) <= 0)
      return false;

    data->mode = parsePlaymode(data->message, side);

    if (data->mode == PLAY_MODE_UNKNOWN) {
      PARSERTRACE ("message from the coach");
      data->sender = SENDER_COACH;
    } else {
      PARSERTRACE("message from the referee");
      data->sender = SENDER_REFEREE;
      if (data->mode == PLAY_MODE_GOAL_US || data->mode == PLAY_MODE_GOAL_THEM)
	if (sscanf(data->message, "goal_%*c_%d", &data->score) <= 0) 
	  return false;
    }
    
    return true;

  } else if(!strncmp(sender, "self", 4)) {
    data->sender = SENDER_SELF;
    PARSERTRACE("message from self");
    msg += len;
    if (sscanf(msg, " \"%[-0-9a-zA-Z .+*/?<>_()]\" ", data->message) <= 0)
      return false;

    return true;

  } else if(!strncmp(sender, "online_coach_left", 17) ||
	    !strncmp(sender, "online_coach_right", 18)) {
    PARSERTRACE("message from " << sender);
    data->sender = SENDER_ONLINECOACH;
    msg += len;
    if (!parseCoachMsg())
      return false;

    return true;
  }
  if (sscanf(msg, " our %d)", &data->sender_num) == 1) {
    data->sender = SENDER_PLAYER_US;
    PARSERTRACE("empty message from our player");

  } else if (!strncmp(msg, "opp)", 4)) {
    data->sender = SENDER_PLAYER_THEM;
    PARSERTRACE("empty message from opp player");

  } else if (sscanf(msg, "%lf our %d \"%[-0-9a-zA-Z .+*/?<>_()]\"", 
	     &data->sender_dir, &data->sender_num, data->message) == 3) {
    data->sender = SENDER_PLAYER_US;
    data->sender_dir = DEG_TO_RAD(data->sender_dir);
    PARSERTRACE("message from our player with a certain direction");


  } else if (sscanf(msg, "%lf opp \"%[-0-9a-zA-Z .+*/?<>_()]\"", 
	     &data->sender_dir, data->message) == 2) {
    data->sender = SENDER_PLAYER_THEM;
    data->sender_dir = DEG_TO_RAD(data->sender_dir);
    PARSERTRACE("message from opp player with a certain direction");


  } else if (sscanf(msg, "%lf \"%[-0-9a-zA-Z .+*/?<>_()]\"", 
	     &data->sender_dir, data->message) == 2) {
    data->sender = SENDER_DIR;
    data->sender_dir = DEG_TO_RAD(data->sender_dir);
    PARSERTRACE("message from a certain direction");

  } else if (sscanf(msg, "(%[-0-9a-zA-Z .+*/?<>_\"]) \"%[-0-9a-zA-Z .+*/?<>_()]\"", 
		    data->sender_name, data->message) == 2) {
    data->sender = SENDER_PLAYER;
    PARSERTRACE("message from a named player");

  } else
    return false;

  PARSERTRACE ("parsing heard finished");

  return true;
}

bool
Parser::parseSee( void ) {
  PARSERTRACE("parsing see data");

  int pnum(0), gnum(0), fnum(0), lnum(0);
  char name[strlen(msg)+1];

  // Invalidate visual information
  data->ball.valid = false;
  for (int i = 0; i < NUM_GOALS; ++i)
    data->goal[i].valid = false;
  for (int i = 0; i < 21; ++i)
    data->player[i].valid = false;
  for (int i = 0; i < NUM_FLAGS; ++i)
    data->flag[i].valid = false;
  for (int i = 0; i < NUM_LINES; ++i)
    data->line[i].valid = false;
  
  while (true) {
    
    /* Format of see msgs
       ((p "<TEAMNAME>" <UNUM>) <DIST> <DIR> <DISTCHG> <DIRCHG> \
         <BDIR> <HDIR> [rcssserver-<POINTDIR>] [t])
			    
       ((p "<TEAMNAME>") <DIST> <DIR> [rcssserver-<POINTDIR>] [t])
     */

    int l(0), len(0), end(0);
    name[0] = 0;
    sscanf(msg, " ( ( %[^)] ) %n %*[^)] ) %n ) %n ", name ,&l, &len, &end);
    if (l == 0 || len == 0) {
      cerr << "Parser: malformed object: l=" << l << " len=" << len << endl;
      cerr << "'" << msg << "'" << endl;
      return false;
    }
    
    PARSERTRACE("object name: '" << name << "'");

    // skip object name
    msg += l;

    // scan data
    SensorObject so = {true, INVALID_DIST, INVALID_DIR, INVALID_DIST, INVALID_DIR};
    double face_dir(INVALID_DIR), body_dir(INVALID_DIR), pointto_dir(INVALID_DIR);
    char tackle(0);

    sscanf(msg, " %lf %lf ", &so.dist, &so.dir);
    
    // If only one parameter is given, it's the direction, not the distance
    // (when receiving low quality visual information)
    if (so.dir == INVALID_DIR) {
      so.dir = so.dist;
      so.dist = INVALID_DIST;
    } else {
      sscanf(msg, " %*f %*f %lf %lf", &so.dist_change, &so.dir_change);
      
      if (so.dist_change == INVALID_DIST) {
	sscanf(msg, " %*f %*f %c", &tackle);

      } else {
	if (so.dir_change == INVALID_DIR) {
	  pointto_dir = so.dist_change;
	  so.dist_change = INVALID_DIST;
	  sscanf(msg, " %*f %*f %*f %c", &tackle);

	} else {
	  sscanf(msg, "%*f %*f %*f %*f %lf %lf %lf %c", 
		 &body_dir, &face_dir, &pointto_dir, &tackle);
	}
      }
    }

    // move to end of object
    msg += len-l;

    PARSERTRACE("object values: " << so.dist << " " << so.dir << " " 
		<< so.dist_change << " "<< so.dir_change << " " 
		<< body_dir << " " << face_dir << " " 
		<< pointto_dir << " " << tackle);

    // convert angles from deg to rad
    if (IsDirValid(so.dir))
      so.dir = DEG_TO_RAD(so.dir);
    if (IsDirValid(so.dir_change))
      so.dir_change = DEG_TO_RAD(so.dir_change);
    if (IsDirValid(body_dir))
      body_dir = DEG_TO_RAD(body_dir);
    if (IsDirValid(face_dir))
      face_dir = DEG_TO_RAD(face_dir);
    if (IsDirValid(pointto_dir))
      pointto_dir = DEG_TO_RAD(pointto_dir);

    // capital letters are felt objects
    name[0] = tolower(name[0]);

    switch (determineObject(name)) {
    case OT_PLAYER:
      if (pnum > 20) { // more than 22 (=21+me) players ??
	cerr << "Parser: to manny players!" << endl;
	return false;
      }
      (SensorObject&)data->player[pnum] = so;
      data->player[pnum].body_dir = body_dir;
      data->player[pnum].face_dir = face_dir;
      data->player[pnum].pointto_dir = pointto_dir;
      data->player[pnum].tackle = (tackle == 't');
      parsePlayer(name, data->player[pnum]);
      PARSERTRACE("Player values: " << data->player[pnum].valid << " "
		  << data->player[pnum].dist << " " << data->player[pnum].dir << " "
		  << data->player[pnum].dist_change << " " << data->player[pnum].dir_change << " "
		  << data->player[pnum].body_dir << " " << data->player[pnum].face_dir
		  << " " << data->player[pnum].pointto_dir
		  << " " << data->player[pnum].tackle);
      ++pnum;
      break;
    case OT_BALL:
      (SensorObject&)data->ball = so;
      PARSERTRACE("Ball values: " << data->ball.valid << " "
		  << data->ball.dist << " " << data->ball.dir << " "
		  << data->ball.dist_change << " " << data->ball.dir_change);
      break;
    case OT_GOAL:
      if (gnum > 1) { // more than 2 goals ??
	cerr << "Parser: to manny goals!" << endl;
	return false;
      }
      (SensorObject&)data->goal[gnum] = so;
      parseGoal(name, data->goal[gnum]);
      PARSERTRACE("Goal values: " << data->goal[gnum].valid << " "
		  << data->goal[gnum].dist << " " << data->goal[gnum].dir << " "
		  << data->goal[gnum].dist_change << " " << data->goal[gnum].dir_change);
      ++gnum;
      break;
    case OT_FLAG:
      if (fnum >= NUM_FLAGS) {
	cerr << "Parser: to manny flags!" << endl;
	return false;
      }
      parseFlag(name, data->flag[fnum]);
      (SensorObject&)data->flag[fnum] = so;
      PARSERTRACE("Flag values: " << data->flag[fnum].valid << " "
		  << data->flag[fnum].dist << " " << data->flag[fnum].dir << " "
		  << data->flag[fnum].dist_change << " " << data->flag[fnum].dir_change);
      ++fnum;
      break;
    case OT_LINE:
      if (lnum >= NUM_LINES) {
	cerr << "Parser: to manny lines!" << endl;
	return false;
      }
      (SensorObject&)data->line[lnum] = so;
      PARSERTRACE("Line values: " << data->line[lnum].valid << " "
		     << data->line[lnum].dist << " " << data->line[lnum].dir << " "
		     << data->line[lnum].dist_change << " " << data->line[lnum].dir_change);
      parseLine(name, data->line[lnum]);
      ++lnum;
      break;
    default:
      cerr << "Parser: unknown object type" << endl;
      return false;
    }

    if (end != 0)
      break;

  };

  PARSERTRACE("parsing see data finished");

  return true;
}

bool
Parser::parseExact( void ) {
  PARSERTRACE("parsing exact data");

  int pnum(0), gnum(0);
  char name[strlen(msg)+1];
  name[0] = 0;
  // Invalidate visual information
  data->exactball.valid = false;
  for (int i = 0; i < NUM_GOALS; ++i)
    data->exactgoal[i].valid = false;
  for (int i = 0; i < 22; ++i)
    data->exactplayer[i].valid = false;
  
  while (true) {
    
    int l(0), len(0), end(0);
    name[0] = 0;
    sscanf(msg, " ( ( %[^)] ) %n %*[^)] ) %n ) %n ", name ,&l, &len, &end);
    if (l == 0 || len == 0) {
      cerr << "Parser: malformed object: l=" << l << " len=" << len << endl;
      cerr << "'" << msg << "'" << endl;
      return false;
    }
    
    PARSERTRACE("object name: '" << name << "'");

    // skip object name
    msg += l;

    // scan data
    ExactObject eo = {true, 0., 0., 0., 0.};
    double face_dir(INVALID_DIR), body_dir(INVALID_DIR), pointto_dir(INVALID_DIR);
    char tackle(0); int l1(0);

    sscanf(msg, " %lf %lf %lf %lf %lf %lf %n", &eo.x, &eo.y, &eo.x_vel, &eo.y_vel,
	   &body_dir, &face_dir, &l1);
  
    if (sscanf(msg+l1, "%lf %c", &pointto_dir, &tackle) <= 0) {
      pointto_dir = INVALID_DIR;
      tackle = (char)0;
      if (sscanf(msg+l1, "%c", &tackle) <= 0)
	tackle = (char)0;
    }
  
    // move to end of object
    msg += len-l;

    // convert angles from deg to rad
    if (IsDirValid(body_dir))
      body_dir = DEG_TO_RAD(body_dir);
    if (IsDirValid(face_dir))
      face_dir = DEG_TO_RAD(face_dir);
    if (IsDirValid(pointto_dir))
      pointto_dir = DEG_TO_RAD(pointto_dir);

    PARSERTRACE("object values: " << eo.x << " " << eo.y << " " 
		<< eo.x_vel << " "<< eo.y_vel << " " 
		<< body_dir << " " << face_dir << " " << pointto_dir
		<< " " << tackle);

    // capital letters are felt objects
    name[0] = tolower(name[0]);

    switch (determineObject(name)) {
    case OT_PLAYER:
      if (pnum > 21) {// more than 22 players ??
	cerr << "Parser: to manny players!" << endl;
	return false;
      }
      (ExactObject&)data->exactplayer[pnum] = eo;
      data->exactplayer[pnum].body_dir = body_dir;
      data->exactplayer[pnum].face_dir = face_dir;
      data->exactplayer[pnum].pointto_dir = pointto_dir;
      data->exactplayer[pnum].tackle = (tackle == 't');
      parsePlayer(name, data->exactplayer[pnum]);
      ++pnum;
      break;
    case OT_BALL:
      (ExactObject&)data->exactball = eo;
      break;
    case OT_GOAL:
      if (gnum > 1) {// more than 2 goals ??
	cerr << "Parser: to manny goals!" << endl;
	return false;
      }
      (ExactObject&)data->exactgoal[gnum] = eo;
      parseGoal(name, data->exactgoal[gnum]);
      ++gnum;
      break;
    default:
      return false;
    }

    if (end != 0)
      break;

  };

  PARSERTRACE("parsing exact data finished");

  return true;
}

bool
Parser::parsePlayerType( void ) {
  int i(0);

  Token playerTypeTokens[] = {
    {" (id %d)%n",                (void*)&data->playerType.player_type,      DT_INT},
    {" (player_speed_max %lf)%n", (void*)&data->playerType.player_speed_max, DT_DOUBLE},
    {" (stamina_inc_max %lf)%n",  (void*)&data->playerType.stamina_inc_max,  DT_DOUBLE},
    {" (player_decay %lf)%n",     (void*)&data->playerType.player_decay,     DT_DOUBLE},
    {" (inertia_moment %lf)%n",   (void*)&data->playerType.inertia_moment,   DT_DOUBLE},
    {" (dash_power_rate %lf)%n",  (void*)&data->playerType.dash_power_rate,  DT_DOUBLE},
    {" (player_size %lf)%n",      (void*)&data->playerType.player_size,      DT_DOUBLE},
    {" (kickable_margin %lf)%n",  (void*)&data->playerType.kickable_margin,  DT_DOUBLE},
    {" (kick_rand %lf)%n",        (void*)&data->playerType.kick_rand,        DT_DOUBLE},
    {" (extra_stamina %lf)%n",    (void*)&data->playerType.extra_stamina,    DT_DOUBLE},
    {" (effort_max %lf)%n",       (void*)&data->playerType.effort_max,       DT_DOUBLE},
    {" (effort_min %lf)%n",       (void*)&data->playerType.effort_min,       DT_DOUBLE},
    {NULL, NULL, DT_NONE}
  };

  PARSERTRACE("parsing player_type");

  while(playerTypeTokens[i].fmt != NULL) {
    switch (readToken(playerTypeTokens[i++])) {
    case -1:
      return false;
    case 0:
      PARAMTRACE(playerTypeTokens[i-1]);
      break;
    case 1: {
      char errmsg[MAX_SERVER_MSG_SIZE];
      int l(0);
      sscanf(msg, " (%[^)]) %n", errmsg, &l);
      msg += l;
      cerr << "Parser::parsePlayerType: skipping unknown token at " 
	   << i-1 << " (" << errmsg << ")" << endl;
      --i;
      break;
    }
    default:;
    }
  }

  PARSERTRACE("parsing player_type finished");

  return true;
}

void
Parser::parsePlayer( const char * name, team_number &teamn,
		     int &number, goalie_player &goalie)
{
  PARSERTRACE("parsing player");
  number = -1;
  char teamname[strlen(name)+1];
  char g[strlen(name)+1];
  teamname[0] = (char)0;
  g[0] = (char)0;

  sscanf(name, " p \"%[-_a-zA-Z0-9]\" %d %[a-z] ", teamname, &number, g);
  
  if (!strcmp(teamname, team))
    teamn = TEAM_US;
  else if (strlen(teamname) > 0)
    teamn = TEAM_THEM;
  else
    teamn = TEAM_UNKNOWN;

  if (number != -1)
    if (!strncmp(g, "goalie", 6))
      goalie = GOALIE_TRUE;
    else
      goalie = GOALIE_FALSE;
  else
    goalie = GOALIE_UNKNOWN;
}

void
Parser::parseGoal( const char * s, goal_type &which, team_number &team ) {
  PARSERTRACE("parsing goal");
  if (!strncmp(s, "g l", 3)) {
    which = GOAL_LEFT;
    team = side == SIDE_LEFT ? TEAM_US : TEAM_THEM;
  } else if (!strncmp(s, "g r", 3)) {
    which = GOAL_RIGHT;
    team = side == SIDE_RIGHT ? TEAM_US : TEAM_THEM;
  } else {
    which = GOAL_UNKNOWN;
    team = TEAM_UNKNOWN;
  }
}

void
Parser::parseFlag( const char * name, SensorFlag &flag ) {
  PARSERTRACE("parsing flag");
  
  if (!strncmp(name, "f t l 50", 8))
    flag.which = FLAG_TOP_LEFT_50;
  else if (!strncmp(name, "f t l 40", 8))
    flag.which = FLAG_TOP_LEFT_40;
  else if (!strncmp(name, "f t l 30", 8))
    flag.which = FLAG_TOP_LEFT_30;
  else if (!strncmp(name, "f t l 20", 8))
    flag.which = FLAG_TOP_LEFT_20;
  else if (!strncmp(name, "f t l 10", 8))
    flag.which = FLAG_TOP_LEFT_10;
  else if (!strncmp(name, "f t r 10", 8))
    flag.which = FLAG_TOP_RIGHT_10;
  else if (!strncmp(name, "f t r 20", 8))
    flag.which = FLAG_TOP_RIGHT_20;
  else if (!strncmp(name, "f t r 30", 8))
    flag.which = FLAG_TOP_RIGHT_30;
  else if (!strncmp(name, "f t r 40", 8))
    flag.which = FLAG_TOP_RIGHT_40;
  else if (!strncmp(name, "f t r 50", 8))
    flag.which = FLAG_TOP_RIGHT_50;
  
  else if (!strncmp(name, "f b l 50", 8))
    flag.which = FLAG_BOTTOM_LEFT_50;
  else if (!strncmp(name, "f b l 40", 8))
    flag.which = FLAG_BOTTOM_LEFT_40;
  else if (!strncmp(name, "f b l 30", 8))
    flag.which = FLAG_BOTTOM_LEFT_30;
  else if (!strncmp(name, "f b l 20", 8))
    flag.which = FLAG_BOTTOM_LEFT_20;
  else if (!strncmp(name, "f b l 10", 8))
    flag.which = FLAG_BOTTOM_LEFT_10;
  else if (!strncmp(name, "f b 0", 5))
    flag.which = FLAG_BOTTOM_0;
  else if (!strncmp(name, "f b r 10", 8))
    flag.which = FLAG_BOTTOM_RIGHT_10;
  else if (!strncmp(name, "f b r 20", 8))
    flag.which = FLAG_BOTTOM_RIGHT_20;
  else if (!strncmp(name, "f b r 30", 8))
    flag.which = FLAG_BOTTOM_RIGHT_30;
  else if (!strncmp(name, "f b r 40", 8))
    flag.which = FLAG_BOTTOM_RIGHT_40;
  else if (!strncmp(name, "f b r 50", 8))
    flag.which = FLAG_BOTTOM_RIGHT_50;

  else if (!strncmp(name, "f l t 30", 8))
    flag.which = FLAG_LEFT_TOP_30;
  else if (!strncmp(name, "f l t 20", 8))
    flag.which = FLAG_LEFT_TOP_20;
  else if (!strncmp(name, "f l t 10", 8))
    flag.which = FLAG_LEFT_TOP_10;
  else if (!strncmp(name, "f l b 10", 8))
    flag.which = FLAG_LEFT_BOTTOM_10;
  else if (!strncmp(name, "f l b 20", 8))
    flag.which = FLAG_LEFT_BOTTOM_20;
  else if (!strncmp(name, "f l b 30", 8))
    flag.which = FLAG_LEFT_BOTTOM_30;
  
  else if (!strncmp(name, "f r t 30", 8))
    flag.which = FLAG_RIGHT_TOP_30;
  else if (!strncmp(name, "f r t 20", 8))
    flag.which = FLAG_RIGHT_TOP_20;
  else if (!strncmp(name, "f r t 10", 8))
    flag.which = FLAG_RIGHT_TOP_10;
  else if (!strncmp(name, "f r b 10", 8))
    flag.which = FLAG_RIGHT_BOTTOM_10;
  else if (!strncmp(name, "f r b 20", 8))
    flag.which = FLAG_RIGHT_BOTTOM_20;
  else if (!strncmp(name, "f r b 30", 8))
    flag.which = FLAG_RIGHT_BOTTOM_30;

  else if (!strncmp(name, "f p l t", 7))
    flag.which = FLAG_P_LEFT_TOP;
  else if (!strncmp(name, "f p l c", 7))
    flag.which = FLAG_P_LEFT_CENTER;
  else if (!strncmp(name, "f p l b", 7))
    flag.which = FLAG_P_LEFT_BOTTOM;
  else if (!strncmp(name, "f p r t", 7))
    flag.which = FLAG_P_RIGHT_TOP;
  else if (!strncmp(name, "f p r c", 7))
    flag.which = FLAG_P_RIGHT_CENTER;
  else if (!strncmp(name, "f p r b", 7))
    flag.which = FLAG_P_RIGHT_BOTTOM;
  
  
  else if (!strncmp(name, "f g l t", 7))
    flag.which = FLAG_GOAL_LEFT_TOP;
  else if (!strncmp(name, "f g l b", 7))
    flag.which = FLAG_GOAL_LEFT_BOTTOM;
  else if (!strncmp(name, "f g r t", 7))
    flag.which = FLAG_GOAL_RIGHT_TOP;
  else if (!strncmp(name, "f g r b", 7))
    flag.which = FLAG_GOAL_RIGHT_BOTTOM;

  else if (!strncmp(name, "f r 0", 5))
    flag.which = FLAG_RIGHT_0;
  else if (!strncmp(name, "f l 0", 5))
    flag.which = FLAG_LEFT_0;
  else if (!strncmp(name, "f t 0", 5))
    flag.which = FLAG_TOP_0;
  else if (!strncmp(name, "f l t", 5))
    flag.which = FLAG_TOP_LEFT;
  else if (!strncmp(name, "f c t", 5))
    flag.which = FLAG_TOP_CENTER;
  else if (!strncmp(name, "f r t", 5))
    flag.which = FLAG_TOP_RIGHT;
  else if (!strncmp(name, "f l b", 5))
    flag.which = FLAG_BOTTOM_LEFT;
  else if (!strncmp(name, "f c b", 5))
    flag.which = FLAG_BOTTOM_CENTER;
  else if (!strncmp(name, "f r b", 5))
    flag.which = FLAG_BOTTOM_RIGHT;

  else if (!strncmp(name, "f c", 3))
    flag.which = FLAG_CENTER;

  else
    flag.which = FLAG_UNKNOWN;
  PARSERTRACE("flag is: ";
	      switch (flag.which) {
	      case FLAG_UNKNOWN: cout << "unknown"; break;
	      case FLAG_TOP_LEFT: cout << "top left"; break;
	      case FLAG_TOP_CENTER: cout << "top center"; break;
	      case FLAG_TOP_RIGHT: cout << "top right"; break;
	      case FLAG_BOTTOM_LEFT: cout << "bottom left"; break;
	      case FLAG_BOTTOM_CENTER: cout << "bottom center"; break;
	      case FLAG_BOTTOM_RIGHT: cout << "bottom right"; break;
	      case FLAG_P_LEFT_TOP: cout << "penalty left top"; break;
	      case FLAG_P_LEFT_CENTER: cout << "penalty left center"; break;
	      case FLAG_P_LEFT_BOTTOM: cout << "penalty left bottom"; break;
	      case FLAG_P_RIGHT_TOP: cout << "penalty right top"; break;
	      case FLAG_P_RIGHT_CENTER: cout << "penalty right center"; break;
	      case FLAG_P_RIGHT_BOTTOM: cout << "penalty right bottom"; break;
	      case FLAG_CENTER: cout << "center flag"; break;
	      case FLAG_GOAL_LEFT_TOP: cout << "goal left top"; break;
	      case FLAG_GOAL_LEFT_BOTTOM: cout << "goal left bottom"; break;
	      case FLAG_GOAL_RIGHT_TOP: cout << "goal right top"; break;
	      case FLAG_GOAL_RIGHT_BOTTOM: cout << "goal right bottom"; break;
	      case FLAG_TOP_LEFT_50: cout << "top left 50"; break;
	      case FLAG_TOP_LEFT_40: cout << "top left 40"; break;
	      case FLAG_TOP_LEFT_30: cout << "top left 30"; break;
	      case FLAG_TOP_LEFT_20: cout << "top left 20"; break;
	      case FLAG_TOP_LEFT_10: cout << "top left 10"; break;
	      case FLAG_TOP_0: cout << "top 0"; break;
	      case FLAG_TOP_RIGHT_10: cout << "top right 10"; break;
	      case FLAG_TOP_RIGHT_20: cout << "top right 20"; break;
	      case FLAG_TOP_RIGHT_30: cout << "top right 30"; break;
	      case FLAG_TOP_RIGHT_40: cout << "top right 40"; break;
	      case FLAG_TOP_RIGHT_50: cout << "top right 50"; break;
	      case FLAG_BOTTOM_LEFT_50: cout << "bottom left 50"; break;
	      case FLAG_BOTTOM_LEFT_40: cout << "bottom left 40"; break;
	      case FLAG_BOTTOM_LEFT_30: cout << "bottom left 30"; break;
	      case FLAG_BOTTOM_LEFT_20: cout << "bottom left 20"; break;
	      case FLAG_BOTTOM_LEFT_10: cout << "bottom left 10"; break;
	      case FLAG_BOTTOM_0: cout << "bottom 0"; break;
	      case FLAG_BOTTOM_RIGHT_10: cout << "bottom right 10"; break;
	      case FLAG_BOTTOM_RIGHT_20: cout << "bottom right 20"; break;
	      case FLAG_BOTTOM_RIGHT_30: cout << "bottom right 30"; break;
	      case FLAG_BOTTOM_RIGHT_40: cout << "bottom right 40"; break;
	      case FLAG_BOTTOM_RIGHT_50: cout << "bottom right 50"; break;
	      case FLAG_LEFT_TOP_30: cout << "left top 30"; break;
	      case FLAG_LEFT_TOP_20: cout << "left top 20"; break;
	      case FLAG_LEFT_TOP_10: cout << "left top 10"; break;
	      case FLAG_LEFT_0: cout << "left 0"; break;
	      case FLAG_LEFT_BOTTOM_10: cout << "left bottom 10"; break;
	      case FLAG_LEFT_BOTTOM_20: cout << "left bottom 20"; break;
	      case FLAG_LEFT_BOTTOM_30: cout << "left bottom 30"; break;
	      case FLAG_RIGHT_TOP_30: cout << "right top 30"; break;
	      case FLAG_RIGHT_TOP_20: cout << "right top 20"; break;
	      case FLAG_RIGHT_TOP_10: cout << "right top 10"; break;
	      case FLAG_RIGHT_0: cout << "right 0"; break;
	      case FLAG_RIGHT_BOTTOM_10: cout << "right bottom 10"; break;
	      case FLAG_RIGHT_BOTTOM_20: cout << "right bottom 20"; break;
	      case FLAG_RIGHT_BOTTOM_30: cout << "right bottom 30"; break;
	      default:;
	      }; cout);
}

void
Parser::parseLine( const char * name, SensorLine &line ) {
  PARSERTRACE("parsing line");
  if (!strncmp(name, "l t", 3))
    line.which = LINE_TOP;
  else if (!strncmp(name, "l l", 3))
    line.which = LINE_LEFT;
  else if (!strncmp(name, "l r", 3))
    line.which = LINE_RIGHT;
  else if (!strncmp(name, "l b", 3))
    line.which = LINE_BOTTOM;
  else
    line.which = LINE_UNKNOWN;
  PARSERTRACE("line is: ";
	      switch (line.which) {
	      case LINE_TOP: cout << "top"; break;
	      case LINE_LEFT: cout << "left"; break;
	      case LINE_RIGHT: cout << "right"; break;
	      case LINE_BOTTOM: cout << "bottom"; break;
	      default:;
	      }; cout);
}

bool
Parser::parseCoachMsg( void ) {

  //  int len(0);

  PARSERTRACE("parsing online coach message...");

//    sscanf(msg, " %d %n", &data->time, &len);
//    if (!len)
//      return false;

//    msg += len;

  if (sscanf(msg, " (freeform \"%[-0-9a-zA-Z .+*/?<>_]\") ", data->message) == 1) {
    COACHTRACE ("recvd freeform message from olcoach");
    data->coachMsgType = CMT_FREEFORM;
  } 

  else if (!strncmp(msg,"(advice", 7)) {
    msg += 7;
    strncpy(data->message, msg, MAX_SAY_MSG_SIZE);
    COACHTRACE ("recvd advice message from olcoach");
    data->coachMsgType = CMT_ADVICE;
  } 

  else if (!strncmp(msg, "(info", 5)) {
    msg += 5;
    strncpy(data->message, msg, MAX_SAY_MSG_SIZE);
    COACHTRACE ("recvd info message from olcoach");
    data->coachMsgType = CMT_INFO;
  } 

  else if (!strncmp(msg, "(meta", 5)) {
    msg += 5;
    strncpy(data->message, msg, MAX_SAY_MSG_SIZE);
    COACHTRACE ("recvd meta message from olcoach");
    data->coachMsgType = CMT_META;
  } 

  else if (!strncmp(msg, "(define", 7)) {
    msg += 7;
    strncpy(data->message, msg, MAX_SAY_MSG_SIZE);
    COACHTRACE ("recvd define message from olcoach");
    data->coachMsgType = CMT_DEFINE;
  }

  else if (!strncmp(msg, "(unsupported_clang)", 19)) {
    msg += 19;
    COACHTRACE ("recvd unsupported_clang from olcoach");
    data->coachMsgType = CMT_NONE;
  }

  else {
    cerr << "Parser: recvd unknown message from olcoach:" << endl;
    cerr << '"'<< data->message << '"' << endl;
    data->coachMsgType = CMT_NONE;
    return false;
  }
  
  return true;
}

int
Parser::readToken( Token & token ) {
  int l(0);
  
  switch (token.dt) {
  case DT_INT: {
    if (sscanf(msg, token.fmt, (int*)token.ptr, &l) != 1)
      return 1;
    break;
  }
  case DT_DOUBLE: {
    if (sscanf(msg, token.fmt, (double*)token.ptr, &l) != 1)
      return 1;
    break;
  }
  case DT_STRING: {
    if (sscanf(msg, token.fmt, (char*)token.ptr, &l) != 1)
      return 1;
    break;
  }
  case DT_BOOL: {
    short b(0);
    if (sscanf(msg, token.fmt, &b, &l) != 1)
      return 1;
    *(bool*)token.ptr = !!b;
    break;
  }
  case DT_NONE: {
    sscanf(msg, token.fmt, &l);
    if (!l)
      return 1;
    break;
  }
  default:
    cerr << "Parser::readToken: unknown data type " << token.dt << endl;
    return -1;
  }

  msg += l;

  return 0;
}

Parser::ObjectType
Parser::determineObject( const char * name ) {
  if (!strncmp(name, "b", 1)) {
    PARSERTRACE("Object is a ball");
    return OT_BALL;
  }
  else if (!strncmp(name, "g", 1)) {
    PARSERTRACE("Object is a goal");
    return OT_GOAL;
  }
  else if (!strncmp(name, "p", 1)) {
    PARSERTRACE("Object is a player");
    return OT_PLAYER;
  }
  else if (!strncmp(name, "f", 1)) {
    PARSERTRACE("Object is a flag");
    return OT_FLAG;
  }
  else if (!strncmp(name, "l", 1)) {
    PARSERTRACE("Object is a line");
    return OT_LINE;
  }
  else {
    PARSERTRACE("Unknown Object");
    return OT_NONE;
  }
}
