#ifndef PARSER_H
#define PARSER_H

#include "sensor.h"
#include "mrbtypes.h"

//#define USE_PARSERTRACE
//#define USE_PARAMTRACE
//#define USE_COACHTRACE

class Parser {

public:

  Parser( const char * team, const client_type ctype,
	  int clangMinVer = 7, int clangMaxVer = 7);
  ~Parser( void );

  bool parseServerParam( const char * msg );
  bool parsePlayerParam( const char * msg );

  bool parseServerString( const char * msg, SensorData &data );
  static play_mode parsePlaymode( const char * msg, const playing_side side );

  inline void setSide( const playing_side _side ) { side = _side; };

  inline int getCLangMinVer( void ) { return clangMinVer; };
  inline int getCLangMaxVer( void ) { return clangMaxVer; };
  
private:
  
  enum ObjectType {
    OT_NONE = 0,
    OT_PLAYER,
    OT_BALL,
    OT_GOAL,
    OT_FLAG,
    OT_LINE
  };

  enum DataType {
    DT_NONE = 0,
    DT_DOUBLE,
    DT_INT,
    DT_STRING,
    DT_BOOL
  };

  struct Token {
    char * fmt;
    void * ptr;
    DataType dt;
  };

  static Token serverParamTokens[], playerParamTokens[];
  char * team;
  playing_side side;
  bool isCoach;
  SensorData * data;
  char * msg;

  int clangMinVer, clangMaxVer;

  bool parsePlayerType( void );
  bool parseHear( void );
  bool parseSenseBody( void );
  bool parseSee( void );
  bool parseExact( void );
  bool parseSay( void );

  inline
  void parsePlayer( const char * s, SensorPlayer &player ) {
    parsePlayer(s, player.team, player.number, player.goalie);
  };

  inline
  void parsePlayer( const char * s, ExactPlayer &player ) {
    parsePlayer(s, player.team, player.number, player.goalie);
  };

  inline
  void parseGoal( const char * s, SensorGoal &goal ) {
    parseGoal(s, goal.which, goal.team);
  };

  inline
  void parseGoal( const char * s, ExactGoal &goal ) {
    parseGoal(s, goal.which, goal.team);
  };

  void parsePlayer( const char * s, team_number &team,
		    int &number, goalie_player &goalie );
  void parseGoal( const char * , goal_type &which, team_number &team );
  void parseFlag( const char * , SensorFlag &);
  void parseLine( const char * , SensorLine &);
  bool parseCoachMsg( void );

  int readToken( Token & token );
  ObjectType determineObject( const char * );
};

#endif //PARSER_H






