//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef PERM_H
#define PERM_H

//////////////////////////////////////////////////////////////////////
//
//    class Permutation 
//
//////////////////////////////////////////////////////////////////////

/*
  This class is just a vector without write permissions
  to store a permutation
 */

class Permutation {

public:

  Permutation( const Permutation & );

  ~Permutation( void ) { delete [] p; };

  inline char operator[]( const char ) const;

  Permutation &operator=( Permutation & );

private:

  char *p, n;

  friend class Permutations;

  Permutation( const char n = 0 );

  char* operator= ( char* perm );

};


//////////////////////////////////////////////////////////////////////
//
//    class Permutations
//
//////////////////////////////////////////////////////////////////////

/*
  This class is a vector of Permutation without to store all possible 
  permutations
 */

class Permutations {
  
public:

  // if generate == false you have to generate() the permutations
  // before accessing them
  Permutations( const char n, const bool generate = true );

  Permutations( const Permutations & );

  // returns permutations of n
  // if n > numPerm() the return value is *this
  Permutations of( const char n );

  ~Permutations( void ) { delete [] p; };

  // generate all permutations
  void generate( void );

  // number of permutations in class
  unsigned int numPerm( void ) const { return fac; };

  // calc n!
  inline unsigned int nfac( const int ) const;

  inline Permutation &operator[] ( const unsigned int ) const;

private:

  char n;
  unsigned int fac;
  bool generated;
  Permutation *p;

  void genPerm( char [], char, unsigned int & );
};

#endif // PERM_H
