//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// $Id: audiomgr.h,v 1.15 2001/09/11 10:01:38 goetz Exp $
//
// $manager: goetz$
//
// Manager for handling audio (spoken) messages with distribution of
// worldmodel.
//
//////////////////////////////////////////////////////////////////////

#ifndef _AUDIOMGR_H_
#define _AUDIOMGR_H_

#include <iostream.h>
#include <stdio.h>
#include <string.h>

#include "Ensure.h"
#include "constants.h"
#include "coordinates.h"
#include "snapshot.h"
#include "audiomsg.h"

//////////////////////////////////////////////////////////////////////
//
// Constants
//
//////////////////////////////////////////////////////////////////////

const int DEFAULT_AUM_QUEUE_SIZE= 20;

//////////////////////////////////////////////////////////////////////
//
// Queuing mechanism for messages (for received messages)
//
//////////////////////////////////////////////////////////////////////

class AudioMsgQueue
{
public:
  /** create message queue with a buffer for <no> messages */
  AudioMsgQueue( const unsigned int no);
  
  ~AudioMsgQueue();
  /** returns an audio message, false if there was none left */
  bool getNextMsg( AudioMsg &msg);
  /** stores a msg into the queue; false if unread msg overwritten */
  void storeMessage(const AudioMsg &msg);
private:
  AudioMsg *queue;
  unsigned int queueLen;
  unsigned int readMsg;
  unsigned int writeMsg;
  unsigned int numMessages;
}; // class AudioMsgQueue

//////////////////////////////////////////////////////////////////////
//
// The manager to control the message flow
//
//////////////////////////////////////////////////////////////////////

class AudioMgr
{
public:
  /** myPlayerNum: 1..11, use 0 for coach only! */
  AudioMgr(const unsigned int myPlayerNum= (unsigned int)(-1), 
	   const bool sideIsLeft= false,
	   const char msgPrefix[4]= DEFAULT_MSG_PREFIX,
	   const unsigned int queueSize= DEFAULT_AUM_QUEUE_SIZE);
  ~AudioMgr();

  /** set player's number and side */
  void setStates(int number, bool sideIsLeft);

  /** encode an audio message */
  const char* SpeakTo( AudioMsg &msg);
  bool getNextMsg( AudioMsg &msg);
  /** store a received message and check checksums */
  bool storeMsg(const AudioMsg &msg);
  
  /** has the agent heard a msg in step stime or stime-1 ? */
  bool heardMsgRecently(unsigned int stime);
private:
  unsigned int myNumber;
  bool mySideIsLeft;
  AudioMsgQueue *myQueue;
  char sendMessage[MAX_SAY_MSG_SIZE];
  char m_prefix[4];
  unsigned int timeOfLastHeard;
}; // class AudioMgr

#endif
