//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
//	$Id: constants.h,v 1.32 2002/06/09 13:04:03 cs Exp $	
//	$Manager: goetz$
// 
//  
//  constants.h - Definition of constants
//  
//  Author: Christian Bauer <cbauer@iphcip1.physik.uni-mainz.de> (1998)
//          Goetz Schwandtner <goetz@informatik.uni-mainz.de> (1999)
//  
//  RoboCup 98, 99 - "Mainz Rolling Brains"
//  
//
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
//  
// NOTES:
//  
//  - The x axis extends from left to right, 
//    the y axis extends from top to bottom
//  - The kick-off point is at (0,0)
//  
// 
//
//////////////////////////////////////////////////////////////////////

#ifndef _CONSTANTS_H_
#define _CONSTANTS_H_

#include <math.h>

#include "coordinates.h"

/*
 *  Technical constants which are not in server.conf
 */

// default soccerserver host and port
const char DEFAULT_HOST[] = "localhost";

// number of Snapshots kept in Worldmodel
const int DEFAULT_SNAPSHOT_DEPTH = 30;

// margin on top of SIMULATOR_STEP, before any messages will be missed
const int COMM_ALARM_MARGIN_USEC = 5000;     

// minimum time interval between any two UDP-sends
const long COMM_MINIMUM_SEND_PAUSE_USEC = 0;

// Maximum cycles we wait before server is considered dead
const int COMM_MAX_DEAD_SERVER_CYCLES = 50;

// protocol version of the agent's snaplogs
const int SNAPLOG_PROTOCOL = 2;

// size of center circle

const double CENTER_CIRCLE_R = 9.15;

// freekick forbidden area
const double FREE_KICK_FORBIDDEN_AREA = 9.15;

const double EPS = 1e-10;

extern int HALF_TIME_IN_CYCLES;

/*
 *  Simulation/server related definitions
 */


// [C] (server-implementation)
// max length of server messages. Depending on implementation of sserver !
const int MAX_SERVER_MSG_SIZE = 8192;

// [C]
// max length of say messages. Change this if value in server.conf has changed!
const int MAX_SAY_MSG_SIZE = 512;

// this struct stores the information sent in a player_param message
// which is parsed during checking into the server
extern struct PlayerParams {
  int player_types;
  int subs_max;
  double pt_max;
  double player_speed_max_delta_min;
  double player_speed_max_delta_max;
  double stamina_inc_max_delta_factor;
  double new_stamina_inc_max_delta_factor;
  double player_decay_delta_min;
  double player_decay_delta_max;
  double inertia_moment_delta_factor;
  double dash_power_rate_delta_min;
  double dash_power_rate_delta_max;
  double new_dash_power_rate_delta_min;
  double new_dash_power_rate_delta_max;
  double player_size_delta_factor;
  double kickable_margin_delta_min;
  double kickable_margin_delta_max;
  double kick_rand_delta_factor;
  double extra_stamina_delta_min;
  double extra_stamina_delta_max;
  double effort_max_delta_factor;
  double effort_min_delta_factor;
} playerParams;

extern double GOAL_WIDTH;
extern double PLAYER_SIZE;
extern double PLAYER_DECAY;
extern double PLAYER_RAND;
extern double PLAYER_WEIGHT;
extern double PLAYER_SPEED_MAX;
extern double PLAYER_ACCEL_MAX;
extern double STAMINA_MAX;
extern double STAMINA_INC_MAX;
extern double RECOVER_DEC_THR;
extern double RECOVER_DEC;
extern double RECOVER_MIN;
extern double EFFORT_DEC_THR;
extern double EFFORT_DEC;
extern double EFFORT_INC_THR;
extern double EFFORT_INC;
extern double EFFORT_MIN;
extern double EFFORT_MAX;
extern double HEAR_MAX;
extern double HEAR_INC;
extern double HEAR_DECAY;
extern double INERTIA_MOMENT;
extern int    SENSE_BODY_STEP;
extern double CATCHABLE_AREA_L;
extern double CATCHABLE_AREA_W;
extern double CATCH_PROBABILITY;
extern int    CATCH_BAN_CYCLES;
extern int    GOALIE_MAX_MOVES;
extern double BALL_SIZE;
extern double BALL_DECAY;
extern double BALL_RAND;
extern double BALL_WEIGHT;
extern double BALL_SPEED_MAX;
extern double BACCEL_MAX;
extern double WIND_FORCE;
extern double WIND_DIR;
extern double WIND_RAND;
extern double WIND_ANGLE;
extern double KICKABLE_AREA;
extern double KICKABLE_MARGIN;
extern double CKICK_MARGIN;
extern double DASH_POWER_RATE;
extern double KICK_POWER_RATE;
extern double KICK_RAND;
extern double VISIBLE_ANGLE;
extern double AUDIO_CUT_DIST;
extern double QUANTIZE_STEP;
extern double QUANTIZE_STEP_L;
extern double MAXPOWER;
extern double MINPOWER;
extern double MAXMOMENT;
extern double MINMOMENT;
extern double MAXNECKMOMENT;
extern double MINNECKMOMENT;
extern double MAXNECKANG;
extern double MINNECKANG;
extern int    PORT;
extern int    COACH_PORT;
extern int    OLCOACH_PORT;
extern int    SAY_COACH_CNT_MAX;
extern int    SAY_COACH_MSG_SIZE;
extern int    CLANG_WIN_SIZE;
extern int    CLANG_DEFINE_WIN;
extern int    CLANG_DEL_WIN;
extern int    CLANG_RULE_WIN;
extern int    CLANG_META_WIN;
extern int    CLANG_ADVICE_WIN;
extern int    CLANG_INFO_WIN;
extern int    CLANG_MESS_DELAY;
extern int    CLANG_MESS_PER_CYCLE;
extern int    SEND_VI_STEP;
extern int    SIMULATOR_STEP;
extern int    SEND_STEP;
extern int    RECV_STEP;
extern int    HALF_TIME;
extern int    SAY_MSG_SIZE;
extern bool   USE_OFFSIDE;
extern double OFFSIDE_ACTIVE_AREA;
extern bool   FORBID_KICK_OFF_OFFSIDE;
extern double OFFSIDE_KICK_MARGIN;
extern int    SLOW_DOWN_FACTOR;
extern double VISIBLE_DISTANCE;
extern int    DROP_BALL_TIME;
extern bool   SYNCH_MODE;
extern int    POINTTO_BAN_CYCLES;
extern int    POINTTO_DURATION;
extern double TACKLE_DIST;
extern double TACKLE_BACK_DIST;
extern double TACKLE_WIDTH;
extern int    TACKLE_EXP;
extern int    TACKLE_CYCLES;
extern double TACKLE_POWER_RATE;
/*
 *  Server constants that are not in the server.conf
 */

const double UNUM_FAR_LENGTH= 20.0;
const double UNUM_TOOFAR_LENGTH= 40.0;
const double TEAM_FAR_LENGTH= 40.0;
const double TEAM_TOOFAR_LENGTH= 60.0;
extern double VISIBLE_DISTANCE;  // sent in server_param msg from sserver

/*
 *  Definitions for the field
 */

const double FIELD_LENGTH = 105.0;
const double FIELD_WIDTH = 68.0;

const double PENALTY_AREA_LENGTH = 16.5;
const double PENALTY_AREA_WIDTH = 40.32;

const double GOAL_AREA_LENGTH = 5.5;
const double GOAL_AREA_WIDTH = 18.32;

extern double GOAL_WIDTH;
const double GOAL_DEPTH = 2.44;


/*
 *  Definitions for flags on the field
 */

// Enumeration of flag types
enum flag_type {
	FLAG_UNKNOWN = -1,	// "anonymous" flag, used in sensor data
				// i.e. we see a flag but don't know which one it is
	FLAG_TOP_LEFT = 0,
	FLAG_TOP_CENTER,
	FLAG_TOP_RIGHT,
	FLAG_BOTTOM_LEFT,
	FLAG_BOTTOM_CENTER,
	FLAG_BOTTOM_RIGHT,

	FLAG_P_LEFT_TOP,	// flags in penalty area
	FLAG_P_LEFT_CENTER,
	FLAG_P_LEFT_BOTTOM,
	FLAG_P_RIGHT_TOP,
	FLAG_P_RIGHT_CENTER,
	FLAG_P_RIGHT_BOTTOM,

	FLAG_CENTER,

	FLAG_GOAL_LEFT_TOP,	// goal posts
	FLAG_GOAL_LEFT_BOTTOM,
	FLAG_GOAL_RIGHT_TOP,
	FLAG_GOAL_RIGHT_BOTTOM,

	FLAG_TOP_LEFT_50,	// flags outside of field
	FLAG_TOP_LEFT_40,
	FLAG_TOP_LEFT_30,
	FLAG_TOP_LEFT_20,
	FLAG_TOP_LEFT_10,
	FLAG_TOP_0,
	FLAG_TOP_RIGHT_10,
	FLAG_TOP_RIGHT_20,
	FLAG_TOP_RIGHT_30,
	FLAG_TOP_RIGHT_40,
	FLAG_TOP_RIGHT_50,

	FLAG_BOTTOM_LEFT_50,
	FLAG_BOTTOM_LEFT_40,
	FLAG_BOTTOM_LEFT_30,
	FLAG_BOTTOM_LEFT_20,
	FLAG_BOTTOM_LEFT_10,
	FLAG_BOTTOM_0,
	FLAG_BOTTOM_RIGHT_10,
	FLAG_BOTTOM_RIGHT_20,
	FLAG_BOTTOM_RIGHT_30,
	FLAG_BOTTOM_RIGHT_40,
	FLAG_BOTTOM_RIGHT_50,

	FLAG_LEFT_TOP_30,
	FLAG_LEFT_TOP_20,
	FLAG_LEFT_TOP_10,
	FLAG_LEFT_0,
	FLAG_LEFT_BOTTOM_10,
	FLAG_LEFT_BOTTOM_20,
	FLAG_LEFT_BOTTOM_30,

	FLAG_RIGHT_TOP_30,
	FLAG_RIGHT_TOP_20,
	FLAG_RIGHT_TOP_10,
	FLAG_RIGHT_0,
	FLAG_RIGHT_BOTTOM_10,
	FLAG_RIGHT_BOTTOM_20,
	FLAG_RIGHT_BOTTOM_30,

	NUM_FLAGS		// Number of flags (excluding the anonymous flag)
};

// Get the absolute position of a flag of given type
inline const Vector2d FLAGPOS(flag_type which)
{
  extern double flag_x[], flag_y[];

  if (which < 0 || which >= NUM_FLAGS)
    return Vector2d();
  else 
    return Vector2d(flag_x[which], flag_y[which]);
}


/*
 *  Definitions for lines on the field
 */

// Enumeration of line types
enum line_type {
	LINE_UNKNOWN = -1,	// "anonymous" line, used in sensor data
				// i.e. we see a line but don't know which one it is
	LINE_TOP = 0,
	LINE_LEFT,
	LINE_RIGHT,
	LINE_BOTTOM,
	NUM_LINES		// Number of lines (excluding the anonymous line)
};

// Is a given line horizontal?
inline bool IsLineHorizontal(line_type which) {return which == LINE_TOP || which == LINE_BOTTOM;}

// Is a given line vertical?
inline bool IsLineVertical(line_type which) {return which == LINE_LEFT || which == LINE_RIGHT;}


/*
 *  Definitions for goals
 */

// Enumeration of goal types ("left" and "right" refers to the
// absolute position on the field)
enum goal_type {
	GOAL_UNKNOWN = -1,	// "anonymous" goal
				// i.e. we see a goal but don't know which one it is
	GOAL_LEFT = 0,
	GOAL_RIGHT,
	NUM_GOALS		// Number of goals (excluding the anonymous goal)
};

// Get the absolute position of the center point of a goal of given type
inline const Vector2d GOALPOS(goal_type which)
{
  extern double goal_x[], goal_y[];
  if (which < 0 || which >= NUM_GOALS)
    return Vector2d();
  else 
    return Vector2d(goal_x[which], goal_y[which]);
}

#endif




