//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// drawMacros.h - Macros to be used in Drawtraces
//
// $Id: drawMacros.h,v 1.6 2001/09/11 10:01:38 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#ifndef _DRAWTRACES_H_
#define _DRAWTRACES_H_

#if !defined(FUNSSELMACROS) && !defined(REPLAYERMACROS)
#define REPLAYERMACROS
#endif

#ifdef REPLAYERMACROS
// draw circle and light circle resp.
#define CIRCLE(pos, rad) "(c " << pos << " " << rad << ")"
#define LCIRCLE(pos, rad) "(c " << pos << " " << rad << "l)"

// draw circle with position and light circle with position resp.
// for pos use a Vector2d !!
#define GCIRCLE(pos, rad) "(c " << pos << " " << rad << "g)"
#define LGCIRCLE(pos, rad) "(c " << pos << " " << rad << "gl)"

// draw position and light position mark
// for pos use a Vector2d !!
#define POSITION(pos) "(p " << pos << ")"
#define LPOSITION(pos) "(p " << pos << " l)"

// draw a line and light line resp
// for start and end use Vector2d !!
#define LINE(start, end) "(l " << start << end << ")"
#define LLINE(start, end) "(l " << start << end << "l)"

// draw a vertical line (eg offside line)
#define VLINE(x) "(l (" << x << ", " << -FIELD_WIDTH/2 << ")(" << x << ", " << FIELD_WIDTH/2 << "))"
#define LVLINE(x) "(l (" << x << ", " << -FIELD_WIDTH/2 << ")(" << x << ", " << FIELD_WIDTH/2 << ") l)"

// draw a corridor and light corridor resp. (without the circle around target)
#define CORRIDOR(source, target, d, alpha) \
LINE( source, \
      source + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) + \
      radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) ) \
LINE( source + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) + \
      radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ), \
      target + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) ) \
LINE( target + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ), \
      target - radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) ) \
LINE( target - radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ), \
      source + radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) + \
      radAng2Vector2d( 0.5 * d, ( target - source ).perp().perp().perp().getAngle() ) ) \
LINE( source + radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) + \
      radAng2Vector2d( 0.5 * d, ( target - source ).perp().perp().perp().getAngle() ), \
      source )
#define LCORRIDOR(source, target, d, alpha) \
LLINE( source, \
       source + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) + \
       radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) ) \
LLINE( source + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) + \
       radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ), \
       target + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) ) \
LLINE( target + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ), \
       target - radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) ) \
LLINE( target - radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ), \
       source + radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) + \
       radAng2Vector2d( 0.5 * d, ( target - source ).perp().perp().perp().getAngle() ) ) \
LLINE( source + radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) + \
       radAng2Vector2d( 0.5 * d, ( target - source ).perp().perp().perp().getAngle() ), \
       source )
 
// draw a text string to a position pos (Vector2d)
// or draw a text into the corner
#define TEXT(text, pos) "(t " << pos << text << ")"
#define CTEXT(text) "(t (" << (-FIELD_LENGTH/2) << ", " << (-FIELD_WIDTH/2+3) << ") " << text << ")"

#endif

#ifdef FUNSSELMACROS

// stop command for server
#define STOP MStop

// marker command
#define MARKER(pos, text, duration) MMarker(pos, text, duration)

// NOTE FOR MACROS WITH STYLE-OPTION:
// there is an enumerate in mcommand.h, which gives the line-style:
// enum MStyle { MSolid=0, MDash=1, MDot=2, MDashDot=3, MDashDotDot=4};

// some of the drawmacros have a version (beginning with S) which 
// accepts an extra line style parameter of type MStyle as last one

// draw circle and light circle resp.
#define CIRCLE(pos, rad) MDCircle(pos, rad)
#define SCIRCLE(pos, rad, style) MDCircle(pos, rad, style)

// draw circle with position and light circle with position resp.
// for pos use a Vector2d !!
#define GCIRCLE(pos, rad) MDCircle(pos, rad, true)
#define SGCIRCLE(pos, rad, style) MDCircle(pos, rad, true, style)

// draw position and light position mark
// for pos use a Vector2d !!
#define POSITION(pos) MDPosition(pos)

// draw a line and light line resp
// for start and end use Vector2d !!
#define LINE(start, end) MDLine(start, end)
#define SLINE(start, end, style) MDLine(start, end, style)

// draw a vertical line (eg offside line)
#define VLINE(x) MDVLine(x)
#define SVLINE(x, style) MDVLine(x, style)

// draw a box with two opposite corners pos1 and pos2
#define BOX(pos1, pos2) MDBox(pos1, pos2)
#define SBOX(pos1, pos2, style) MDBox(pos1, pos2, style)

// draw a corridor and light corridor resp. (without the circle around target)
#define CORRIDOR(source, target, d, alpha) MDCorridor(source, target, d, alpha)
#define SCORRIDOR(source, target, d, alpha, style) MDCorridor(source, target, d, alpha, style)

// draw a text string to a position pos (Vector2d)
// or draw a text into the corner
#define TEXT(text, pos) MDLabel(pos, text)

#endif

#endif
