//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*- ///////////////////////////////////////////////////////
//
// mcommand.C - some classes to send commands to special monitor
//              controlling the game
//
// $Id: mcommand.C,v 1.15 2002/01/31 17:20:38 goetz Exp $
// Time-stamp: <2001-05-07 15:30:28 goetz>
//////////////////////////////////////////////////////////////////////

#include "coordinates.h"
#include "mcommand.h"
#include "snapshot.h"

#include <iostream.h>
#include <string>
#include <strstream.h>
#include "stdio.h"

#include "FunsselStrings.h"

// class MCommand

class ostream& operator<< (class ostream& os, const MCommand &mc)
{
  return os << string(mc);
}

MCommand::MCommand( int duration, const string &encoded)
  : duration(duration) 
{
  ostrstream ost;
  ost << encoded << (char)0;
  char *str= ost.str();
  this->encoded= str;
  delete str;
}

// class MMarker
MMarker::MMarker( const Vector2d &pos, const char *label, int duration)
  : MCommand( duration )
{
  char buf[201];
  const char *tmp= label;
  if (label == 0)
    tmp= "";

  sscanf(tmp, "%200[^\"]", buf);

  ostrstream ost;
  ost << FMC_MARK " " << duration << " " 
      << pos << '"' << buf << '"' << " )" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MLog
MLog::MLog( const char *text, int duration)
  : MCommand(duration, string(FMC_LOG " "+string(text)+string(")") ))
{}

// class MSnapshot
MSnapshot::MSnapshot( const Snapshot & sn)
{
  ostrstream ost;
  ost << FMC_SNAP " ";
  sn.binaryWriteToStreamOSI( ost );
  ost << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}  

// class MDCircle
MDCircle::MDCircle(const Vector2d &pos, double rad, bool hasGrid, 
		   MStyle style,  int duration)
  : MCommand(duration)
{
  ostrstream ost;
  ost << FMC_CIRCLE " " << duration << " " << pos << " " 
      << rad << " " << hasGrid << " " << style << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// MDPosition
MDPosition::MDPosition(const Vector2d &pos, int duration)
  : MCommand( duration )
{
  ostrstream ost;
  ost << FMC_POS " " << duration << " " << pos << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MDLine
MDLine::MDLine(const Vector2d &start, const Vector2d &end, MStyle style, int duration)
  : MCommand( duration )
{
  ostrstream ost;
  ost << FMC_LINE " " << duration << " " << start 
      << " " << end << " " << style << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MDVLine
MDVLine::MDVLine(double x, MStyle style, int duration)
  : MCommand( duration )
{
  ostrstream ost;
  ost << FMC_VLINE " " << duration << " " << x << " " << style  << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MDBox
MDBox::MDBox(const Vector2d &start, const Vector2d &end, MStyle style, int duration)
  : MCommand( duration )
{
  ostrstream ost;
  ost << FMC_BOX " " << duration << " " << start 
      << " " << end << " " << style << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MDCorridor 
MDCorridor::MDCorridor( const Vector2d &source, const Vector2d &target,
			double diam, const Angle &alpha, MStyle style,
			int duration)
  : MCommand( duration )
{
  ostrstream ost;
  ost << FMC_CORR " " << duration << " " << source << " " << target 
      << " " << diam << " " << alpha << " " << style  << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MDLabel
MDLabel::MDLabel( const Vector2d &pos, const string &text, int duration )
  : MCommand( duration )
{
  ostrstream ost;
  ost << FMC_LABEL " " << duration << " " << pos << " " << text << ")"
      << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MStop

MStop::MStop() 
  : MCommand( 1, string(FMC_STOP ")") )
{}

// class MDropBall
MDropBall::MDropBall( const Vector2d &pos ) 
{
  ostrstream ost;
  ost << FMC_DROPBALL " " << duration << " " << pos << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MFreekickUs
MFreekickUs::MFreekickUs( const Vector2d &pos )
{
  ostrstream ost;
  ost << FMC_FREEKUS " " << duration << " " << pos << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MFreekickThem
MFreekickThem::MFreekickThem( const Vector2d &pos )
{
  ostrstream ost;
  ost << FMC_FREEKTHEM " " << duration << " " << pos << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}

// class MSetPlayer
MSetPlayer::MSetPlayer( bool ourTeam, int playerNo, const Vector2d &pos, 
			const Angle &ang)
{
  ostrstream ost;
  ost << FMC_SETPLYR " " << duration << " " << playerNo << " ";
  if (ourTeam)
    ost << 1;
  else
    ost << 0;
  ost << " " << pos << " " << ang << ")" << (char)0;
  char *str= ost.str();
  encoded = str;
  delete str;
}
