//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: mrbtypes.h,v 1.12 2002/06/08 19:33:05 gauer Exp $
// 
//  mrbtypes.h - declaration of enum types for sserver information
// 
//////////////////////////////////////////////////////////////////////


#ifndef _MRBTYPES_H_
#define _MRBTYPES_H_
#include <iostream.h>

enum view_width {
  VIEW_WIDTH_NARROW,
  VIEW_WIDTH_NORMAL,
  VIEW_WIDTH_WIDE,
  VIEW_WIDTH_NONE
};

enum view_quality {
  VIEW_QUALITY_LOW,
  VIEW_QUALITY_HIGH
};

enum playing_side {
  SIDE_UNKNOWN = -1,
  SIDE_LEFT,
  SIDE_RIGHT
};

enum play_mode {
  PLAY_MODE_UNKNOWN = -1,
  PLAY_MODE_BEFORE_KICK_OFF = 0,
  PLAY_MODE_TIME_OVER,
  PLAY_MODE_HALF_TIME,
  PLAY_MODE_TIME_EXTENDED,
  PLAY_MODE_PAUSE,
  PLAY_MODE_PLAY_ON,

  PLAY_MODE_DROP_BALL,
  PLAY_MODE_HUMAN_JUDGE,


  PLAY_MODE_KICK_OFF_US,
  PLAY_MODE_KICK_OFF_THEM,

  PLAY_MODE_KICK_IN_US,
  PLAY_MODE_KICK_IN_THEM,

  PLAY_MODE_FREE_KICK_US,
  PLAY_MODE_FREE_KICK_THEM,

  PLAY_MODE_CORNER_KICK_US,
  PLAY_MODE_CORNER_KICK_THEM,

  PLAY_MODE_GOAL_KICK_US,
  PLAY_MODE_GOAL_KICK_THEM,

  PLAY_MODE_PENALTY_KICK_US,
  PLAY_MODE_PENALTY_KICK_THEM,


  PLAY_MODE_GOAL_US,
  PLAY_MODE_GOAL_THEM,

  PLAY_MODE_GOALIE_CATCH_US,
  PLAY_MODE_GOALIE_CATCH_THEM,

  PLAY_MODE_OFFSIDE_US,
  PLAY_MODE_OFFSIDE_THEM,

  PLAY_MODE_BACKPASS_US,
  PLAY_MODE_BACKPASS_THEM,

  PLAY_MODE_FREEKICK_FAULT_US,
  PLAY_MODE_FREEKICK_FAULT_THEM,

 
  PLAY_MODE_FOUL_CHARGE_US,
  PLAY_MODE_FOUL_CHARGE_THEM,

  PLAY_MODE_FOUL_PUSH_US,
  PLAY_MODE_FOUL_PUSH_THEM,

  PLAY_MODE_FOUL_MULTIPLE_ATTACK_US,
  PLAY_MODE_FOUL_MULTIPLE_ATTACK_THEM,

  PLAY_MODE_FOUL_BALLOUT_US,
  PLAY_MODE_FOUL_BALLOUT_THEM
};

enum CoachMsgType {
  CMT_FREEFORM,
  CMT_META,
  CMT_INFO,
  CMT_ADVICE,
  CMT_DEFINE,
  CMT_NONE
};

inline
class ostream & operator << (class ostream &os, const play_mode pm)
{
 switch (pm)
   {
   case PLAY_MODE_UNKNOWN:
     return os << "PLAY_MODE_UNKNOWN";
   case PLAY_MODE_BEFORE_KICK_OFF:
     return os << "PLAY_MODE_BEFORE_KICK_OFF";
   case PLAY_MODE_TIME_OVER:
     return os << "PLAY_MODE_TIME_OVER";
   case PLAY_MODE_HALF_TIME:
     return os << "PLAY_MODE_HALF_TIME";
   case PLAY_MODE_TIME_EXTENDED:
     return os << "PLAY_MODE_TIME_EXTENDED";
   case PLAY_MODE_PAUSE:
     return os << "PLAY_PAUSE";
   case PLAY_MODE_PLAY_ON:
     return os << "PLAY_MODE_PLAY_ON";

   case PLAY_MODE_DROP_BALL:
     return os << "PLAY_MODE_DROP_BALL";
   case PLAY_MODE_HUMAN_JUDGE:
     return os << "PLAY_MODE_HUMAN_JUDGE";

   case PLAY_MODE_KICK_OFF_US:
     return os << "PLAY_MODE_KICK_OFF_US";
   case PLAY_MODE_KICK_OFF_THEM:
     return os << "PLAY_MODE_KICK_OFF_THEM";

   case PLAY_MODE_KICK_IN_US:
     return os << "PLAY_MODE_KICK_IN_US";
   case PLAY_MODE_KICK_IN_THEM:
     return os << "PLAY_MODE_KICK_IN_THEM";

   case PLAY_MODE_FREE_KICK_US:
     return os << "PLAY_MODE_FREE_KICK_US";
   case PLAY_MODE_FREE_KICK_THEM:
     return os << "PLAY_MODE_FREE_KICK_THEM";

   case PLAY_MODE_CORNER_KICK_US:
     return os << "PLAY_MODE_CORNER_KICK_US";
   case PLAY_MODE_CORNER_KICK_THEM:
     return os << "PLAY_MODE_CORNER_KICK_THEM";

   case PLAY_MODE_GOAL_KICK_US:
     return os << "PLAY_MODE_GOAL_KICK_US";
   case PLAY_MODE_GOAL_KICK_THEM:
     return os << "PLAY_MODE_GOAL_KICK_THEM";

   case PLAY_MODE_PENALTY_KICK_US:
     return os << "PLAY_MODE_PENALTY_KICK_US";
   case PLAY_MODE_PENALTY_KICK_THEM:
     return os << "PLAY_MODE_PENALTY_KICK_THEM";

   case PLAY_MODE_GOAL_US:
     return os << "PLAY_MODE_GOAL_US";
   case PLAY_MODE_GOAL_THEM:
     return os << "PLAY_MODE_GOAL_THEM";

   case PLAY_MODE_GOALIE_CATCH_US:
     return os << "PLAY_MODE_GOALIE_CATCH_US";
   case PLAY_MODE_GOALIE_CATCH_THEM:
     return os << "PLAY_MODE_GOALIE_CATCH_THEM";

   case PLAY_MODE_OFFSIDE_US:
     return os << "PLAY_MODE_OFFSIDE_US";
   case PLAY_MODE_OFFSIDE_THEM:
     return os << "PLAY_MODE_OFFSIDE_THEM";

   case PLAY_MODE_BACKPASS_US: 
     return os << "PLAY_MODE_BACKPASS_US";
   case PLAY_MODE_BACKPASS_THEM:
     return os << "PLAY_MODE_BACKPASS_THEM";

   case PLAY_MODE_FREEKICK_FAULT_US:
     return os << "PLAY_MODE_FREEKICK_FAULT_US";
   case PLAY_MODE_FREEKICK_FAULT_THEM:
     return os << "PLAY_MODE_FREEKICK_FAULT_THEM";

   case PLAY_MODE_FOUL_CHARGE_US:
     return os << "PLAY_MODE_FOUL_CHARGE_US";
   case PLAY_MODE_FOUL_CHARGE_THEM:
     return os << "PLAY_MODE_FOUL_CHARGE_THEM";

   case PLAY_MODE_FOUL_PUSH_US:
     return os << "PLAY_MODE_FOUL_PUSH_US";
   case PLAY_MODE_FOUL_PUSH_THEM:
     return os << "PLAY_MODE_FOUL_PUSH_THEM";

   case PLAY_MODE_FOUL_MULTIPLE_ATTACK_US:
     return os << "PLAY_MODE_FOUL_MULTIPLE_ATTACK_US";
   case PLAY_MODE_FOUL_MULTIPLE_ATTACK_THEM:
     return os << "PLAY_MODE_FOUL_MULTIPLE_ATTACK_THEM";

   case PLAY_MODE_FOUL_BALLOUT_US:
     return os << "PLAY_MODE_FOUL_BALLOUT_US";
   case PLAY_MODE_FOUL_BALLOUT_THEM:
     return os << "PLAY_MODE_FOUL_BALLOUT_THEM";

   default:
     return os << "PLAY_MODE: UNRECOGNIZED !?!";
   }
}


#endif
