//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*-////////////////////////////////////////////////////////
//
// $Id: player.C,v 1.25 2002/06/03 12:33:18 cm Exp $
//
// player.C - interface for player client
//
///////////////////////////////////////////////////////////////////////


#include "player.h"
#include "Config.h"

#undef TRACE
#define TRACE(cmd)

Player::Player (const char* team_name,             
		bool        isGoalie, 
		const char* host,
		int         port,
		const char* configfile,
		bool        logging_snap,
		int         reconnectNumber,
                bool        FUNSSELmode,
		int         clangMinVer,
		int         clangMaxVer,
		int         numSnapshots) :

  skills(NULL), 
  comm(NULL),
  world(NULL)
{
  TRACE (cout << "Player: initializing config reader" << endl);

  // try to initialize config file
  if (!Config::readFile(configfile)) {
    cerr << "PANIC: Could not read config file: " << configfile << endl;
    exit(-1);
  }

  TRACE (cout << "Player: generating Comm" << endl);

  TRACE (cout << "Player: initializing Comm" << endl);
  
  if (!isGoalie)
    comm = new Comm(team_name, 
		    CLIENT_FIELDPLAYER,
		    host, port,
		    reconnectNumber,
		    FUNSSELmode,
		    clangMinVer,
		    clangMaxVer);
  else
    comm = new Comm(team_name, 
		    CLIENT_GOALIE,
		    host, port,
		    reconnectNumber,
		    FUNSSELmode,
		    clangMinVer,
		    clangMaxVer);
  
  comm_error ce;

  if (comm->isServerAlive())
    ce = comm->connect();
  else
    ce = COMM_ERROR;

  switch (ce) {
  case (COMM_OK): 
    
    TRACE (cout << "Player: generating Worldmodel and AdvSkills" << endl);
    
    world  = new Worldmodel (numSnapshots, comm);
    skills = new AdvSkills (comm);
    
    writeSnaps = false;
    if (logging_snap) {
      writeSnaps = true;
      // open snap_log file
      ostrstream tmp;
      tmp << SNAP_LOG_FILENAME << getMyTeam() << getMyNumber() 
	  << SNAP_LOG_EXTENSION;
      char* filename = tmp.str();
      out.open(filename, ios::out);
      // Version control
      out << SNAPLOG_PROTOCOL << endl;

      delete filename;
    }
    
    break;
  case (COMM_ERROR): 
    cerr << "Player: unspecified error when initializing !" << endl; break;
  case (COMM_ERROR_SERVER_NOT_FOUND):
    cerr << "Player: could not find server's host !"<< endl; break;
  case (COMM_ERROR_SOCKET_INIT):
    cerr << "Player: could not create and bind socket !" << endl; break;
  case (COMM_ERROR_SERVER_INIT):
    cerr << "Player: could not deliver checkin message to server !" 
	 << endl; break;
  case (COMM_ERROR_PARSE):
    cerr << "Player: could not parse init response !" << endl; break;
  case (COMM_ERROR_NO_DATA):
    cerr << "Player: did not receive response from server !" << endl; break;
  }
} 


Player::~Player () {

  TRACE (cout << "Player: entering destructor, closing down" << endl);
  delete (skills);
  delete (world);
  delete (comm);
  
  if (writeSnaps) out.close();

  TRACE (cout << "Player: leaving destructor" << endl);

}


void 
Player::play() {

  if (comm->isServerAlive()) {
    TRACE (cout << "Player: started" << endl);
    
    while ( comm->isServerAlive() && !comm->isGameOver() ) {
      
      TRACE (cout << "Player: waiting msgs" << endl);
      comm->drainSocket();

      if (comm->isActionNeeded()) {
	
	TRACE (cout << "Player: need action for time " << getTime() << endl); 
	
	TRACE (cout << "Player: adjusting worldmodel" << endl);
	// world->adjust();   // to be done in mastercontrol
	
	TRACE (cout << "Player: getting snapshot of time " 
	       << getTime() << endl);

	TRACE (cout << "Player: acting" << endl);
	
	act();

	Snapshot* snap = world->getSnapshot(getTime());
	
	if (snap == NULL) {
	  cerr << "Player (" << getMyTeam() << " No." << getMyNumber()
	       <<"): cannot generate Snapshot of time " << getTime() << endl;
	} else {
	  if ( (writeSnaps) && comm->isTimeRunning() ) {
	    if (comm->inFUNSSELmode())
	      Comm::sendMCmd(MSnapshot(*snap));
	    else {
	      snap->binaryWriteToStreamOSI(out);
	      out << endl;
	    }
	  }
	  
	} // got a snapshot
	
      } else {
	TRACE(cout << "Player: do not need new action yet" << endl);
      }
      
    } // while server alive
    
    TRACE (cout << "Player: server is gone, leaving" << endl);
    
  } else { 
    // server never was alive
    cerr << "Player: could not connect to sserver, closing down" << endl;
  }
  
} //play
