//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player PassEval
//
// \$Id: passeval.h,v 1.33 2002/06/03 12:30:19 cm Exp $
// \$manager: schappel$
//
//////////////////////////////////////////////////////////////////////

#ifndef __PASSEVAL_H__   
#define __PASSEVAL_H__

#include "grade.h"
#include "worldmodel.h"
#include "goalshoteval.h"
#include "AdvSkills.h"


// the class PropOf
//////////////////////////

class PropOf {
 public:
  int      number;    // number of partner
  Vector2d pos;       // pos to pass to; not necessary pos of partner! 
  int      xIndex;    // indicates grid position of pass
  int      yIndex;
  double   dist;      // distance between player and target pos
  double   vel;       // velocity that ball should reach at position
  int      free;      // radius of free circle around pos
  int      corridor;  // free corridor width
  bool     offside;   // partner is not in an offside-positon
  bool     perhaps;   // true, if pass makes sense
  Grade    grade;     // grade of the pass from player to pos

  // shout the PropOf
  friend ostream& operator << ( class ostream &os,  const PropOf prop)
  {
    return os << " pass to  " << prop.number << " at pos " 
	      << prop.pos << " with dist " << prop.dist  
	      << " free: " << prop.free << " corridor: " 
	      << prop.corridor << " offside: " << prop.offside 
	      << " grade: " << prop.grade  
	      << " result is perhaps: " << prop.perhaps;
  }
};

// the class EvalPoint
//////////////////////////
class EvalPoint
{
  friend class GoalEvalPoints;

 private:
  EvalPoint(Vector2d _pos, Grade _grade, EvalPoint *evalpoint = NULL)
    {
      pos = _pos;
      grade = _grade;
      nextpoint = evalpoint;
    }
  Vector2d pos;
  Grade grade;
  EvalPoint *nextpoint;
};

// the class GoalEvalPoints
//////////////////////////

class GoalEvalPoints
{
 public:
  // constructor
  GoalEvalPoints()
    {
      pointlist = NULL;
    }

  // destructor
  ~GoalEvalPoints()
    { 
      deleteAll();
    }
  
  // are there any points for which the goalshoteval-routine
  // was called
  bool isEmpty()
    {
      return pointlist == NULL ? true : false;
    }

  // which timestep do the elements in the list belong to
  int getTime()
    {
      return time;
    }

  // set timestep
  void setTime(int _time)
    {
      time = _time;
    }
  
  // insert a new point
  void insert(Vector2d pos, Grade grade)
    {
      EvalPoint *newpoint = new EvalPoint(pos, grade, pointlist);
      pointlist = newpoint;
    }

  // delete all points
  void deleteAll()
    {
      EvalPoint *help = pointlist;
      while (help)
	{
	  EvalPoint *tmp = help;
	  help = help->nextpoint;
	  delete tmp;
	}
      pointlist = NULL;
    }

  // look for a certain point
  bool allreadyComputed(Vector2d pos, Grade &grade)
    {
      EvalPoint *help = pointlist;
      bool found = false;
      while (help && !found)
	{
	  if (help->pos == pos)
	    {
	      found=true;
	      grade = help->grade;
	    }
	  else
	    help = help->nextpoint;
	}
      return found;
    }
		   

 private:
  
  EvalPoint *pointlist;
  int time;
};
  


// the class PassEval
//////////////////////////

class PassEval
{
public:
  // constructor
  PassEval(const GoalShotEval* _goalshoteval);
  // destructor
  ~PassEval();

  // this methods needs a worldSnap, a playernumber and a partnernumber;
  // it grades the passes from player to partner on the basis of the worldSnap,
  // considering also rangepasses;
  // the position where the pass would be played to is reported in posOfPass;
  Grade ourEval(const Snapshot* worldSnap, int plnumber, int panumber, 
		Vector2d &posOfPass) const;
  
  // just for calling the above method without posOfPass
  Grade ourEval(const Snapshot* worldSnap, int plnumber, int panumber) const
    { Vector2d v; return ourEval(worldSnap, plnumber, panumber, v); };

  // does the same as above but reports PropOf of best pass
  PropOf ourEvalProp(const Snapshot* worldSnap, int plnumber, 
		     int panumber) const;

  // this methods does the same, but for players of the other team;
  // it takes no features of the other team into account;
  // does not consider rangepasses
  Grade otherEval(const Snapshot* worldSnap, int plnumber, int panumber) const;

  // get the "better" pass out of two;
  // true, if good is better than bad, else false;
  bool betterPass(const Snapshot *wS, PropOf good, PropOf bad) const;

private:
  // this module is used by the passmodule of a player, too!
  friend class PassModule;

  // some things this module needs:
  /////////////////////////////////////////

  const GoalShotEval* goalshoteval;
 
  GoalEvalPoints *goalevalpoints;
  
  // the dangerzone for freestanding-check, probably adaptive?
  double DANGERZONE;

  // to decide, when an player is on the same level with another player
  // I don't want to do it with the exact x-pos
  double SAMELEVEL_LENGTH;
  
  int MAXKICK_DURATION;      // maximal time for preparing a kick
  
  // in some standard situations like corner kicks and kick ins
  // the ball must be kicked immediately and may not be placed
  // around the player
  int STDSITKICK_DURATION;   

  double MAXPASS_LENGTH; // maximal length of a pass

  double RPASS_LATTICELENGTH;
  
  // length of a pass that is too short to play it if the sit. is not dangerous
  double SHORTPASS_LENGTH;

  // length of a pass that is too short to play it for the goalie
  double SHORTPASS_LENGTH_FOR_GOALIE;

  // distance to the sides of the field
  double SECURITY_DIST;

  // distance to the sides of the field that shell be kept for range passes
  double SIDELINE_DIST;

  // some private methods
  /////////////////////////////////////////

  // answer the question, if the player can do a safe kick/fixed kick/superkick
  // if there is enough time for a superkick,  return SUPERKICK_DURATION
  //                               fixed kick, return FIXEDKICK_DURATION
  //                               kick away,  return KICKAWAY_DURATION
  // player is the number of the player that shall pass;
  int check_for_safekick(const Snapshot *wS, int player) const;
 
  // get the PropOf of a pass from player no. "plnumber" to "target"
  // the receiver of the pass is its teammate with no. "panumber"
  // if we == true, the method works for our team, else for the opponent
  bool get_pPass(bool we, const Snapshot *wS, int plnumber,
		 const Vector2d target, int panumber, PropOf &prop,
		 bool isGoalie) const;

  // get the PropOf of "best" r-pass from player no. "plnumber" to player no. 
  // "panumber"
  bool get_rPass(bool we, const Snapshot *wS, int plnumber, int panumber, 
		 PropOf &prop, bool isGoalie) const;

  // calculate the Grade for a passpartner regarding his and the player's 
  // position
  // if we == true, look at our team, else at the other team
  Grade calculateGrade(bool we, const Snapshot *wS, const Vector2d &player,
		       const Vector2d &partner, PropOf &prop,
		       bool isGoalie) const;

};


#endif
