//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module PassModule
//
// \$Id: passmodule.h,v 1.50 2002/06/03 12:32:20 cm Exp $
// \$manager: schappel$
//
//////////////////////////////////////////////////////////////////////

#ifndef __PASSMODULE_H__   
#define __PASSMODULE_H__

#include "passeval.h"
#include "goalshoteval.h"

#include "grade.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"
#include "drawMacros.h"
#include "mcommand.h"

#include <strstream.h>

// the class PassModule
//////////////////////////

class PassModule
{
public:
  // constructor of module PassModule
  PassModule(PassEval* _passeval, const GoalShotEval* _goalshoteval);
  ~PassModule();
  
  // feed information about new world and tactics into module
  bool feedbackWorld(const Snapshot* _worldSnap, 
		     TacticalSnapshot* _tacticalSnap);

  // write information into String (for player communication)
  bool getSayMsg( char* msg );

  // Do the evaluation for the next action
  // For definition of Grade: see grade.h
  Grade evaluate(AdvSkills *skills, const Snapshot *_worldSnap,
                 TacticalSnapshot *_tacticalSnap);

  /** Decide whether a turn head should be done.
   * This is always called after act, so the data of act and
   * evaluate should be used to calculate the TurnHeadCommand
   * which is then executed by the Mastercontrol.
   * skills should only be used to calculate the command,
   * NOT TO SEND IT!!
   */
  Grade turnHeadEvaluate(AdvSkills* skills, AdvSkills::TurnHeadCommand &thc);
 
  // Perform the action for this module
  bool act(AdvSkills *skills, const Snapshot *_worldSnap, 
	   TacticalSnapshot *_tacticalSnap);

private:
  // recall of this module needed, if true
  bool callMeAgain;

  // we want to use our own passeval-method
  PassEval* passeval;

  // and perhaps the goalshoteval-method
  const GoalShotEval* goalshoteval;

  // for saving the current Snapshot
  const Snapshot *worldSnap;

  // for saving the current tactical Snap
  TacticalSnapshot *tacticalSnap;

  // time of latest snapshot the module received;
  // to check if feedback, evaluate and act work with the same snapshot
  int lastSnapshotTime;

  // time available for a kick
  int time_for_kick;

  // property-array for playerpasses
  PropOf pprops[12];

  // property array for rangepasses:
  PropOf rprops[12];

  // information on the passtarget
  /////////////////////////////////////////
  bool     target_selected;
  bool     target_breed;     // true: pass to player; false: rangepass
  int      target_number;
  Vector2d target_pos;

  double   target_vel;       // desired velocity when ball leaves player
  int      target_corridor;  // width of free corridor to target

  Vector2d target_source;

  // emergency kick away
  /////////////////////////
  Vector2d em_vec;    // direction of emergency-kick-away
  bool     em_case;   // true if we have an emergency kick situation

  // information concerning the looking after ball
  ///////////////////////////////////////////////////
  bool look_act;              // true, if act shall only look for ball
  bool look_for_ball;         // true, if player shall look for ball
  int  look_for_ball_time;    // time, looking for ball started
  Vector2d look_for_ball_pos; // position, where player is looking for the ball
  
  Vector2d lookAtPos;         // position where the player wants to look at
  Angle    lookAtAngle;       // angle to turn the head

  Grade turnHeadGrade;        // grade for turning the head towards a pass target

  // information about the kick-commands
  //////////////////////////////////////////
  AdvSkills::ExtKickCommand kickparams;


  // some private methods
  /////////////////////////////////////////


  // this method helps to gather some information
  // it fills the pprop- and the rprop-arrays;
  bool getPasses(bool isGoalie);

  // calculate emergency kick away vector (=em_vec)
  Vector2d get_ePass(play_mode playMode);

  // lock the pass target
  Grade lockPass(PropOf prop);

}; // class PassModule

#endif

