//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                              //
//                                                                                              //
//                             POSITIONMODULE by F. Schulz  (2000)                              //
//                                               F. Flentge (2001)                              //
//                                                                                              //
//                                                                                              //
//////////////////////////////////////////////////////////////////////////////////////////////////
 
// [C] constant dependend code
// [C]* multiple constants
#include "PositionModule.h"
#include "constants.h"
#include "minmax.h"
    

#include "PositionFunctions.C"

//#define POSACTIONTRACE

#ifdef POSACTIONTRACE
#warning: POSM debug trace ACTIONTRACE is activated!
#define PACTIONTR(output) Comm::sendMCmd(MLog(output))
#else
#define PACTIONTR(output)
#endif


const double max_useable_Stamina[4]={500,100,70,30};//max Stamina per cycle
const double save_Stamina[4]={0,.2,.3,.5}; 
  //factor how much stamina is saved for higher priorities

const double MaxNeckAngle=M_PI/4;
const double MinNeckAngle=-M_PI/4;
const double NormalViewWidth=M_PI/4;

const  double FREEPOSAREA_LENGHT=FIELD_LENGTH/8.;
const  double FREEPOSAREA_WIDTH=FIELD_WIDTH/3.;
// the distance from the ball, in which the players goto free positions during
// standartsit. 

PositionModule::PositionModule(bool _isGoalie, 
			       const PassEval *_passEval,  
			       const GoalShotEval *_goalShotEval,
			       PositionEval *_positionEval) : 
  posGoal(0,0),
  priority(3), 
  oldPosGoal(0,0),
  middleOfGoal(-FIELD_LENGTH/2, 0), 
  nearestFriendNumber(UNKNOWN), 
  internalTime(0), 
  nearestOther(UNKNOWN),
  passEval(_passEval), 
  goalShotEval(_goalShotEval),
  positionEval(_positionEval),
  LastBestPos(1000,1000),
  lastActive(0)
{
  tS=NULL;
  wS=NULL;
  isGoalie = _isGoalie;
  ballDist[0] = 0;
  double v = BALL_SPEED_MAX;
  for (int i=1; i<50; i++){
    ballDist[i] = ballDist[i-1] + v;
    v *= BALL_DECAY;
  }
  leftCorner = Vector2d(-FIELD_LENGTH/2,-GOAL_WIDTH/2);
  rightCorner= Vector2d(-FIELD_LENGTH/2,GOAL_WIDTH/2);   
}

PositionModule::~PositionModule()
{
}

bool PositionModule::feedbackWorld(const Snapshot* _wS, TacticalSnapshot* _tS)
{
  // just to undo the unused-parameter-warnings
  if (_wS==NULL){
    cerr  << "POS: feedback -> no wS available" << endl;
    return false;
  }

  // is there a tacticalSnap?
  if(_tS == NULL){
    cerr << _wS->time << "POS: feedback -> no tacticalSnap available" << endl;
    return false;
  }

  // Now everything is checked and fine.
  
  // save the Snapshots 
  wS=_wS;
  tS=_tS;
  
  // positionEval needs the snapshots as well
  positionEval->CalMinMaxX(wS);
  positionEval->CalPlayerPositions(wS);

  oldPosGoal = posGoal;
   
  // SET OFFSIDE VARIABLES
  ourOffsideLine = wS->ourOffsideLine(MAXAGE);
  otherOffsideLine = wS->otherOffsideLine(MAXAGE);
 //   Comm::sendMCmd(LINE(Vector2d(ourOffsideLine,-FIELD_WIDTH/2),Vector2d(ourOffsideLine,FIELD_WIDTH/2))); 
//    Comm::sendMCmd(LINE(Vector2d(otherOffsideLine,-FIELD_WIDTH/2),Vector2d(otherOffsideLine,FIELD_WIDTH/2))); 

  //calculate max. useable stamina
  availableStamina = max (0.0, wS->senseBody.stamina - 
			  EFFORT_DEC_THR * STAMINA_MAX - 2 );


  // shorten the code
  me = wS->my;
  nearestFriendNumber = wS->ourTeamClosestTo(me.pos,20,2);

  internalTime++; 

  myTactNo = tS->ourTeamTacticalNo(me.number);
  if(myTactNo == UNKNOWN) myTactNo = me.number;
  
  nearestOther = wS->otherTeamClosestTo(me.pos);
#ifdef POSACTIONTRACE
  ostrstream temp; 
  switch(tS->ourAction.state){
  case TA_REACT:
    temp <<"posm: ourAction.state TA_REACT performer: ";
    break;
  case TA_PASS:
    temp <<"posm: ourAction.state TA_PASS performer: ";
    break;
  case TA_CONTROL:
    temp << "posm: ourAction.state TA_CONTROL performer: ";
    break;
  default:
    temp <<"posm: ourAction.state= was wei ich performer: ";
  };
  temp << tS->ourAction.performer << char(0);
  { char * str= temp.str();
  PACTIONTR(MLog(str));
  delete str; }

  ostrstream tmp;
  tmp << "posm: Resp. players: " << tS->firstResponsiblePlayer << ", " 
      <<tS->secondResponsiblePlayer << ", " << tS->thirdResponsiblePlayer 
      << " dang. opp: " << tS->firstDangerousOpponent << ", " 
      << tS->secondDangerousOpponent << ", "
      << tS->thirdDangerousOpponent << char(0);
  { char * str = tmp.str();
  PACTIONTR(MLog(str));
  delete str; }
#endif
  return true;
} 


Grade PositionModule::evaluate(AdvSkills* skills,
			       const Snapshot* _wS,
			       TacticalSnapshot* _tS)
{

  Grade g;
  g.set_BAD();
  coveredOpponent = -1;
  lookObject=NULL;
  lookPos.setXY(0,0);
  didIAct=0; // for turnHeadEval
  action=DEFAULT;
  
  // standard position is my position
  posGoal = me.pos;
  
  // do we have a valid snapshot?
  if(wS == NULL)
    {
      cerr << "POS: eval -> no wS available" << endl;
      g.set_IMPOSSIBLE();
      returnGRADE(g);
    }
  
  // are the skills available?
  if(skills == NULL)
    {
      cerr << wS->time << "POS: eval -> no skills available" << endl;
      g.set_IMPOSSIBLE();
      returnGRADE(g);
    }
  
  // is there a tacticalSnap?
  if(tS == NULL){
    cerr << wS->time << "POS: eval -> no tacticalSnap available" << endl;
    g.set_IMPOSSIBLE();
    returnGRADE(g);
  }
  
  // Now everything is checked and fine, so we can start to think.
  
  // when did I tackle the last time? May I move?
  if (wS->lastTackleTime+TACKLE_CYCLES>=wS->time-3)
    {
      // I am not allowed to move
      PACTIONTR("posm: I can't move due to a tackle command!");
      action=TACKLE;
      return(Grade::PERFECT);
    }


  // my ball ?
  if (skills->kickSafe(wS)){ 
    action=NO_ACTION;
    POSTRACE("I can kick ball");
    posGoal = wS->ball.pos;
    priority = 2;
    goToPrecision = KICKABLE_MARGIN/2;
    action=NO_ACTION;
    g.set_BAD();
    EvalGrade=g;
    returnGRADE(g);
  }

  if (!isGoalie)
    {
      CalcStdPriority();
      g.set_GOOD();
      //g.set_ABSOLUTE();
      posGoal=positionEval->GetPlayerPosition(wS->my.number);

      goToPrecision = 3;
      
      if ((tS->ourAction.state==TA_REACT) && 
	  (tS->firstResponsiblePlayer!=UNKNOWN) &&
	  // (tS->secondResponsiblePlayer!=UNKNOWN) &&
	  (//attack player if 
	   //I am firstResp. player
	   ((tS->firstResponsiblePlayer==me.number) ||
	   ((tS->secondResponsiblePlayer==me.number) &&
	    (tS->firstResponsiblePlayer==wS->ourGoalieNo)))
	   ||
	   //I am second resp. and first resp. behind first dangerous opp.
	   ((tS->secondResponsiblePlayer==me.number) &&
	    ((tS->ourTeamDefender(tS->firstResponsiblePlayer)) ||
	     (tS->ourTeamDefMidfield(tS->firstResponsiblePlayer))) &&
	    (tS->firstDangerousOpponent!=UNKNOWN) &&
	    (wS->ourTeam(tS->firstResponsiblePlayer).pos.getX()>
	     wS->otherTeam(tS->firstDangerousOpponent).pos.getX()))
	   ) &&
/*	  !(
	    // der Spieler soll nicht hingehen, wenn beide schon hinter
	    //dem Gegner sind, er selbst ein Abwehrspieler ist 
	    //(der Mittelfeldspieler hat schon abgesichert und wird 
	    //wahrscheinlich der hintere sein) und er nicht firstRespPlayer
	    //ist
	  ((wS->ourTeam(tS->firstResponsiblePlayer).pos.getX()>
	     wS->otherTeam(tS->firstDangerousOpponent).pos.getX()) &&
	      ((wS->ourTeam(tS->secondResponsiblePlayer).pos.getX()>
		wS->otherTeam(tS->firstDangerousOpponent).pos.getX())) &&
	      (me.number!=tS->firstResponsiblePlayer) &&
	      (tS->ourTeamDefender(me.number))
	   )) &&
*/
	  (wS->playMode==PLAY_MODE_PLAY_ON))
	{
	  //Bedingung nochmal berdenken; Abwehrspieler drfen sich nicht
	  //zu weit rausziehen lassen


	  action=PLAYER_ATTACK;

	  if ((tS->ourTeamDefender(me.number)) ||
	      (tS->ourTeamDefMidfield(me.number)))
	    g.set_PERFECT();
	  else
	    g.set_GOOD();
	}
      else if (evalSupportRespPlayer())
	{
	  action=SUPPORT_RESP_PLAYERS;
#ifdef POSACTIONTRACE
	  ostrstream temp;
	  temp << "posm: supp. player no: " << SupportPlayerNo << char(0);
	  char * str=temp.str();
	  PACTIONTR(str);
	  delete str;
#endif
	}
      else if (searchFreePosition())
      {
	  action=GOTO_FREE_POSITION;
      }
      
      else if (shouldICover())
	  {
	      action=COVER_OPPONENT;
	  }
       
    }
  else 
      g =  goalieEvaluate(skills);

#ifdef POSACTIONTRACE
  switch (action)
    {
    case DEFAULT : PACTIONTR("planned action: default");
      break;
    case NO_ACTION: PACTIONTR("planned action: no action");
      break;
    case SUPPORT_RESP_PLAYERS :PACTIONTR("planned action: supp. resp. player");
      break;
    case PLAYER_ATTACK: PACTIONTR("planned action: player attack");
      break;
    case GOTO_FREE_POSITION: PACTIONTR("planned action: free pos.");
      break;
    case COVER_OPPONENT: PACTIONTR("planned action: cover opp.");
      break;
    default: PACTIONTR("planned action: unknown");
    };
#endif


  EvalGrade=g;//I want to know the grade for the turnHeadEvaluate
  returnGRADE(g);
}

bool PositionModule::evalSupportRespPlayer()
{
  int a;
  int b;
  bool act=false;
  int resp[3];

  resp[0]=tS->firstResponsiblePlayer;
  resp[1]=tS->secondResponsiblePlayer;
  resp[2]=tS->thirdResponsiblePlayer;

  for(b=0;(b<3) && (!act);b++)
    for (a=0;(a<1) && (!act);a++)// only first resp.
	if ((resp[a]==positionEval->GetPlayerWatchNo(me.number,b)) &&
	    (resp[a]!=UNKNOWN))
	    act=true;
	    
  SupportPlayerNo=positionEval->GetPlayerWatchNo(me.number,b-1);
  if (SupportPlayerNo!=UNKNOWN) lookObject=&wS->ourTeam(SupportPlayerNo);
  return act;
}

bool PositionModule::shouldICover(){
    if (tS->myTacticalNo == UNKNOWN) return false;
    int opp2cover =wS->otherTeamClosestTo(me.pos);
    if (opp2cover == UNKNOWN) return false;
    if (me.pos.getX()>-FIELD_LENGTH+PENALTY_AREA_LENGTH+10) return false;
    if (tS->ourTeamOffender(tS->myTacticalNo)) return false;
    if (tS->ourTeamOffMidfield(tS->myTacticalNo)) return false;
    if ( (tS->ourAction.state == TA_CONTROL) && (tS->ourBallControl == BC_SAFE) 
	 && (tS->ourTeamDefMidfield(tS->myTacticalNo)) )return false;
    if (wS->ourTeamClosestTo(wS->otherTeam(opp2cover).pos) != me.number) return false;
    if (wS->otherTeam(opp2cover).age > 5) return false;
    if ( (wS->otherTeam(opp2cover).state != VISUAL)
	 && (wS->otherTeam(opp2cover).state != AUDIAL) ) return false;
    if (me.distTo( wS->otherTeam(opp2cover).pos ) > 10) return false;
    return true;
}

bool PositionModule::searchFreePosition() 
{ 

  //  if (positionEval->IsInPlayerArea(me.number,me.pos) )
  //  {
  if (((wS->playMode == PLAY_MODE_FREE_KICK_US) ||(wS->playMode == PLAY_MODE_KICK_IN_US)||
       (wS->playMode == PLAY_MODE_CORNER_KICK_US)) && 
      (tS->firstResponsiblePlayer!=UNKNOWN) && (wS->ourGoalieNo!=UNKNOWN) &&
      (tS->firstResponsiblePlayer!=wS->ourGoalieNo))
	{
	  if (wS->ball.pos.getX()<me.pos.getX())
	    {
              lastActive = wS->time; 
              return true ;
	    }
	  else
	    if (((wS->ball.pos.getX()-FREEPOSAREA_LENGHT)< me.pos.getX()) && 
                (wS->ball.pos.getY()-FREEPOSAREA_WIDTH<me.pos.getY())&&(wS->ball.pos.getY()+FREEPOSAREA_WIDTH>me.pos.getY())) 
	      {
		lastActive=wS->time;
	        return true ;
              }
	    else 
	      {
	
#ifdef FREEPOSTRACE
  ostrstream tmp;
  tmp << "NOT FreePos -> Player too far away " << char(0) ;
  char * str = tmp.str();
  FREEPOSTRC(MLog(str));
  delete str;
#endif
		return false ;
	      }
	} 
      else
        if((wS->playMode == PLAY_MODE_PLAY_ON)&&
          ((tS->ourAction.state == TA_CONTROL ) || 
	  (tS->ourAction.state == TA_PASS)))
          if ((tS->ourTeamOffender(me.number) )||(tS->ourTeamOffMidfield(me.number)))
	    {
	      lastActive = wS->time;
              return true ;
	    }
	  else 
	   {	  
	  
#ifdef FREEPOSTRACE
  ostrstream tmp;
  tmp << "NOT FreePos -> Defense Player " << char(0) ;
  { char * str= tmp.str();
  FREEPOSTRC(MLog(str));
  delete str; }
#endif
	     return false ; 
	   }
        else 
         {

#ifdef FREEPOSTRACE
  ostrstream tmp;
  tmp << "NOT FreePos -> Action not control or pass " << char(0) ;
  { char * str = tmp.str();
  FREEPOSTRC(MLog(str));
  delete str; }
#endif
	    return false ; 
	 }
      // }
      // else
      //{
      // LastBestPos.setX(1000) ;
      //LastBestPos.setY(1000) ;
      //#ifdef FREEPOSTRACE
      //ostrstream tmp;
      //  tmp << "NOT FreePos -> Player not in Playerarea " << char(0) ;
      // char * str = tmp.str();
      // FREEPOSTRC(MLog(str));
      // delete str; 
      //#endif
      //return false ;
      // }
}

Grade PositionModule::turnHeadEvaluate(AdvSkills *skills , 
				       AdvSkills::TurnHeadCommand &thc)
{
  Grade g;
  g.set_AVERAGE();

  //some variables have to be reseted, if we didn't act
  if (!didIAct)
    {
      if (action!=PLAYER_ATTACK) taReportTime=0;
      if (action!=GOTO_FREE_POSITION) 
	LastBestPos.setXY(1000,1000);
    };


  switch (action)
    {
    case PLAYER_ATTACK: 
      if (tS->firstDangerousOpponent!=UNKNOWN)
	{
	  if ( (wS->ball.age>3) || 
	       ( (me.pos.getX()>FIELD_LENGTH/2-PENALTY_AREA_LENGTH-2) && (wS->ball.age>1)) )
	       
	    {
	      thc=skills->turnHead(wS,wS->ball);
	      thc.vwidth=VIEW_WIDTH_NARROW;
	      PACTIONTR(MLog("posm: turning head to ball"));
	    }
	  else
	    {
	      thc=skills->turnHead(wS,wS->otherTeam(tS->firstDangerousOpponent));
	      if (wS->otherTeam(tS->firstDangerousOpponent).age>3)
		thc.vwidth=VIEW_WIDTH_NORMAL;
	      else
		thc.vwidth=VIEW_WIDTH_NARROW;
	      
	      PACTIONTR(MLog("posm: turning head to firstDangrousOpponent"));
	    };
	  g.set_GOOD();
	}
      else
	{
	  PACTIONTR(MLog("posm: firstDangerous Opponent is not set"));
	  g.set_IMPOSSIBLE();
	};
      break;
    case TACKLE:
      {
	// I am not allowed to move, and I tackled, so I don't know
	// the new speed of the ball. Therefore I look at the ball, since
	// I can't do anything useful
	PACTIONTR(MLog("posm: TACKLE -> turning head to ball"));
	thc=skills->turnHead(wS,wS->ball);
	thc.vwidth=VIEW_WIDTH_NARROW;
	// if turn head is ok return perfect, if not try normal behavior
	if ((thc.state == AdvSkills::THOK) || (thc.state ==AdvSkills::THFINISHED))
	  {
	    g.set_PERFECT();
	    break;
	  }
	PACTIONTR(MLog("posm: TACKLE-> turning head to ball not possible"));
      }

    case SUPPORT_RESP_PLAYERS:
	    g.set_GOOD();//no break
	default:
	    if ((isGoalie) ) 
		// (wS->ball.pos.getX()<-FIELD_LENGTH/2.+PENALTY_AREA_LENGTH+5))  
		// Goalie looks at ball all the time and our players whenever the
		//ball is next to our penalty area
	    {
		thc = skills->turnHead(wS,wS->ball);
		PACTIONTR(MLog("posm: turning head to ball"));
	    }
	    else
	    {   
		if ( (wS->ball.age>3) || 
		     ((me.distTo(wS->ball)<8*BALL_SPEED_MAX) && (wS->ball.age>2)) ||
		     ((me.distTo(wS->ball)<4*BALL_SPEED_MAX) && (wS->ball.age>1)) ){
		    thc = skills->turnHead(wS,wS->ball);
		    POSTRACE("turning head to ball");
		    PACTIONTR(MLog("posm: turning head to ball"));
		    if (thc.state == AdvSkills::THOK) break;
		} 
		if (coveredOpponent != -1) {
		    //desiredHeadAng = (wS->otherTeam(coveredOpponent).pos - me.pos).getAngle();
		    thc = skills->turnHead(wS,(wS->otherTeam(coveredOpponent).pos 
					       - me.pos).getAngle());
		    POSTRACE("turning head to opponent "<<coveredOpponent);
		    PACTIONTR(MLog("posm: turning head to coveredOpponent"));
		    if (thc.state == AdvSkills::THOK) break;
		}
		if (lookObject!=NULL)
		{
		    thc= skills->turnHead(wS,*lookObject);
		    PACTIONTR(MLog("posm: turning head to lookObject"));
		    if (thc.state == AdvSkills::THOK) break;
		}
		if (lookPos != Vector2d(0,0))
		{
		    PACTIONTR(MLog("posm: turning head to lookPos"));
		    //desiredHeadAng = (lookGoal.pos - me.pos).getAngle();
		    thc = skills->turnHead(wS, (lookPos - me.pos).getAngle());
		    thc.vwidth=VIEW_WIDTH_NORMAL;
		    if (thc.state == AdvSkills::THOK) break;
		}
		if ( (wS->otherTeamClosestTo(me.pos) != UNKNOWN) &&
		     ((wS->otherTeam(wS->otherTeamClosestTo(me.pos)).state != VISUAL) ||
		     (wS->otherTeam(wS->otherTeamClosestTo(me.pos)).age>2))){
		    thc = skills->turnHead(wS,wS->otherTeam(wS->otherTeamClosestTo(me.pos)));
		    PACTIONTR(MLog("posm: turning head to closest Opponent"));
		    if (thc.state == AdvSkills::THOK) break;
		}
		if ( (tS->firstDangerousOpponent != UNKNOWN) &&
		     ( (wS->otherTeam(tS->firstDangerousOpponent).state != VISUAL) ||
		       (wS->otherTeam(tS->firstDangerousOpponent).age > 2)) ){
		    thc = skills->turnHead(wS,wS->otherTeam(tS->firstDangerousOpponent));
		    PACTIONTR(MLog("posm: turning head to firstDangerousOpponent"));
		    if (thc.state == AdvSkills::THOK) break;
		}
		if ( (tS->secondDangerousOpponent != UNKNOWN) &&
		     ( (wS->otherTeam(tS->secondDangerousOpponent).state != VISUAL) ||
		       (wS->otherTeam(tS->secondDangerousOpponent).age > 3)) ){
		    thc = skills->turnHead(wS,wS->otherTeam(tS->secondDangerousOpponent));
		    PACTIONTR(MLog("posm: turning head to secondDangerousOpponent"));
		    if (thc.state == AdvSkills::THOK) break;
		}
		
		// detect oldest object
		WorldMovedObject lookGoal = wS->ball;
		for(int i = 1; i <= 11; i++){
		  if((wS->ourTeam(i).age != ETERNAL)
		     &&(wS->ourTeam(i).age > lookGoal.age)
		     &&(i != me.number) && 
		     (abs(me.bodyAngleTo(wS->ourTeam(i).pos))<
		      MaxNeckAngle+0.9*NormalViewWidth))
		    lookGoal = wS->ourTeam(i);
		  if((wS->otherTeam(i).age != ETERNAL)&&
		     (wS->otherTeam(i).age > lookGoal.age) &&
		     (abs(me.bodyAngleTo(wS->otherTeam(i).pos))<
		      MaxNeckAngle+0.9*NormalViewWidth))
		    lookGoal = wS->otherTeam(i);
		}
		POSTRACE("oldest Age: " << lookGoal.age << " at: " << lookGoal.pos);
		PACTIONTR(MLog("posm: turning head to oldest object"));
		//desiredHeadAng = (lookGoal.pos - me.pos).getAngle();
		thc = skills->turnHead(wS, (lookGoal.pos - me.pos).getAngle());
		thc.vwidth=VIEW_WIDTH_NORMAL;
		
	    };
    }//switch
  if (!didIAct)
      g.downgrade();
  
  returnGRADE(g);
}

//////////////////////////////////////////////////////////////////////////
//                 The Goalie                                           //
///////////////////////////////////////////////////////////////////////// 
#include "PositionGoalieEval.C"



bool PositionModule::act(AdvSkills* skills,
			 const Snapshot* _wS, TacticalSnapshot* _tS)
{
    didIAct=1; //for turnHeadEval

    // are the skills available?
    if(skills == NULL)
    {
	cerr << "POS: act -> no skills available" << endl;
	return false;
    }

    // do we have a valid snapshot?
    if(wS == NULL)
    {
	cerr << "POS: act -> snapshot is NULL" << endl;
	skills->noAction();
	return true;
    }

    // just to undo the unused-parameter-warnings
    // as soon as the tactical snapshot will be used in this methode it must 
    // be tested not to equal NULL
    if (tS == NULL)
    {
	cerr << "POS: act -> tactSnapshot is NULL" << endl;
	skills->noAction();
	return true;
    }

   
    POSDRAWTRACE(POSITION(posGoal));
    POSTRACE("Going to position: " << posGoal);
    POSTRACE("priority " << priority << "; Przision: " << goToPrecision << 
	     "; Stamina to use: " << calcUseableStamina() << 
	     "; overall Stamina:" << wS->senseBody.stamina);
    if (isGoalie) return GoalieAct(skills); 



    switch(action)
      {
      case TACKLE:
	{
	  skills->noAction();
	  return (true);
	  break;
	};
      case PLAYER_ATTACK:
	{
	    //report tatctical action
	    // if any player of our team except for me is near the
	    // ball, be quiet. I do not use the reachTime but real
	    // distance since reachTime is instable and I don't want
	    // to fill up our EmChannel with nonsens
	    int clst = wS->ourTeamClosestTo(wS->ball.pos);
	    if ((clst != UNKNOWN) && (clst == me.number) &&
		(wS->ourTeam(clst).distTo(wS->ball) < 2*BALL_SPEED_MAX))
	    {
	    
		if ((tS->ourAction.state != TA_REACT) ||
		    (taReportTime>6) || (taReportTime==0))
		{
		    // same ta as before. Occasionally remind others
		    // else be quiet
		    tS->reportOurTacticalAction(TA_REACT, me.number);
		    taReportTime=1;
		}
		else taReportTime+=1;
	    };
	    playerAttack(tS->firstDangerousOpponent);
	    //stdact();
	    break;
	}
	case COVER_OPPONENT: 
	    posGoal=coverOpponent();
	    stdact();
	    break;
	case SUPPORT_RESP_PLAYERS: 
	    SupportPlayer();
	    stdact();
	    break;
	case GOTO_FREE_POSITION:
	    // gotoFreePos();
	    posGoal=gotoFreePos();
	    stdact();
	    break;
	default:
	    stdact();
    };

    return true;
}

void PositionModule::stdact()
{
#ifdef POSACTIONTRACE
    ostrstream temp;
    temp << "Goint to position: "<< posGoal<<" precision: " << goToPrecision
	 << " with priority: " << priority   
	 << " Stamina to use: " << calcUseableStamina() << char(0);
    { char * str= temp.str();
    PACTIONTR(MLog(str));
    delete str; }
#endif
    // keep care of standard situation areas
    respectStandardSituations(wS, tS);
    // avoid player collisions
    avoidPlayerCollision(wS);
    AdvSkills::TurtleCommand tc 
	= skills->goToPosition(wS, posGoal, goToPrecision, calcUseableStamina());
  
  
    if(tc.state == AdvSkills::TFINISHED)
    {
      
	AdvSkills::TurtleCommand tc //zum testen
	    = skills->goToPosition(wS, posGoal, 2, 10);
	

	if (tc.state==AdvSkills::TFINISHED)
	{
	  AdvSkills::TurnCommand tuc; 
	    //if nothing to do, turn body to goal or ball
	  if (me.pos.getX() > FIELD_LENGTH/2 - PENALTY_AREA_LENGTH-1){
	    Vector2d corner = Vector2d(FIELD_LENGTH/2, 0.85 * GOAL_WIDTH/2);
	    if (wS->ball.pos.getY()>0)
	      corner = Vector2d(FIELD_LENGTH/2, - 0.85 * GOAL_WIDTH/2);
	    tuc = skills->turn(wS,me.bodyDir+
			       me.bodyAngleTo(corner));
	  }
	  else{
		tuc = skills->turn(wS,me.bodyDir+me.bodyAngleTo(wS->ball));
	  }
	    if(tuc.state == AdvSkills::RFINISHED)
		skills->noAction();
	    else
		skills->exec(wS, tuc);
	}
	else skills->exec(wS,tc);
    }
    else
	skills->exec(wS,tc);

};

////////////////////////////////////////////////////////////////////
//                                                                //
// functions used by more than one eval-function                  //
//                                                                //
////////////////////////////////////////////////////////////////////


bool PositionModule::otherPlayerFree(const Snapshot *_wS,int num, double distance){
  
    // opponent "num" valid ? and our nearest is known
    if(
	(wS->otherTeam(num).age < MAXAGE)
	&&
	(wS->ourTeamClosestTo(wS->otherTeam(num).pos) != UNKNOWN)
	){
	// in standard situations players near to the ball are NOT free
	if(
	    (wS->playMode == PLAY_MODE_KICK_IN_THEM)
	    ||
	    (wS->playMode == PLAY_MODE_FREE_KICK_THEM)
	    ||
	    (wS->playMode == PLAY_MODE_CORNER_KICK_THEM)
	    ||
	    (wS->playMode == PLAY_MODE_GOAL_KICK_THEM)
	    ){
	    if(dist(wS->otherTeam(num).pos, wS->ball.pos) < FREE_KICK_FORBIDDEN_AREA)
		return false;
	}
    
	// our player near to opponent "num"
	if (dist(wS->otherTeam(num).pos,
		 wS->ourTeam(wS->ourTeamClosestTo(wS->otherTeam(num).pos)).pos)
	    < distance){ 
	    // if our player is not nearer to another player
	    if  (wS->otherTeamClosestTo(wS->ourTeam(wS->ourTeamClosestTo(wS->otherTeam(num).pos)).pos)
		 == num)
		return false;
	    else
		return true;
	}
    }

    // in doubt, return true
    return true;
}


bool PositionModule::ourPlayerFree(const Snapshot *_wS,int num, double distance){

    // player valid, opponent near ?
    if ((wS->ourTeam(num).age < MAXAGE)
	&&(wS->otherTeamClosestTo(wS->ourTeam(num).pos) != UNKNOWN))
	if (dist(wS->ourTeam(num).pos,
		 wS->otherTeam(wS->otherTeamClosestTo(wS->ourTeam(num).pos)).pos)
	    < distance)
	    return false;
	else
	    return true;
    // in doubt, return true
    return true;  
}


double PositionModule::calcUseableStamina(){
    double useableStam=max_useable_Stamina[priority];
    double maxStam=wS->senseBody.stamina-(1+save_Stamina[priority])*
	EFFORT_DEC_THR*STAMINA_MAX;
		 
    return max(min(useableStam,maxStam),0.);
}


void PositionModule:: CalcStdPriority()
{
    priority=2; //standard: not to fast and not to slow
    if (wS->playMode==PLAY_MODE_PLAY_ON)
    {
	priority=1; 
	if ((tS->ourTeamDefender(me.number)) && 
	    (wS->ball.pos.getX()<me.pos.getX()))
	    priority=0;
	//defenders have to hurry if the ball is behind them

	if ((tS->ourTeamOffender(me.number)) &&
	    (me.pos.getX()-5<wS->ourOffsideLine(MAXAGE)) &&
	    (positionEval->GetOffensive()<.5))
	    priority=3;
	//offenders may save stamina if the ball is in our defense
	   
    }
};

////////////////////////////////////////////////////////////////////
//                                                                //
//              function used by act                              //
//                                                                //
////////////////////////////////////////////////////////////////////



void PositionModule::respectStandardSituations(const Snapshot* _wS,
					       const TacticalSnapshot* ){
    switch(wS->playMode){
	case PLAY_MODE_UNKNOWN : 
	    break;
	case PLAY_MODE_BEFORE_KICK_OFF :
	    posGoal.setX(min(posGoal.getX(),-PLAYER_SIZE));
	    break;
	case PLAY_MODE_TIME_OVER :
	    posGoal = me.pos;
	    break;
	case PLAY_MODE_PLAY_ON : 
	    break;
	case PLAY_MODE_KICK_OFF_US : 
	case PLAY_MODE_KICK_OFF_THEM :
	    posGoal.setX(min(posGoal.getX(),-PLAYER_SIZE));    
	    break;
	case PLAY_MODE_KICK_IN_US : 
	    break;
	case PLAY_MODE_KICK_IN_THEM : 
	    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
	       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		if((posGoal - me.pos).getAngle() - (wS->ball.pos - me.pos).getAngle() > 0)//right
		    posGoal = (wS->ball.pos - me.pos).perp()*(1/norm(wS->ball.pos - me.pos));
		else // left
		    posGoal = (me.pos - wS->ball.pos).perp()*(1/norm(me.pos - wS->ball.pos));
	    break;
	case PLAY_MODE_FREE_KICK_US : 
	    break;
	case PLAY_MODE_FREE_KICK_THEM :
	    if( (!wS->ball.inOtherPenaltyArea()) 
		&& (wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
		&&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		if((posGoal - me.pos).getAngle() - (wS->ball.pos - me.pos).getAngle() > 0)//right
		    posGoal = (wS->ball.pos - me.pos).perp()*(1/norm(wS->ball.pos - me.pos));
		else // left
		    posGoal = (me.pos - wS->ball.pos).perp()*(1/norm(me.pos - wS->ball.pos));
	    break;    
	case PLAY_MODE_CORNER_KICK_US : 
	    break;
	case PLAY_MODE_CORNER_KICK_THEM :
	    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
	       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		posGoal = me.pos;
	    break;
	case PLAY_MODE_GOAL_KICK_US : 
	    break;
	case PLAY_MODE_GOAL_KICK_THEM :
	    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
	       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		posGoal = me.pos;
	    break;
	case PLAY_MODE_GOAL_US :
	    posGoal = me.pos;
	    break;
	case PLAY_MODE_DROP_BALL :
	    break;
	case PLAY_MODE_GOAL_THEM :
	    posGoal = me.pos;
	    break;    
	case PLAY_MODE_OFFSIDE_US :
	    break;
	case PLAY_MODE_OFFSIDE_THEM :
	    if((wS->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
	       &&(dist(wS->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		if((posGoal - me.pos).getAngle() - (wS->ball.pos - me.pos).getAngle() > 0)//right
		    posGoal = (wS->ball.pos - me.pos).perp()*(1/norm(wS->ball.pos - me.pos));
		else // left
		    posGoal = (me.pos - wS->ball.pos).perp()*(1/norm(me.pos - wS->ball.pos));
	    break;    
	case PLAY_MODE_HALF_TIME :
	    posGoal = me.pos;
	    break;        
	case PLAY_MODE_TIME_EXTENDED:
	    posGoal = me.pos;    
	    break;
    default:
      break;
    }

}


void PositionModule::avoidPlayerCollision(const Snapshot *_wS){

    if (nearestOther!=UNKNOWN)
	posGoal=goAroundOpponent(posGoal,wS->otherTeam(nearestOther),2,
				 (me.pos-posGoal).getX());
    if (nearestFriendNumber!=UNKNOWN)
	posGoal=goAroundOpponent(posGoal,wS->ourTeam(nearestFriendNumber),2,
				 (me.pos-posGoal).getX());

    // if otherPlayer or ourPlayer (incl. me!) in my way
    /*  if(dist(posGoal, me.pos) > PLAYER_SIZE){
	if(
	(wS->otherTeamInCorridor(me.pos, me.pos +
	radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle()),
	4*PLAYER_SIZE) > 0)
	||
	(wS->ourTeamInCorridor(me.pos, me.pos +
	radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle()),
	4*PLAYER_SIZE) > 1)
	){
	if((wS->ball.pos - me.pos).getAngle() > 0) // ball right of me
	// ball right => side step right of player
	// to avoid interference with free kick areas
	posGoal = radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle())
	+ radAng2Vector2d(5*PLAYER_SIZE, (posGoal-me.pos).getAngle()).perp();
	else // vice versa
	posGoal = radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle())
	- radAng2Vector2d(5*PLAYER_SIZE, (posGoal-me.pos).getAngle()).perp();
	}
	}
    */
}
