//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// $Id: standardsituationmodule.C,v 1.84 2002/06/14 07:57:32 gauer Exp $
// $manager: gauer$
//
//////////////////////////////////////////////////////////////////////

#include "standardsituationmodule.h"
#include "Config.h"

// include the outsourced methods
#include "stdPlan.C"
#include "stdCalc.C"
#include "stdGoalie.C"
#include "stdKicksUs.C"
#include "stdKicksThem.C"

// ------------------------------------------------------------------

StandardSituationModule::StandardSituationModule(bool isGoalie, 
					         const PassEval *_passEval, 
						 const GoalShotEval *_goalShotEval, 
						 const PositionEval *_positionEval) :
  kickUsPlayer(-1),
  goalShotEval(_goalShotEval),
  passEval(_passEval),
  positionEval(_positionEval)
{
  // the agents modules
  if (goalShotEval == NULL)
    {
      ERRORTRACE1( "STANDARDSITUATIONMODULE: goalShotEval == NULL" );
    }
  if (passEval == NULL)
    {
      ERRORTRACE1( "STANDARDSITUATIONMODULE: passEval == NULL" );
    }

  readAgent_conf();


  goalieMovesInCurrentSituation = 0;
  goalieDoesFreeKick = false;
  goalieMayMove = false;
  goalieShouldMoveAgain = false;
  nextCycleDoNothing = false;
  weAreKickingOff = false;
  actedThisCycle = false;
  wantToTurnHead = false;
  timeLeft = DROP_BALL_TIME - 10; //[C]
  stdSitFinished = false;

#ifdef USE_MARKER_TRACE
  playmode = PLAY_MODE_UNKNOWN;
#endif

  lastTime_PLAY_ON = -1;
  lastTime_KICK_OFF_US = -1 - standardWaitingTime;

  if (isGoalie)    
    OTHERTRACE( "StandardSituationModule (STDSIT) starting ... as goalie" )
  else
    OTHERTRACE( "StandardSituationModule (STDSIT) starting ... as fieldplayer" )
  iAmGoalie = isGoalie;
  isSpecialSituation = false;
  maxAge = 20;
  waitingTime = -1;
  nextTurnAction = TURN_TO_LEFT_SIDE;

  // calculating the positions where the goalie should move when doing
  // a free kick after moving
  goalieMovePositions = GOALIE_MOVE_POSITIONS;
  if (isGoalie)
    {
      OTHERTRACE( "STDSIT: calculating Goalie Move positions)" )
      // [C]*
      double yStep = (PENALTY_AREA_WIDTH - 5.0) / (goalieMovePositions - 1.0);
      for (int i=0; i<goalieMovePositions; i++)
	{
	  goalieMovePos[i].setXY( -FIELD_LENGTH/2.0 + PENALTY_AREA_LENGTH - 1.0, 
				  -PENALTY_AREA_WIDTH / 2.0 + 2.5 + i * yStep  );
	  OTHERTRACE( i << ": " << goalieMovePos[i] )
	}
    }

  // calculating Permutations
  OTHERTRACE("setting up permutations");
  Perm = new Permutations(5);
}

StandardSituationModule::
~StandardSituationModule()
{
  delete Perm;
  OTHERTRACE( "StandardSituationModule leaving ..." )
}  


bool 
StandardSituationModule::
feedbackAudio(AudioMsg &msg)
{
  if (!msg.getState() || !msg.isCoachMsg()) {
    DECISIONTRACE("got bad msg");
    return false;
  }

  Int3B player(12);
  bool ok(false);

  msg >> ok;

  if (!ok) {
    DECISIONTRACE("got wrong msg");
    return false;
  }

  msg >> player;

  if (player < 1 || player > 11) {
    DECISIONTRACE("got bad player" << kickUsPlayer);
    return false;
  }
  else
    kickUsPlayer = int(player);

  DECISIONTRACE("according to coach acting Player is " << kickUsPlayer);

  return true;
};

bool 
StandardSituationModule::
feedbackWorld(const Snapshot* worldSnap, 
	      TacticalSnapshot* tacticalSnap)
{
  if (worldSnap == NULL)
    ERRORTRACE( "feedbackWorld: worldSnap == NULL" );
  if (tacticalSnap == NULL)
    ERRORTRACE( "feedbackWorld: tacticalSnap == NULL" );

  if (worldSnap->playMode == PLAY_MODE_PLAY_ON ||
      playmode != worldSnap->playMode)
    {
      stdSitFinished = false;
      kickUsPlayer = -1;
      DECISIONTRACE("playmode != last playmode setting stdSitFinished to " << stdSitFinished);
    }

#ifdef USEDRAWTRACE
  // for drawing free-kick-circles
  switch (worldSnap->playMode)
    {
    case PLAY_MODE_OFFSIDE_THEM:
    case PLAY_MODE_OFFSIDE_US:
    case PLAY_MODE_FREE_KICK_THEM:
    case PLAY_MODE_FREE_KICK_US:
    case PLAY_MODE_CORNER_KICK_US:
    case PLAY_MODE_CORNER_KICK_THEM:
    case PLAY_MODE_KICK_IN_US:
    case PLAY_MODE_KICK_IN_THEM:
      DRAWTRACE( CIRCLE(worldSnap->ball.pos, FREE_KICK_FORBIDDEN_AREA) );
    default: ;
    }
#endif   

  if (worldSnap->playMode != playmode) {
    playmode = worldSnap->playMode;
#ifdef USE_MARKER_TRACE
    if (worldSnap->my.number == MARKER_NUM) {
      ostrstream out;
      out << playmode << (char)0;
      char * str = out.str();
      Comm::sendMCmd(MMarker(Vector2d(0,0), str, 1));
      delete str;
    }
#endif
  }

  return true;
}


Grade 
StandardSituationModule::
evaluate(AdvSkills* skills,
	 const Snapshot* worldSnap, 
	 TacticalSnapshot* tacticalSnap)
{
  isSpecialSituation = false;
  actedThisCycle = false;
  Grade value;
  
  OTHERTRACE("TA: " << tacticalSnap->ourAction.state );
  OTHERTRACE("Tt: " << tacticalSnap->ourAction.startingTime );
  OTHERTRACE("TP: " << tacticalSnap->ourAction.performer );
  OTHERTRACE("TR: " << tacticalSnap->ourAction.receiver );
  OTHERTRACE("TQ: " << tacticalSnap->ourAction.quality );

  if (skills == NULL)
    {
      ERRORTRACE( "eval: skills == NULL" );
      value.set_IMPOSSIBLE();
      returnGRADE(value);
    }
  if (worldSnap == NULL)
    {
      ERRORTRACE( "eval: worldSnap == NULL" );
      value.set_IMPOSSIBLE();
      returnGRADE(value);
    }
  if (tacticalSnap == NULL)
    {
      ERRORTRACE( "eval: tacticalSnap == NULL" );
      value.set_IMPOSSIBLE();
      returnGRADE(value);
    }

  if (stdSitFinished) {
    value.set_BAD();
    DECISIONTRACE("stdsit finished returning BAD");
    returnGRADE(value);
  }

  // the evaluation
  switch (worldSnap->playMode)
    {
      // case PLAY_MODE_UNKNOWN:
    case PLAY_MODE_TIME_OVER:
      value.set_IMPOSSIBLE();
      break;
    case PLAY_MODE_TIME_EXTENDED:
      timeLeft = DROP_BALL_TIME - 10; //[C]
      value.set_ABSOLUTE();
      weAreKickingOff = (worldSnap->ourSide == SIDE_LEFT);
      if (weAreKickingOff)
	DECISIONTRACE( "Extended time, our kick off" )
      else
	DECISIONTRACE( "Extended time, other kick off" )
      break;
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_FREE_KICK_US :
      timeLeft--;
      value = evalFreeKickUs(skills, worldSnap, tacticalSnap);
      if (worldSnap->my.number == kickUsPlayer)
	value.set_ABSOLUTE();
      break;
    case PLAY_MODE_OFFSIDE_US :
    case PLAY_MODE_FREE_KICK_THEM :
      {
	timeLeft--;
	// Do we need more than one player to build a wall
	bool iWillAct(false);
	if (!iAmGoalie) {
	  actCount = calcPlayersAndPos4FreeKickThem(worldSnap, tacticalSnap, 
						    actPlayer, targetPos);
	  for (int i = 0; i < actCount; i++) 
	    if (iWillAct = worldSnap->my.number == actPlayer[i]) break;
	}

	if ((iAmGoalie &&
	     sqrnorm(worldSnap->ourGoal.pos - worldSnap->ball.pos) <= sqr(maxWallDist)) ||
	    iWillAct)
	  {
	    // if the opponent goalie will perform a free kick in his penalty area
	    // do not build a wall (because he will move!)
	    if (worldSnap->ball.inOtherPenaltyArea() &&
		tacticalSnap->firstDangerousOpponent == worldSnap->otherGoalieNo)
	      value.set_BAD();
	    else
	      value.set_ABSOLUTE();
	  }
	else
	  value.set_BAD();
      }
    break;
    case PLAY_MODE_GOAL_US :
      timeLeft--;
      weAreKickingOff = false;
      value.set_ABSOLUTE();
      DECISIONTRACE( "Oh we scored a goal (through the other goalies trouserholders)" );
      break;
    case PLAY_MODE_GOAL_THEM :
      timeLeft--;
      weAreKickingOff = true;
      value.set_ABSOLUTE();
      DECISIONTRACE( "Oh my god, the other team scored a goal" );
      break;      
    case PLAY_MODE_KICK_OFF_US :
      value.set_ABSOLUTE();
      weAreKickingOff = true;
      break;
    case PLAY_MODE_KICK_OFF_THEM :
      timeLeft = DROP_BALL_TIME - 10; //[C]
      weAreKickingOff = false;
      value.set_ABSOLUTE();
      break;
    case PLAY_MODE_BEFORE_KICK_OFF :
      timeLeft = DROP_BALL_TIME - 10; //[C]
      if  ( ( (worldSnap->ourSide == SIDE_LEFT) && 
	      (worldSnap->time == 0) ) ||
	    ( (worldSnap->ourSide == SIDE_RIGHT) &&  
	      (worldSnap->time == HALF_TIME_IN_CYCLES) )  ||
	    ( (worldSnap->ourSide == SIDE_LEFT) && 
	      (worldSnap->time == 2*HALF_TIME_IN_CYCLES) ) 
	    )
	{
	  // start of one of the half-times or the extended time
	  weAreKickingOff = true;
	}
      if  ( (worldSnap->time == 0)  ||
	    (worldSnap->time == HALF_TIME_IN_CYCLES ) ||
	    (worldSnap->time == 2*HALF_TIME_IN_CYCLES) 
	    )
	lastTime_PLAY_ON = worldSnap->time-1;  // the goalie should not catch if a new 
                                               // half-time just started
      
      value.set_ABSOLUTE();
      if (weAreKickingOff)
	DECISIONTRACE( "Check your trikots, we are before our kick-off" )
      else
	DECISIONTRACE( "Check your trikots, we are before their kick-off" )
      break;
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_CORNER_KICK_US :
      value = evalRemainingKicksUs(skills, worldSnap, tacticalSnap, worldSnap->playMode);
      if (worldSnap->my.number == kickUsPlayer)
	value.set_ABSOLUTE();
      break;
    case PLAY_MODE_GOAL_KICK_US :
      timeLeft--;
      {
	value.set_BAD();

	if (tacticalSnap->ourAction.state == TA_PASS &&
	    tacticalSnap->ourAction.receiver != worldSnap->my.number &&
	    tacticalSnap->ourAction.performer != worldSnap->my.number) {
	  stdSitFinished = true;
	  returnGRADE(value);
	}

	Vector2d leftPos(-FIELD_LENGTH/2.+PENALTY_AREA_LENGTH-3., -PENALTY_AREA_WIDTH/2.0+3.),
	  rightPos(-FIELD_LENGTH/2.+PENALTY_AREA_LENGTH-3., +PENALTY_AREA_WIDTH/2.0-3.);
	switch(getOurTacticalSystem(tacticalSnap))
	  {
	  case 442:
	    targetPos[1] = leftPos;
	    targetPos[4] = rightPos;
	    if (worldSnap->my.number == 2 ||
		worldSnap->my.number == 5)
	      value.set_ABSOLUTE();
	    break;
	  case 343:
	  case 352:
	  default:
	    targetPos[1] = leftPos;
	    targetPos[3] = rightPos;
	    if (worldSnap->my.number == 2 ||
		worldSnap->my.number == 4)
	      value.set_ABSOLUTE();
	    break;
	  }
	if (worldSnap->my.inOurPenaltyArea())
	  value.set_ABSOLUTE();
	
	if (iAmGoalie)
	  {
	    if (nextCycleDoNothing)
	      {
		value.set_BAD();
		break;
	      }
	    if (waitingTime < 0)
	      waitingTime = standardWaitingTime;
	    value.set_ABSOLUTE();
	    goalieMayMove = false;
	    goalieShouldMoveAgain = false;
	  }
	else
	  {
	    if (waitingTime < 0)
	      waitingTime = 8; //[C]
	  }
      }
      break;
    case PLAY_MODE_KICK_IN_THEM :
      {
	timeLeft--;
	actCount = calcPlayersAndPos4KickInThem(worldSnap, tacticalSnap, 
						actPlayer, targetPos);
	                      
	bool iWillAct(false);                      
	for (int i = 0; i < actCount; i++) 
	  if (iWillAct = worldSnap->my.number == actPlayer[i]) break;
	
	if (iWillAct)
	  value.set_ABSOLUTE();
	else
	  value.set_BAD();
	break;  
      } 
    case PLAY_MODE_CORNER_KICK_THEM :
      {
	timeLeft--;
	actCount = calcPlayersAndPos4CornerKickThem(worldSnap, tacticalSnap, 
						    actPlayer, targetPos);
	
	bool iWillAct(false);
	for (int i = 0; i < actCount; i++) 
	  if (iWillAct = worldSnap->my.number == actPlayer[i]) break;
	
	if (iWillAct)
	  value.set_ABSOLUTE();
	else
	  value.set_BAD();
	break;
      }
    case PLAY_MODE_GOAL_KICK_THEM :
      value.set_IMPOSSIBLE();
      timeLeft--;
      break;
    case PLAY_MODE_HALF_TIME :
      timeLeft = DROP_BALL_TIME - 10; //[C]
      value.set_ABSOLUTE();
      weAreKickingOff = (worldSnap->ourSide == SIDE_RIGHT);
      if (weAreKickingOff)
	DECISIONTRACE( "Half-Time, our kick-off")
      else
	DECISIONTRACE( "Half-Time, other kick-off")
      break;
    case PLAY_MODE_PLAY_ON:
      timeLeft = DROP_BALL_TIME - 10; //[C]
      waitingTime = -1;
      weAreKickingOff = false;
      goalieMovesInCurrentSituation = 0;
      // looking for standardsituations that have been in the last cycles
      if (worldSnap->time > lastTime_PLAY_ON)
	nextCycleDoNothing = false;
      lastTime_PLAY_ON = worldSnap->time;
      if (myPlan.planExists())
	value.set_ABSOLUTE();
      else
	value.set_IMPOSSIBLE();
      break;
    case PLAY_MODE_DROP_BALL :
      goalieMovesInCurrentSituation = 0;
      value.set_IMPOSSIBLE();
      break;
    case PLAY_MODE_PENALTY_KICK_US:
      if (worldSnap->time == 0)
        value.set_ABSOLUTE();
      else 
	value.set_IMPOSSIBLE();
      break;
    case PLAY_MODE_PENALTY_KICK_THEM:
      if (worldSnap->time == 0)
	value.set_ABSOLUTE();	
      else
	value.set_IMPOSSIBLE();
      break;
    default:
      value.set_IMPOSSIBLE();
      ERRORTRACE( "Playmode Unkown !?" );
      break;
    } 
  DECISIONTRACE( "evaluate returning " << value << ", playmode = " << worldSnap->playMode );
  OTHERTRACE( "Time left: " << timeLeft );

  returnGRADE(value);
}

Grade 
StandardSituationModule::
turnHeadEvaluate(AdvSkills *skills, AdvSkills::TurnHeadCommand &thc)
{
  Grade g;

  thc = thcmd;

  if (actedThisCycle && wantToTurnHead) {
    g.set_ABSOLUTE();
  }
  else {
    g.set_IMPOSSIBLE();
  }

  returnGRADE(g);
}

bool 
StandardSituationModule::
act(AdvSkills* skills,
    const Snapshot* worldSnap,
    TacticalSnapshot* tacticalSnap)
{

  DECISIONTRACE("kickUsPlayer: " << kickUsPlayer);

  actedThisCycle=true;
  wantToTurnHead=false;
  if (skills == NULL)
    {
      ERRORTRACE( "act: skills == NULL" );
      return false;
    }
  if (worldSnap == NULL)
    {
      ERRORTRACE( "act: wordlSnap == NULL" );
      skills->noAction();
      return false;
    }
  if (tacticalSnap == NULL)
    {
      ERRORTRACE( "act: tacticalSnap == NULL" );
      skills->noAction();
      return false;
    }

  // do we have to speak because we see the ball in a new standardsituation
  // where the ball might have been moved 
  // [C]
  if ( (worldSnap->ball.age < worldSnap->time - lastTime_PLAY_ON) &&
       (worldSnap->time - lastTime_PLAY_ON < 9)
     )

    isSpecialSituation = true;
  else
    isSpecialSituation = false;

  switch (worldSnap->playMode)
    {
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_GOAL_THEM :
      return doingKickOff(skills, worldSnap, tacticalSnap);
    case PLAY_MODE_KICK_IN_US :
      {
	bool iWillAct(false);
	if (12 > kickUsPlayer && kickUsPlayer > 0)
	  actPlayer[0] = kickUsPlayer;
	for (int i = 0; i < actCount; i++)
	  if(iWillAct = worldSnap->my.number == actPlayer[i]) break;

	if (iWillAct)
	  return doingKickInUs(skills, worldSnap,
			       actPlayer, targetPos, actCount);
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
	break;
      }
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_KICK_IN_THEM :
      {
	bool iWillAct(false);
	for (int i = 0; i < actCount; i++)
	  if(iWillAct = worldSnap->my.number == actPlayer[i]) break;

	if (iWillAct)
	  return doingFieldPlayerKicksThem(skills, worldSnap,
					   actPlayer, targetPos, actCount);
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
	break;
      }
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_FREE_KICK_US :
      {
	// if the goalie does it all ...
	if ( goalieDoesFreeKick )
	  return doingGoalieFreeKickUs(skills, worldSnap, tacticalSnap);
	    
	if (iAmGoalie)
	  {
	    // check if our goalie should walk back to our goal
	    Vector2d target(-FIELD_LENGTH/2.0+3.0, 0.0);
	    AdvSkills::TurtleCommand 
	      tcmd = skills->goToPosition(worldSnap, target, 2.0);
	    if (tcmd.state != AdvSkills::TFINISHED)
	      {
		skills->exec(worldSnap, tcmd);
		DECISIONTRACE( "Goalie walking back to our goal @ " << target );
	      }
	    else
	      {
		// turn goalie towards our ball if necessary
		AdvSkills::TurnCommand 
		  tcmd = skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->ball.pos) 
				      + worldSnap->my.bodyDir);
		if (tcmd.state == AdvSkills::RANGLE ||
		    tcmd.state == AdvSkills::ROK)
		  {
		    skills->exec(worldSnap, tcmd);
		    DECISIONTRACE( "Goalie turning towards the ball" );
		  }
		else 
		  {
		    skills->noAction();
		    DECISIONTRACE( "Goalie doing nothing" );
		  }
	      }
	    lookAround(skills, worldSnap, worldSnap->ball.pos);
	    DECISIONTRACE( "Goalie looking around the ball" );
	    return true;
	  } // if (iAmGoalie)

	bool iWillAct(false);
	if (12 > kickUsPlayer && kickUsPlayer > 0)
	  actPlayer[0] = kickUsPlayer;
	for (int i = 0; i < actCount; i++)
	  if(iWillAct = worldSnap->my.number == actPlayer[i]) break;

	if (iWillAct)
	  // if we are close to the edge of the field
	  // [C]
	  if (fabs(fabs(worldSnap->ball.pos.getY())-FIELD_WIDTH/2) < 2)
	    // do a kick in
	    return doingKickInUs(skills, worldSnap,
				 actPlayer, targetPos, actCount);
	  else
	    // do a free kick
	    return doingFreeKickUs(skills, worldSnap,
				   actPlayer, targetPos, actCount);
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
      break;
      }
    case PLAY_MODE_OFFSIDE_US :    
    case PLAY_MODE_FREE_KICK_THEM :
      {
	if (iAmGoalie && 
	    sqrnorm(worldSnap->ourGoal.pos-
		    worldSnap->ball.pos) <= sqr(maxWallDist))
	  return doingGoalieFreeKickThem(skills, worldSnap);
	
	bool iWillAct(false);
	for (int i = 0; i < actCount; i++)
	  if(iWillAct = worldSnap->my.number == actPlayer[i]) break;

	if (iWillAct)
	  {
	    DECISIONTRACE( " Doing fieldplayer freekick them" );
	    return doingFieldPlayerKicksThem(skills, worldSnap, 
					     actPlayer, targetPos, actCount);
	  }
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
	break;
      }
    case PLAY_MODE_CORNER_KICK_US :
      {
	bool iWillAct(false);
	if (12 > kickUsPlayer && kickUsPlayer > 0)
	  actPlayer[0] = kickUsPlayer;
	for (int i = 0; i < actCount; i++)
	  if(iWillAct = worldSnap->my.number == actPlayer[i]) break;

	if (iWillAct)
	  return doingCornerKickUs(skills, worldSnap, 
				   actPlayer, targetPos, actCount);
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
	break;
      }
    case PLAY_MODE_GOAL_KICK_US :
      if (iAmGoalie)
	return doingGoalieGoalKickUs(skills, worldSnap, tacticalSnap);
      else
	return doingFieldPlayerGoalKickUs(skills, worldSnap, tacticalSnap);
//     case PLAY_MODE_GOAL_KICK_THEM :

    case PLAY_MODE_HALF_TIME :
      {
	skills->noAction();
	DECISIONTRACE( " half-time -> act does not do a thing" );
	return true;
      }
    case PLAY_MODE_PLAY_ON:
      {
	if (myPlan.planExists())
	  {
	    if (!myPlan.executePlan(skills, worldSnap))
	      {
		DECISIONTRACE ( " act: plan failed => doing nothing ");
		skills->noAction();
		return false;
	      }
	    return true;
	  }
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing nothing" );
	    skills->noAction();
	    return false;
	  }	
      }
    break;
    case PLAY_MODE_PENALTY_KICK_US:
	if (worldSnap->time == 0)
	{
	    if (worldSnap->my.distTo(penaltyKickFieldplayerPos) > 0.5)
	    {
		DECISIONTRACE( "Moving to " << penaltyKickFieldplayerPos );
		skills->basicMove(penaltyKickFieldplayerPos.getX(), penaltyKickFieldplayerPos.getY());
		return true;
	    }
	    else
	    {
		DECISIONTRACE( "Have already moved, trying to face the goal" );
		AdvSkills::TurnCommand tcmd = 
		    skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->otherGoal)+
				 worldSnap->my.bodyDir);
		if (tcmd.state != AdvSkills::RFINISHED)
		    skills->exec(worldSnap, tcmd);
		else
		    skills->noAction();
		thcmd = skills->turnHead(worldSnap, worldSnap->my.bodyDir);
		wantToTurnHead = true;
		return true;
	    }
	}
	else
	{
	    DECISIONTRACE( "I am not responsible for " << worldSnap->playMode );
	    skills->noAction();
	    return false;
	}

    case PLAY_MODE_PENALTY_KICK_THEM:
	if (worldSnap->time == 0)
	{
	    if (worldSnap->my.distTo(penaltyKickGoaliePos) > 0.5)
	    {
		DECISIONTRACE( "Moving to " << penaltyKickGoaliePos);
		skills->basicMove(penaltyKickGoaliePos.getX(), penaltyKickGoaliePos.getY());
		return true;
	    }
	    else
	    {
		DECISIONTRACE( "Have already moved, trying to face the opponent" );
		AdvSkills::TurnCommand tcmd = 
		    skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->otherGoal)+
				 worldSnap->my.bodyDir);
		if (tcmd.state != AdvSkills::RFINISHED)
		    skills->exec(worldSnap, tcmd);
		else
		    skills->noAction();
		thcmd = skills->turnHead(worldSnap, worldSnap->my.bodyDir);
		wantToTurnHead = true;
		return true;
	    }
	}
	else
	{
	    DECISIONTRACE( "I am not responsible for " << worldSnap->playMode );
	    skills->noAction();
	    return false;
	}
	
    default:
      skills->noAction();
      DECISIONTRACE( "act does not do a thing" );
      return false;
    }
  return false;
}


// read data from the agent's config file (agent.conf)
void 
StandardSituationModule::
readAgent_conf( void )
{
  goalieStaminaFactor 
    = Config::getValueAsDouble("std_goalieStaminaFactor");
  wallStaminaFactor 
    = Config::getValueAsDouble("std_wallStaminaFactor");
  passiveStaminaFactor 
    = Config::getValueAsDouble("std_passiveStaminaFactor");
  activeStaminaFactor 
    = Config::getValueAsDouble("std_activeStaminaFactor");
  standardWaitingTime 
    = Config::getValueAsInt("std_maxWaitingTime");
  waitingTimeAfterGoalieMove
    = Config::getValueAsInt("std_waitingTimeAfterGoalieMove");
  lookAroundAngle
    = Config::getValueAsDouble("std_lookAroundAngle");
  maxSecondDist
    = Config::getValueAsDouble("std_maxSecondDist");
  minWallDist
    = Config::getValueAsDouble("std_minWallDist");
  maxWallDist
    = Config::getValueAsDouble("std_maxWallDist");
  ballAvoidDist
    = Config::getValueAsDouble("std_ballAvoidDist");
  stabilitySearchDist
    = Config::getValueAsDouble("std_stabilitySearchDist");
  penaltyKickGoaliePos 
    = Config::getValueAsVector("std_PenaltyKickGoaliePos");
  penaltyKickFieldplayerPos 
    = Config::getValueAsVector("std_PenaltyKickFieldplayerPos");

  char valueString[100];   

  // reading start positions for the different tactical systems
  for (int i=0; i<11; i++)
    {
      // 4-4-2
      sprintf(valueString, "std_442_KickOffPosWeKickOff_%d", i+1);
      kickOffPosWeKickOff_442[i] 
	= Config::getValueAsVector(valueString);
      sprintf(valueString, "std_442_KickOffPosTheyKickOff_%d", i+1);
      kickOffPosTheyKickOff_442[i] 
	= Config::getValueAsVector(valueString);

      // 3-4-3
      sprintf(valueString, "std_343_KickOffPosWeKickOff_%d", i+1);
      kickOffPosWeKickOff_343[i] 
	= Config::getValueAsVector(valueString);
      sprintf(valueString, "std_343_KickOffPosTheyKickOff_%d", i+1);
      kickOffPosTheyKickOff_343[i] 
	= Config::getValueAsVector(valueString);

      // 3-5-2
      sprintf(valueString, "std_352_KickOffPosWeKickOff_%d", i+1);
      kickOffPosWeKickOff_352[i] 
	= Config::getValueAsVector(valueString);
      sprintf(valueString, "std_352_KickOffPosTheyKickOff_%d", i+1);
      kickOffPosTheyKickOff_352[i] 
	= Config::getValueAsVector(valueString);
    }

  // reading player types and count for stdsits
  for (int t = 0; t < 2; t++)
    {
      const char *str1;
      switch (t)
	{
	case 0:
	  str1 = "our";
	  break;
	case 1:
	  str1 = "other";
	  break;
	default:
	  str1 = "looser";
	}
      for (int fp = 0; fp < 3; fp++)
	{
	  const char *str2;
	  switch (fp)
	    {
	    case 0:
	      str2 = "leftField";  
	      break;
	    case 1:
	      str2 = "middleField";
	      break;
	    case 2:
	      str2 = "rightField";
	      break;
	    default:
	      str2 = "looser";
	    }
	  for (int ts = 0; ts < 3; ts++)
	    {
	      const char *str3;
	      switch (ts)
		{
		case 0:
		  str3 = "442";  
		  break;
		case 1:
		  str3 = "352";
		  break;
		case 2:
		  str3 = "343";
		  break;
		default:
		  str3 = "looser";
		}

	      sprintf(valueString, "std_%s_%s_%s_defenders", str3, str1, str2);
	      players[t][fp][ts].defenders = Config::getValueAsInt(valueString);

	      sprintf(valueString, "std_%s_%s_%s_midfield", str3, str1, str2);
	      players[t][fp][ts].midfield = Config::getValueAsInt(valueString);

	      sprintf(valueString, "std_%s_%s_%s_offenders", str3, str1, str2);
	      players[t][fp][ts].offenders = Config::getValueAsInt(valueString);
	    }
	}
    }
}

bool 
StandardSituationModule::
goingTo(Vector2d &target, 
	const double stamina,
	AdvSkills *skills,
	const Snapshot *worldSnap, 
	const Vector2d *m = NULL,
	const double r = 0)
{
  if (m && r>0)
    calcWayTo(worldSnap->my.pos, target, stamina, *m, r, worldSnap);
  walkAroundPlayers(worldSnap, target, stamina, m, r);

  // and run to it
  AdvSkills::TurtleCommand cmd = 
    skills->goToPosition(worldSnap, target, PLAYER_SIZE, stamina);

  if (cmd.state != AdvSkills::TFINISHED)
    {
      // we are not at the target yet, so run and look around
      skills->exec(worldSnap, cmd);
      lookAround(skills, worldSnap, target);
      DECISIONTRACE( " Going towards " << target << " angle: " 
		     << RAD_TO_DEG((target-worldSnap->my.pos).getAngle()) );
      DRAWTRACE( POSITION(target) );
      DRAWTRACE( LINE(worldSnap->my.pos, target) );
      return true;
    }
  else
    {
      // we arrived at the target, so turn to ball
      AdvSkills::TurnCommand tcmd = 
	skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->ball)+
		     worldSnap->my.bodyDir);
      if (tcmd.state != AdvSkills::RFINISHED)
	{
	  DECISIONTRACE( " Turning to ball " );
	  skills->exec(worldSnap, tcmd);
	  thcmd = skills->turnHead(worldSnap, worldSnap->ball.pos);
	  wantToTurnHead=true;
	  return true;
	}
    }
  // we reached the target and are looking to ball, so we are !goingTo()
  return false;
}

bool 
StandardSituationModule::
goingBackwardsTo(Vector2d &target, 
		 const double stamina,
		 AdvSkills *skills,
		 const Snapshot *worldSnap, 
		 const Vector2d *m = NULL,
		 const double r = 0)
{
  if (m && r>0)
    calcWayTo(worldSnap->my.pos, target, stamina, *m, r, worldSnap);
  walkAroundPlayers(worldSnap, target, stamina, m, r);

  AdvSkills::TurtleCommand cmd = 
    skills->goToPosition(worldSnap, target, PLAYER_SIZE, stamina, 
			 true, (worldSnap->ball.pos - worldSnap->my.pos)
			 .getAngle());

  if (cmd.state != AdvSkills::TFINISHED)
    {
      // we are not at the target yet, so run
      skills->exec(worldSnap, cmd);
      lookAround(skills, worldSnap, target);
      DECISIONTRACE( " Going towards " << target << " angle: " 
		     << RAD_TO_DEG((target-worldSnap->my.pos).getAngle()) );
      DRAWTRACE( POSITION(target) );
      return true;
    }
  else
    {
      AdvSkills::TurnCommand tcmd = 
	skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->ball)+
		     worldSnap->my.bodyDir);
      if (tcmd.state == AdvSkills::RANGLE ||
	  tcmd.state == AdvSkills::ROK)
	{
	  DECISIONTRACE( " Turning to ball " );
	  skills->exec(worldSnap, tcmd);
	  thcmd = skills->turnHead(worldSnap, worldSnap->ball.pos);
	  wantToTurnHead=true;
	  return true;
	}
    }
  // we reached the target and are looking to ball, so we are !goingBackwardsTo()
  return false;
}


// check if players of the other team are in a circle with radius r around pos
bool 
StandardSituationModule::
posIsFree(const Vector2d &pos,
	  const Snapshot *worldSnap, 
	  const double r = 3.) const
{
  double sqr_r = sqr(r);
  for (int i = 1; i <= 11; i++)
    if (sqrnorm(worldSnap->otherTeam(i).pos-pos) < sqr_r)
      return false;
  return true;
}


void 
StandardSituationModule::
walkAroundPlayers(const Snapshot *worldSnap, Vector2d &target, 
		  const double stamina = STAMINA_INC_MAX, 
		  const Vector2d *m = NULL, 
		  const double r = 0) const
{
  if (sqrnorm(target - worldSnap->my.pos) < sqr(PLAYER_SIZE+0.1))
    return;

  double sqr_r = sqr(r);
  // that is how far we can go in this cycle using stamina
  // + a litte more as we don't want goToPosition() returns 
  // FINISHED
  // [C]
  double stretch = stamina*DASH_POWER_RATE+norm(worldSnap->my.v)+1;
  // the angle we use to turn
  // [C]
  const double angle = DEG_TO_RAD(15.);
  // the length of the corridor
  // [C]
  const double length = 5;
  // when turning the corridor its length should be
  // increased by addLength to check if there is
  // a wall of players in front of me
  double addLength;
  // the direction we want to go to
  Vector2d dir = target-worldSnap->my.pos;
  dir *= 1./norm(dir);
  Vector2d mypos = worldSnap->my.pos,
    target1, target2;

  // how many players are in front of me
  // I am always in front of me!
  // [C]*
  int ourCount = worldSnap->ourTeamInCorridor(mypos-0.1*dir, 
					      mypos+length*dir,
					      4*PLAYER_SIZE+0.2);
  int otherCount = worldSnap->otherTeamInCorridor(mypos-0.1*dir,
						  mypos+length*dir,
						  4*PLAYER_SIZE+0.2);
  if (ourCount + otherCount > 1)
    {
      DRAWTRACE( CORRIDOR(mypos-0.1*dir, mypos+length*dir, 4*PLAYER_SIZE+0.2, PI_2 ) );
      DECISIONTRACE( "need to walk around players" );
      OTHERTRACE( " ourCount: " << ourCount << " otherCount: " << otherCount );
      // turning to the right till there is a way
      for (int i = 1; i < (2.*M_PI)/angle; i++)
	{
	  OTHERTRACE( ourCount+otherCount << " Player(s) in front of me " );
	  dir.turn(angle);
	  if (i < M_PI/angle)
	    addLength = length/cos(angle)-length;
	  else
	    addLength = 0;
	  // [C]*
	  if ( (ourCount = worldSnap->ourTeamInCorridor(mypos-0.1*dir, 
							mypos+(length+addLength)*dir, 
							4*PLAYER_SIZE+0.2))+
	       (otherCount = worldSnap->otherTeamInCorridor(mypos-0.1*dir, 
							    mypos+(length+addLength)*dir, 
							    4*PLAYER_SIZE+0.2)) <= 1)
	    {
	      // [C]*
	      DRAWTRACE( CORRIDOR(mypos-0.1*dir, mypos+(length+addLength)*dir, 
				  4*PLAYER_SIZE+0.2, PI_2 ) );
	      target1 = mypos + stretch*dir;
	      break;
	    }
	}

      dir = target-mypos;
      dir *= 1./norm(dir);
      // [C]*
      ourCount = worldSnap->ourTeamInCorridor(mypos-0.1*dir, 
					      mypos+length*dir, 4*PLAYER_SIZE+0.2);
      // [C]*
      otherCount = worldSnap->otherTeamInCorridor(mypos-0.1*dir, 
						  mypos+length*dir, 4*PLAYER_SIZE+0.2);
      // turning to the left till there is a way
      for (int i = 1; i < (2.*M_PI)/angle; i++)
	{
	  OTHERTRACE( ourCount+otherCount << " Player(s) in front of me " );
	  dir.turn(-angle);
	  if (i < M_PI/angle)
	    addLength = length/cos(angle)-length;
	  else
	    addLength = 0;
	  // [C]*
	  if ( (ourCount = worldSnap->ourTeamInCorridor(mypos-0.1*dir, 
							mypos+(length+addLength)*dir, 
							4*PLAYER_SIZE+0.2))+
	       (otherCount = worldSnap->otherTeamInCorridor(mypos-0.1*dir, 
							    mypos+(length+addLength)*dir, 
							    4*PLAYER_SIZE+0.2)) <= 1)
	    {
	      // [C]*
	      DRAWTRACE( CORRIDOR(mypos-0.1*dir, mypos+(length+addLength)*dir, 4*PLAYER_SIZE+0.2, PI_2 ) );
	      target2 = mypos + stretch*dir;
	      break;
	    }
	}

      // check if one of those targets is in the circle
      if ((m && r>0) && sqrnorm(target1-*m) < sqr_r)
	{
	  target = target2;
	  DECISIONTRACE( " going to target1 " << target1 << " as target2 in circle " );
	  // [C]
	  DRAWTRACE( LINE(mypos,mypos+10*(target-mypos)) );
	  return;
	}
      if ((m && r>0) && sqrnorm(target2-*m) < sqr_r)
	{
	  target = target1;
	  DECISIONTRACE( " going to target2 " << target2 << " as target1 in circle " );
	  // [C]
	  DRAWTRACE( LINE(mypos,mypos+10*(target-mypos)) );
	  return;
	}
	
      // which way is the shortest
      // [C]
      if (sqr(norm(target-target1)+2) < sqrnorm(target-target2))
	{
	  target = target1;
	  DECISIONTRACE( " going to target1 " << target1 );
	}
      else
	{
	  target = target2;
	  DECISIONTRACE( " going to target2 " << target1 );
	}
      // [C]
      DRAWTRACE( LINE(mypos,mypos+10*(target-mypos)) );
    }
}

void 
StandardSituationModule::
lookAround(AdvSkills *skills,
	   const Snapshot *worldSnap,
	   const Vector2d &target)
{
  wantToTurnHead=true;
  // look to the left or the right of target, to get more see information 

  Angle turnHeadDir;

  // [C]
  if ( sqrnorm(target - worldSnap->my.pos) < 1.0 )  
    turnHeadDir = worldSnap->my.bodyDir;
  else
    turnHeadDir = worldSnap->my.faceAngleTo(target) + 
      worldSnap->my.faceDir;
  
  // decide where to look next
  if ( nextTurnAction == TURN_TO_LEFT_SIDE )
    turnHeadDir += DEG_TO_RAD(lookAroundAngle);
  else
    turnHeadDir += DEG_TO_RAD(-lookAroundAngle);
  thcmd = skills->turnHead(worldSnap, turnHeadDir);
  
  if ((thcmd.state == AdvSkills::THOK) || 
      (thcmd.state == AdvSkills::THANGLE))
    {
      if ( nextTurnAction == TURN_TO_LEFT_SIDE )
	{
	  DECISIONTRACE( " turning head to the left " );
	  DECISIONTRACE( " turnHeadDir: " << RAD_TO_DEG(turnHeadDir) 
			 << " faceDir: " 
				 << RAD_TO_DEG(worldSnap->my.faceDir));
	  nextTurnAction = TURN_TO_RIGHT_SIDE;
	}
      else
	{
	  DECISIONTRACE(" turning head to the right ");
	  DECISIONTRACE( "turnHeadDir: "<< RAD_TO_DEG(turnHeadDir)
			 << " faceDir: " 
			 << RAD_TO_DEG(worldSnap->my.faceDir));
	  nextTurnAction = TURN_TO_LEFT_SIDE;
	}
    }
  else {
    DECISIONTRACE( " player is not turning, because state = " 
		   << thcmd.state );
    wantToTurnHead=false;
  }
}

int 
StandardSituationModule::
getOurTacticalSystem(const TacticalSnapshot *tacticalSnap) const
{
  // check what our tactical system is
  if ( (tacticalSnap->ourSystem.numDefenders == 4) &&
       (tacticalSnap->ourSystem.numMidfield == 4) &&
       (tacticalSnap->ourSystem.numOffenders == 2) )
    return 442;
  if ( (tacticalSnap->ourSystem.numDefenders ==3) &&
       (tacticalSnap->ourSystem.numMidfield == 4) &&
       (tacticalSnap->ourSystem.numOffenders == 3) )
    return 343;
  if ( (tacticalSnap->ourSystem.numDefenders ==3) &&
       (tacticalSnap->ourSystem.numMidfield == 5) &&
       (tacticalSnap->ourSystem.numOffenders == 2) )
    return 352;
  return 343;        //our default system
}

inline
bool 
StandardSituationModule::
mustAvoidBall(const Snapshot *worldSnap) const
{
  return (sqrnorm(worldSnap->my.pos - worldSnap->ball.pos) < sqr(ballAvoidDist) );
}

void 
StandardSituationModule::
avoidBall(AdvSkills *skills, const Snapshot *worldSnap, 
	  const TacticalSnapshot *tacticalSnap)
{
  if (!mustAvoidBall(worldSnap))
    {
      ERRORTRACE( "avoidBall: player is not closer than ballAvoidDist=" << ballAvoidDist 
		  << " so doing nothing" );
      skills->noAction();
    }

  Vector2d mypos(worldSnap->my.pos);
  Vector2d walkTarget(mypos);


  // check what my tactical task is
  int myTacticalNo = tacticalSnap->myTacticalNo;

  if (myTacticalNo == UNKNOWN)
    {
      DECISIONTRACE( "  tacticalNo unknown !!" );
      // [C]*
      if (mypos.getY() < worldSnap->ball.pos.getY() )
	walkTarget.setXY(mypos.getX() + 2.*ballAvoidDist, mypos.getY() - ballAvoidDist);
      else
	walkTarget.setXY(mypos.getX() + 2.*ballAvoidDist, mypos.getY() + ballAvoidDist);
      
    }
  else if (tacticalSnap->ourTeamDefender(myTacticalNo))
    {
      // [C]
      walkTarget.setX(mypos.getX() - 1.5*ballAvoidDist);
    }
  else if (tacticalSnap->ourTeamMidfield(myTacticalNo))
    {
      if (mypos.getY() < worldSnap->ball.pos.getY() )
	// [C]
	walkTarget.setY(mypos.getY() - 1.5*ballAvoidDist);
      else
	// [C]
	walkTarget.setY(mypos.getY() + 1.5*ballAvoidDist);	
    }
  else if (tacticalSnap->ourTeamOffender(myTacticalNo))
    {
      // [C]
      walkTarget.setX(mypos.getX() + 1.5*ballAvoidDist);      
    }
  else // I am goalie
    {
      // [C]
      walkTarget.setX(mypos.getX() - 2.0*ballAvoidDist);
    }

  // check if walkTarget in the field 
  if (walkTarget.getY() > FIELD_WIDTH/2.0)
    // [C]
    walkTarget.setY(walkTarget.getY() - 2.*ballAvoidDist);
  if (walkTarget.getY() < -FIELD_WIDTH/2.0)
    // [C]
    walkTarget.setY(walkTarget.getY() + 2.*ballAvoidDist);
  
  if (walkTarget.getX() > FIELD_LENGTH/2.0)
    // [C]
    walkTarget.setX(walkTarget.getX() - 3.*ballAvoidDist);
  if (walkTarget.getX() < -FIELD_LENGTH/2.0)
    // [C]
    walkTarget.setX(walkTarget.getX() + 3.*ballAvoidDist);
  
  DECISIONTRACE( "avoidBall: I am near the ball, but do not want to" );
  DECISIONTRACE( " do the freekick, walking to " << walkTarget );

  calcWayTo(mypos, walkTarget, STAMINA_INC_MAX, worldSnap->ball.pos, 
	    BALL_SIZE+PLAYER_SIZE, worldSnap);
  DECISIONTRACE( " after calcWayto: walking to " << walkTarget );

  skills->exec(worldSnap, skills->goToPosition(worldSnap, walkTarget, PLAYER_SIZE));
  return;
}
