//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// $Id: standardsituationmodule.h,v 1.73 2002/06/14 07:57:32 gauer Exp $
// $manager: gauer$
//
//////////////////////////////////////////////////////////////////////


// better change this to upper case:
#ifndef __STANDARDSITUATIONMODULE_H__
#define __STANDARDSITUATIONMODULE_H__

#include "grade.h"
#include "Grid.h"
#define FUNSSELMACROS
#include "drawMacros.h"
#include "mcommand.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"

#include "passeval.h"
#include "goalshoteval.h"
#include "PositionEval.h"
#include "Perm.h"

const int GOALIE_MOVE_POSITIONS = 15;

// in algorithms searching players, this constant is used for distances of not wished players
const double NOWHERE_DIST = 100.0 * sqrt( sqr(FIELD_LENGTH) + sqr(FIELD_WIDTH) );

const double eps = 1e-9;

#undef USEDECISIONTRACE
#undef USEOTHERTRACE
#undef USECMDTRACE
#undef USEERRORTRACE
#undef USEPLANTRACE
#undef USEDRAWTRACE
#undef USEPUSHTRACE
#undef USE_MARKER_TRACE

// activate if you want to have a markerlog
// containing markers at each playmode change
//#define USE_MARKER_TRACE
// number of player setting markers
#define MARKER_NUM 2
#ifdef USE_MARKER_TRACE
#warning *** STDSIT: MARKER_TRACE enabled
#endif

#ifdef STDSIT_TRACES
#include <strstream.h>

#define USEDECISIONTRACE
//#define USEOTHERTRACE
//#define USECMDTRACE
//#define USEERRORTRACE
//#define USEPLANTRACE
//#define USEDRAWTRACE
//#define USEPUSHTRACE

#endif


#ifdef USEDECISIONTRACE
#define DECISIONTRACE(out) {ostrstream ost; \
                            ost << "STD: " << out << (char)0; \
                            char * str = ost.str(); \
                            Comm::sendMCmd(MLog(str)); \
                            delete str;}
#warning DECISIONTRACE enabled
#else
#define DECISIONTRACE(cmd) ;
#endif

#ifdef USEOTHERTRACE
#define OTHERTRACE(out) {ostrstream ost; \
                         ost << "STD: " << out << (char)0; \
                         char * str = ost.str(); \
                         Comm::sendMCmd(MLog(str)); \
                         delete str;}
#warning OTHERTRACE enabled
#else 
#define OTHERTRACE(cmd) ;
#endif

#ifdef USECMDTRACE
#define CMDTRACE(cmd) {ostrstream ost; cmd; ost << (char)0; \
                       char * str = ost.str(); \
                       Comm::sendMCmd(MLog(str)); \
                       delete str;}
#warning CMDTRACE enabled
#else 
#define CMDTRACE(cmd) 
#endif

#ifdef USEERRORTRACE
#define ERRORTRACE(out) cerr <<  worldSnap->time << " STDSIT[" << worldSnap->my.number << "]: " << out << endl
#warning ERRORTRACE enabled
#else
#define ERRORTRACE(cmd) 
#endif

#ifdef USEERRORTRACE1
#define ERRORTRACE1(out) cerr << "STDSIT: " << out << endl
#warning ERRORTRACE1 enabled
#else
#define ERRORTRACE1(cmd) 
#endif

#ifdef USEPLANTRACE
#define PLANTRACE(cmd) {ostrstream ost; \
                        ost << "STD: " << cmd << (char)0; \
                        char * str = ost.str(); \
                        Comm::sendMCmd(MLog(str)); \
                        delete str;}
#warning PLANTRACE enabled
#else
#define PLANTRACE(cmd)
#endif

#ifdef USEDRAWTRACE
#define DRAWTRACE(output) Comm::sendMCmd(output);
#warning DRAWTRACE enabled
#else
#define DRAWTRACE(output)
#endif

#ifdef USEPUSHTRACE
#define PUSHTRACE(cmd) {ostrstream ost; cmd; ost << (char)0; \
                        char * str = ost.str(); \
                        Comm::sendMCmd(MLog(str)); \
                        delete str;}
#warning PUSHTRACE enabled
#else
#define PUSHTRACE(output)
#endif

// our memory for a short time after a standardsituation
class AfterStandardSituationPlan
{
public:
  AfterStandardSituationPlan() : 
    thePlan (PLAN_NOTHING),
    maxSuperKickTime (10) {};

  enum plan {
    PLAN_SUPERKICK,
    PLAN_RUN_TO_POSITION,
    PLAN_NO_ACTION,
    PLAN_TURNING,
    PLAN_NOTHING    
  };
  
  // does superkick. execution will be interrupted if not ready after "maxSuperKickTime" cycles
  void setSuperkick(AdvSkills::ExtKickCommand kickCommand_)
    { thePlan = PLAN_SUPERKICK; kickCommand = kickCommand_; 
      time = maxSuperKickTime; };

  // player will run to "pos_" using "stamina_" stamina per cycle
  void setRunToPosition(Vector2d pos_, double stamina_, int time_)
    { thePlan = PLAN_RUN_TO_POSITION; pos = pos_; stamina = stamina_; 
      time = time_; };

  // player will do nothing for "time_" cycles
  void setNoAction(int time_)            
    { thePlan = PLAN_NO_ACTION; time = time_; };

  // player will turn to the (absolute) direction "dir_" and wait there 
  // for "time_" cycles
  void setTurning(Angle dir_, int time_) 
    { thePlan = PLAN_TURNING; dir = dir_; time = time_; };

  // player has no plan
  void setNothing()                      
    { thePlan = PLAN_NOTHING; };

  // is there any plan to be done ?
  bool planExists()                      
    { return (thePlan != PLAN_NOTHING) && (time > 0); };

  // executes the plan, returns false if it fails 
  bool executePlan(AdvSkills *skills, const Snapshot *worlSnap);

  // what plan will be done ?
  plan getPlan()                         
    { return thePlan; }

private:


  plan thePlan;
  Angle dir;
  int time;
  Vector2d pos;
  double stamina;
  int maxSuperKickTime;

  AdvSkills::ExtKickCommand kickCommand;
};


class StandardSituationModuleInfo;



class StandardSituationModule
{
public:
  /** constructor of module StandardSituationModule */
  StandardSituationModule( bool isGoalie,  
			   const PassEval *_passEval, 
			   const GoalShotEval *_goalShotEval,
			   const PositionEval *_positionEval);

  ~StandardSituationModule();
  
  /** feed information from coach into module
   */
  bool feedbackAudio(AudioMsg &msg);

  /** feed information about new world and tactics into module.
   */
  bool feedbackWorld(const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap);

  /** Do the evaluation for the next action
   * For definition of Grade: see grade.h
   */
  Grade evaluate(AdvSkills* skills,
                 const Snapshot* worldSnap,
		 TacticalSnapshot* tacticalSnap);

  /** Decide whether a turn head should be done.
   * This is always called after act, so the data of act and
   * evaluate should be used to calculate the TurnHeadCommand
   * which is then executed by the Mastercontrol.
   * skills should only be used to calculate the command,
   * NOT TO SEND IT!!
   */
  Grade turnHeadEvaluate(AdvSkills* skills,
			 AdvSkills::TurnHeadCommand &thc);
  
  /** Perform the action for this module */
  bool act(AdvSkills* skills,
           const Snapshot* worldSnap, 
	   TacticalSnapshot* tacticalSnap);


  bool sendSpecialSituation() { return isSpecialSituation; };

private:
  int kickUsPlayer;

  bool actedThisCycle;
  bool wantToTurnHead;
  bool weAreKickingOff;
  bool iAmGoalie; 
  bool isSpecialSituation;
  bool stdSitFinished;
  int maxAge;
  int timeLeft;  // how much time is left until a drop ball is done by the referre
  play_mode playmode;
  AdvSkills::TurnHeadCommand thcmd;

  int lastTime_PLAY_ON;               // the last time the playmode was PLAY_ON
  int lastTime_KICK_OFF_US;           // the last time the playmode was KICK_OFF_US

  // needed for KickOfThem and KickOfUs
  // to avoid multiple calculation of this information
  int actPlayer[11], actCount;
  Vector2d targetPos[11];             // the targets where players should walk to:
                                      // actPlayer[0] will walk to targetPos[0], 
                                      // actPlayer[1] will walk to targetPos[1]...

  Permutations* Perm;                 // used for calculating short ways

  int waitingTime;                    // time for the first player to wait for the second
                                      // (only for active standardsituations where we 
                                      // have the ball)

  int maxKickOffUsWaitingTime;        // the maximum time we wait before doing a kick off

  double minWallDist;                 // if the ball is closer to our goal than
                                      // do not build a wall
  double maxWallDist;                 // if the ball is closer to our goal than
                                      // build a wall out of two players

  double maxSecondDist;               // max way the second player can have to be needed

  int standardWaitingTime;            // the time we want to wait before doing a standard 
                                      // situation by default
  int waitingTimeAfterGoalieMove;     // the time the goalie waits after a move

  double wallStaminaFactor;           // determines how fast the players are running
  double passiveStaminaFactor;        // when building walls or doing other dash-commands
  double goalieStaminaFactor;         // players should dash (STAMINA_INC_MAX) * ...Factor
  double activeStaminaFactor;         // if they have enough stamina

  double lookAroundAngle;             // angle for lookAround()
  double ballAvoidDist;               // the distance to the ball the players will have in
                                      // standardsituations if they are not responsible

  double stabilitySearchDist;         // when players are searched, this constant is used to avoid
                                      // stability problems (e.g. every timestep other players are
                                      // responsible for the same action) [see calcClosestPlayerTo]
  
  // for doing goalie's kick off
  int goalieMovePositions;
  Vector2d goalieMovePos[GOALIE_MOVE_POSITIONS];
  int goalieMovesInCurrentSituation;  // the number of move commands the goalie
                                      // has sent in the current standardsituation
  bool goalieDoesFreeKick;
  bool goalieMayMove;                 // goalie may move in free kick-situation 
                                      // (should be true after a catch-command)
  bool goalieShouldMoveAgain;         // something wrong with the last move command

  bool nextCycleDoNothing;            // next time the goalie does not do anything
                                      // also the evaluation will be bad
                                      // used for the goalie's free-kick only

  AfterStandardSituationPlan myPlan;  // my plan for the time after standardsituations

  // the agents modules 
  const GoalShotEval *goalShotEval;
  const PassEval     *passEval;
  const PositionEval *positionEval;

  enum turn_action {TURN_TO_LEFT_SIDE, TURN_TO_RIGHT_SIDE};
  turn_action nextTurnAction;

  // the positions before the kick-off, different systems
  Vector2d kickOffPosWeKickOff_442[11];       
  Vector2d kickOffPosTheyKickOff_442[11];
  Vector2d kickOffPosWeKickOff_343[11];       
  Vector2d kickOffPosTheyKickOff_343[11];
  Vector2d kickOffPosWeKickOff_352[11];       
  Vector2d kickOffPosTheyKickOff_352[11];

  // the position the goalie has to move when starting a penalty kick
  Vector2d penaltyKickGoaliePos;

  // the position the goalie has to move when starting a penalty kick
  Vector2d penaltyKickFieldplayerPos;

  // the number and type of players for stdsits
  struct playerTypeCount {
    int defenders; 
    int midfield;     
    int offenders; 
  } players[2][3][3]; 
  // 2 teams 3 parts of the field 3 tactical systems

  // does what it says
  void readAgent_conf( void );

  // calcs the number of each player type to do a stdsit 
  void calcDesiredPlayerTypes(const Snapshot *ws, const TacticalSnapshot *ts,
			      const Vector2d &target, int &noDef, 
			      int &noMid, int &noOff) const;

  // pushes one player to a distance of 9.15 around the given "pos".
  void pushPlayer(WorldPlayer &player, const Vector2d &pos) const;

  // calculates the positions where the players start before a kick-off
  Vector2d calcKickOffTarget(int playerNumber, const Snapshot *worldSnap, 
			     const TacticalSnapshot *tacticalSnap) const;

  // calculates where the goalie should move and where to kick the ball 
  // note: the ball should be kicked to a player of our team standing in our 
  // penalty area
  void calcGoalieMoveAndKick(const Snapshot *worldSnap, 
			     const TacticalSnapshot *tacticalSnap,
			     Vector2d &moveTarget,
			     Vector2d &kickTarget,
			     const bool goalieMayMoveOnceMore) const;

  // walks to target using stamina without entering the
  // circle defined by center point m and radius r
  // true if still walking
  // false if target reached and turned to ball
  // if false no action is sent (even not skills->noAction()!)
  bool goingTo(Vector2d &target, const double stamina,
	       AdvSkills *skills,
	       const Snapshot *worldSnap, 
	       const Vector2d *m = NULL,
	       const double r = 0);

  // walks to target but walks backwards while looking to the ball
  // don't forget walking backwards needs twice as much stamina than
  // walking forwards
  // true while walking
  // false when target reached
  bool goingBackwardsTo(Vector2d &target, const double stamina,
			AdvSkills *skills,
			const Snapshot *worldSnap, 
			const Vector2d *m = NULL,
			const double r = 0);

  // does all things neccesary for kick-off situations (US+THEM)
  bool doingKickOff(AdvSkills *skills,
		    const Snapshot *worldSnap,
		    TacticalSnapshot *tacticalSnap);

  // does all things neccesary for Goalie's free-kick situations (THEM)
  bool doingGoalieFreeKickThem(AdvSkills *skills,
			       const Snapshot *worldSnap);

  // does all things neccesary for Goalie's free-kick situations (THEM)
  bool doingFieldPlayerKicksThem(AdvSkills *skills,
				 const Snapshot *worldSnap,
				 const int players[], 
				 const Vector2d target[],
				 const int count);

  bool doingFreeKickUs(AdvSkills *skills,
		       const Snapshot *worldSnap,
		       const int players[],
		       const Vector2d target[],
		       const int count);

  bool doingKickInUs(AdvSkills *skills,
		     const Snapshot *worldSnap,
		     const int players[],
		     const Vector2d target[],
		     const int count);

  bool doingCornerKickUs(AdvSkills *skills,
			 const Snapshot *worldSnap,
			 const int players[],
			 const Vector2d target[], const int count);

  // calc the intersection of the bisector of the angle between the lines 
  // from the ball to the edges of the goal and
  // the line between the edges of the goal
  Vector2d calcPossibleGoalPoint(const Snapshot *worldSnap) const;

  // calc closest players to targets
  // the players are  taken out of the
  // given number of player types
  // depends on our tactical system
  // and the field part ball is in
  // goalie is never a result
  int calcClosestDesiredPlayersTo(const Vector2d target[],
				   const int count,
				   const Snapshot *worldSnap,
				   const TacticalSnapshot *ts,
				   int players[]) const;

  void calcClosestPlayerTo(const Vector2d &target, 
			   const int team[], const int count,
			   const Snapshot *ws, 
			   int &player) const;

  // uses the positions where the players should be according to 
  // positionEval->GetPlayerPos() instead of the real position
  void calcShouldBeClosestPlayerTo(const Vector2d &target, 
				   const int team[], const int count,
				   const Snapshot *ws, 
				   int &player) const;

  // returns number of and positions calculated
  // the Vector2d pos[] has to have length 11
  int calcOurKicksThemPositions(const Snapshot*, 
				const TacticalSnapshot* , 
				Vector2d pos[]) const;

  int calcPlayersAndPos4FreeKickThem(const Snapshot *worldSnap, 
				     const TacticalSnapshot *ts,
				     int players[],
				     Vector2d target[]) const;

  int calcPlayersAndPos4FreeKickUs(const Snapshot *worldSnap, 
				   const TacticalSnapshot *ts,
				   int players[],
				   Vector2d target[]) const;

  int calcPlayersAndPos4KickInThem(const Snapshot *worldSnap, 
				   const TacticalSnapshot *ts,
				   int players[],
				   Vector2d target[]) const;

  int calcPlayersAndPos4KickInUs(const Snapshot *worldSnap, 
				 const TacticalSnapshot *ts,
				 int players[],
				 Vector2d target[]) const;

  int calcPlayersAndPos4CornerKickThem(const Snapshot *worldSnap, 
				       const TacticalSnapshot *ts,
				       int players[],
				       Vector2d target[]) const;

  int calcPos4CornerKickUs(const Snapshot *worldSnap, 
			   Vector2d target[]) const;

  int calcPlayersAndPos4CornerKickUs(const Snapshot *worldSnap, 
				     const TacticalSnapshot *ts,
				     int players[],
				     Vector2d target[]) const;

  // calcs the point where to go next if "to" is not reachebale without entering 
  // the circle
  // defined by "m" and "r"
  // the result will be returned in "to"
  void calcWayTo(const Vector2d &from, Vector2d &to, const double stamina,
		 const Vector2d &middle, const double &radius, 
		 const Snapshot *worldSnap) const;

  // if there are players in my way from my.pos to target
  // target is changed to a position to walk to instead
  // m is the middle of a circle with radius r the players can not walk in
  // target and *m have to be in absolute coordinates
  void walkAroundPlayers(const Snapshot *worldSnap,  Vector2d &target,
			 const double stamina = STAMINA_INC_MAX,
			 const Vector2d *m = NULL, const double r = 0) const;

  // turns the players head to the left and the right of target
  void lookAround(AdvSkills *skills,
		  const Snapshot *ws,
		  const Vector2d &target);

  // check if players of the other team are in a cricle with radius r around pos
  bool posIsFree(const Vector2d &pos, 
		 const Snapshot *worldSnap, 
		 const double r = 3.) const;

  // goalie doing goalkick
  bool doingGoalieGoalKickUs(AdvSkills *skills,
			     const Snapshot *worldSnap,
			     TacticalSnapshot *tacticalSnap);

  // fieldplayer finishing goalkick
  bool doingFieldPlayerGoalKickUs(AdvSkills *skills,
				  const Snapshot *worldSnap,
				  TacticalSnapshot *tacticalSnap);

  // goalies goal-kick-situation
  bool doingGoalieFreeKickUs(AdvSkills *skills,
			     const Snapshot *worldSnap,
			     TacticalSnapshot *tacticalSnap);

  // if all passes are bad, this should be used to make an emergency kick-off.
  // routine scans our and other player in the left and the right third and returns an
  // "optimal" move-position
  Vector2d calcEmergencyMovePosition(const Snapshot *worldSnap) const;


  // calculates where to pass the ball when we have kick-off
  Vector2d calcKickOffKickTarget(const Snapshot *worldSnap, 
				 const TacticalSnapshot *tacticalSnap,
				 int &passReceiver) const;

  // changes the players so that the ways to walk are shorter
  void calcShortWays(const Snapshot* worldSnap,
		     int players[], const Vector2d target[],
		     const int count) const;


  double calcUsableStamina( const double &factor,
			     const Snapshot * );

  // gets our tactical system and returns an int (442, 343, 352, ...)
  int getOurTacticalSystem(const TacticalSnapshot *tacticalSnap) const;  

  // calculates the intersection of two evens 
  Vector2d calcIntersection(const Vector2d &A, const Vector2d &B, 
			    const Vector2d &C, const Vector2d &D) const;

  // calculates the shortest distance from A to B around the circle (M,r)
  // worldsnap is only needed for drawing nice pics
  // note: this only works if AB and (M,r) intersect
  double calcDistAroundCircle(const Vector2d &A, const Vector2d &B, 
			      const Vector2d &M, const double r,
			      const Snapshot *worldSnap) const;

  // does player have to avoid the ball ?
  inline bool mustAvoidBall(const Snapshot *worldSnap) const;

  // avoids the ball: if player is near to the ball, run away
  void avoidBall(AdvSkills *skills, const Snapshot *worldSnap,
		 const TacticalSnapshot *tacticalSnap);

  // does the evaluation for PLAY_MODE_FREE_KICK_US
  Grade evalFreeKickUs(AdvSkills* skills,
		       const Snapshot* worldSnap, 
		       const TacticalSnapshot* tacticalSnap);

  // does the evaluation for PLAY_MODE_KICK_IN_US and
  // PLAY_MODE_CORNER_KICK_US
  Grade evalRemainingKicksUs(AdvSkills* skills,
			     const Snapshot* worldSnap, 
			     const TacticalSnapshot* tacticalSnap, 
			     const play_mode playMode);

}; // class StandardSituationModule

#endif

