//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// $Id: stdGoalie.C,v 1.21 2002/06/14 07:57:32 gauer Exp $
// $manager: gauer$
//
//////////////////////////////////////////////////////////////////////

#include "standardsituationmodule.h"
#include "Config.h"

void 
StandardSituationModule::
calcGoalieMoveAndKick(const Snapshot *worldSnap, 
		      const TacticalSnapshot *tacticalSnap,
		      Vector2d &moveTarget,
		      Vector2d &kickTarget,
		      const bool goalieMayMoveOnceMore) const
{
  if (moveTarget == kickTarget); // to avoid warnings
  if (tacticalSnap);             // to avoid warnings

  kickTarget.setXY(0.0, 0.0); // in the case that something goes wrong ...
  // the grades for all movePositions and the best found grade

  int bestPassPartner = UNKNOWN;

  // advanced ourEval:
  PropOf props;
  props.perhaps = false;
  props.grade.set_BAD();
  PropOf actProps;


  if (!goalieMayMoveOnceMore)
    {
      DECISIONTRACE( "calcGoalieMoveAndKick: goalie must not move" );
      // goalie must not move
      for (int passPartner=1; passPartner<=11; passPartner++)
	{
	  if (passPartner == worldSnap->ourGoalieNo)
	    continue;   // goalie does not want the ball
	  if (worldSnap->ourTeam(passPartner).age > maxAge)
	    continue;   // old players are no good !!
	  actProps = passEval->ourEvalProp(worldSnap, worldSnap->ourGoalieNo, passPartner);
	  if (passEval->betterPass(worldSnap, actProps, props))
	    {
	      props           = actProps;
	      bestPassPartner = props.number;
	      kickTarget      = props.pos;
	    }
	} // passPartner      
      DECISIONTRACE( "bestPassPartner = " << bestPassPartner << ", grade = " 
		     << props.grade);
      return;
    }
  
  // goalie can do one more move-command

  Grade positionGrades[goalieMovePositions];
  int bestMovePos     = 0;

  Snapshot simSnap = *worldSnap;

  WorldPlayer ourPlayer;
  WorldPlayer otherPlayer;  

  for (int posNumber=0; posNumber<goalieMovePositions; posNumber++)
    {
      // copy positions of all players
      PUSHTRACE( ost << "the pushings while goalieMove to " 
		 << goalieMovePos[posNumber] );
      for (int playerNumber=1; playerNumber<=11; playerNumber++)
	{
	  // pushing our player
	  ourPlayer   = worldSnap->ourTeam(playerNumber);
	  PUSHTRACE( ost << "ourTeam[" << playerNumber << "].pos = " << ourPlayer.pos
                     << " -> " );
	  pushPlayer(ourPlayer, goalieMovePos[posNumber]);
	  simSnap.setOurTeam(playerNumber, ourPlayer); 
	  PUSHTRACE( ost << simSnap.ourTeam(playerNumber).pos);
	  
	  // pushing other player
	  otherPlayer = worldSnap->otherTeam(playerNumber); 
	  PUSHTRACE( ost << "otherTeam[" << playerNumber << "].pos = " << otherPlayer.pos
		   << " -> " );
	  pushPlayer(ourPlayer, goalieMovePos[posNumber]);
	  simSnap.setOtherTeam(playerNumber, otherPlayer);
	  PUSHTRACE( ost << simSnap.otherTeam(playerNumber).pos );
	} // playerNumber

      // moving our goalie
      ourPlayer = simSnap.ourTeam(simSnap.ourGoalieNo);
      ourPlayer.pos = goalieMovePos[posNumber];
      simSnap.setOurTeam(simSnap.ourGoalieNo, ourPlayer);
      simSnap.my = simSnap.ourTeam(simSnap.ourGoalieNo);
      // for debugging pushing-algorithm
      PUSHTRACE(
		if (posNumber == 1)
		{
		  // [C]
		  double sizeFactor = 2;
		  for (int playerNumber=1; playerNumber <= 11; playerNumber++)
		    {
		      if (simSnap.ourTeam(playerNumber).age < maxAge)
			DRAWTRACE( CIRCLE(simSnap.ourTeam(playerNumber).pos, 
					  sizeFactor*PLAYER_SIZE ) );
		      if (simSnap.otherTeam(playerNumber).age < maxAge)
			DRAWTRACE( CIRCLE(simSnap.otherTeam(playerNumber).pos, 
					   sizeFactor*PLAYER_SIZE ) );
		    }
		  DRAWTRACE( GCIRCLE(simSnap.ball.pos, sizeFactor*BALL_SIZE) );
		  DRAWTRACE( GCIRCLE(simSnap.my.pos, sizeFactor*PLAYER_SIZE) );
		}		
		);
      
      // look for the best pass
      positionGrades[posNumber].set_IMPOSSIBLE();
      for (int passPartner=1; passPartner<=11; passPartner++)
	{
	  if (passPartner == simSnap.ourGoalieNo)
	    continue;   // goalie does not want the ball
	  if (simSnap.ourTeam(passPartner).age > maxAge)
	    continue;   // old players are no good !!
	  
	  actProps = passEval->ourEvalProp(&simSnap, worldSnap->ourGoalieNo, passPartner);
	  if (passEval->betterPass(&simSnap, actProps, props))
	    {
	      props           = actProps;
	      bestPassPartner = props.number;
	      kickTarget      = props.pos;
	      bestMovePos     = posNumber;
	    }
	  if (props.grade.is_better_than(positionGrades[posNumber]))
	    positionGrades[posNumber] = props.grade;
	} // passPartner      
      DECISIONTRACE( "moving to " << goalieMovePos[posNumber] 
		     << " => grade=" << positionGrades[posNumber]
		     << ", bestGrade=" << props.grade
		     );
    } // posNumber
  
  
  // now we move to goalieMovePos[bestMovePos] and think of passing to passPartner
  moveTarget = goalieMovePos[bestMovePos];
  DECISIONTRACE( "goalie has done kick-off decision: move to " << moveTarget << " ..." );
  DRAWTRACE(POSITION(moveTarget));
  DECISIONTRACE( " ... perhaps trying to kick to player " 
		 << bestPassPartner << " at " << kickTarget
		 << " grade= " << props.grade );
  DRAWTRACE(POSITION(kickTarget));
  DRAWTRACE(SLINE(moveTarget,kickTarget,MDash));
  
  if (props.grade.is_worse_than(Grade::BAD))
    {
      DECISIONTRACE( "best pass is bad, so think about moving ..." );
      moveTarget = calcEmergencyMovePosition(worldSnap);
      DECISIONTRACE( "move to " << moveTarget );
    }
}


void
StandardSituationModule::
pushPlayer(WorldPlayer &player, 
	   const Vector2d &pos) const
{
  if ( (player.age > maxAge)  // too old players are ignored 
       || ( player.distTo(pos) > FREE_KICK_FORBIDDEN_AREA) )
    return;
  
  // player has to be pushed outside the circle and his speed has to be set to 0.
  player.v.setXY(0.0, 0.0);
  Vector2d distToPos = player.pos;
  distToPos -= pos;
  double length = distToPos.getLength();
  // [C]
  if (length < 0.1) 
    {
      // player is very close, so place him in
      // front of me
      player.pos.setXY(pos.getX(), 
			pos.getY()+FREE_KICK_FORBIDDEN_AREA);
    }
  else
    {
      // player is not so close, so pushing him the server-like way
      distToPos.stretch(FREE_KICK_FORBIDDEN_AREA / length);
      player.pos = pos;
      player.pos += distToPos;
    }
}

Vector2d 
StandardSituationModule::
calcEmergencyMovePosition(const Snapshot *worldSnap) const
{
  double valueLeft = 0.0, valueRight = 0.0;

  // [C]*
  valueLeft -= 3.0 * 
    worldSnap->otherTeamInCorridor(Vector2d(-FIELD_LENGTH/2.0,-FIELD_WIDTH/3.0), 
				   Vector2d(-FIELD_WIDTH/6.0, -FIELD_WIDTH/3.0), 
				   FIELD_WIDTH/3.0);
  valueLeft += 1.0 * 
    worldSnap->ourTeamInCorridor(Vector2d(-FIELD_LENGTH/2.0,-FIELD_WIDTH/3.0), 
				 Vector2d(-FIELD_WIDTH/6.0, -FIELD_WIDTH/3.0), 
				 FIELD_WIDTH/3.0);

  valueRight -= 3.0 * 
    worldSnap->otherTeamInCorridor(Vector2d(-FIELD_LENGTH/2.0,FIELD_WIDTH/3.0), 
				   Vector2d(-FIELD_WIDTH/6.0, FIELD_WIDTH/3.0), 
				   FIELD_WIDTH/3.0);
  valueRight += 1.0 * 
    worldSnap->ourTeamInCorridor(Vector2d(-FIELD_LENGTH/2.0,FIELD_WIDTH/3.0), 
				 Vector2d(-FIELD_WIDTH/6.0, FIELD_WIDTH/3.0), 
				 FIELD_WIDTH/3.0);

  // [C]*
  if (valueLeft > valueRight)  // left position is better
    return Vector2d(-FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH - 1.5, 
		    -PENALTY_AREA_WIDTH / 2.0 + 1.5);
  else
    return Vector2d(-FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH - 1.5, 
		    PENALTY_AREA_WIDTH / 2.0 - 1.5);
}


bool 
StandardSituationModule::
doingGoalieGoalKickUs(AdvSkills *skills,
		      const Snapshot *worldSnap,
		      TacticalSnapshot *tacticalSnap)
{
  // I am goalie
  if ( !skills->kickSafe(worldSnap) )
    {
      DECISIONTRACE( "Goalie wants to do the goalkick, but has to go to the ball first" );
      DECISIONTRACE( "dist2Ball=" << norm(worldSnap->my.pos - worldSnap->ball.pos) );
      Vector2d targetPos = worldSnap->ball.pos;
      // [C]
      targetPos.setX(targetPos.getX()-PLAYER_SIZE-BALL_SIZE);
      if ( (targetPos.getX() < worldSnap->my.pos.getX()+PLAYER_SIZE/2.0) ||
	   (fabs(targetPos.getY()-worldSnap->my.pos.getY()) > 1.) )
	{
	  double yPos = targetPos.getY();
	  if (worldSnap->my.pos.getY() < yPos) 
	    yPos -= 0.5;
	  else
	    yPos += 0.5;
	  targetPos.setXY(targetPos.getX()-2., yPos);
	}
	  
#ifdef USEDRAWTRACE
      // [C]
      double r = BALL_SIZE+PLAYER_SIZE;
      // [C]
      Vector2d m = (r-1.5*BALL_SIZE)/norm(worldSnap->ball.pos-targetPos)*
	(worldSnap->ball.pos-targetPos) + 
	worldSnap->ball.pos;
      DRAWTRACE( CIRCLE(m, r) );
#endif
      if (goingTo(targetPos, STAMINA_INC_MAX, skills, worldSnap,  &(worldSnap->ball.pos), 
		  BALL_SIZE+PLAYER_SIZE))	
	{
	  DECISIONTRACE( "Goalie going to " << targetPos );
	}
      else
	{
	  DECISIONTRACE( "goalie has reached ball, waitingTime = " << waitingTime );
	  skills->noAction();
	}
      lookAround(skills, worldSnap, worldSnap->ball.pos);
      waitingTime = standardWaitingTime; 
      return true;
    } // running to ball (if)

  waitingTime--;
  // now getting some wm-info and let the positionmodule do it's job
  // (we want many many passpartners when doing the freekick)
  // scheme:
  //         looking to the others goal            (waitingTime >  0)
  //         turning head until we can use say message and one of
  //           our players is in the penalty area  (waitingTime == 0)
  //         kicking the ball to one of our players in our penalty area
  //           incl. turning body towards this player


  if (waitingTime > 0)
    {
      Angle bodyTurnDir = worldSnap->my.bodyAngleTo(worldSnap->otherGoal.pos)+worldSnap->my.bodyDir;
      AdvSkills::TurnCommand 
	tcmd = skills->turn(worldSnap, bodyTurnDir);	  
      if (tcmd.state != AdvSkills::RFINISHED)
	{
	  skills->exec(worldSnap, tcmd);
	  DECISIONTRACE( "Goalie wants to wait because waitingTime= " 
			 << waitingTime
			 << " turning to others goal" ); 
	}
      else if (waitingTime > 0)
	skills->noAction();
	  
      DECISIONTRACE( "Goalie wants to wait because waitingTime= " 
		     << waitingTime ); 
      lookAround(skills, worldSnap, worldSnap->otherGoal.pos);
      return true;
    }	  

  // waitingTime should be == 0 now
  // goalie should look for our players in our penalty area
  // if there is anyone, just turn to him and see right into his eye
      
  Grade playerGrades[12][12], grade;
  int ourPlayer;
  Vector2d playerPos, dummyPos;
  Snapshot hsnap(*worldSnap), *ws(&hsnap);
  for (ourPlayer=1; ourPlayer<=11; ourPlayer++)
    {
      for (int j = 0; j < 12; j++)
	playerGrades[ourPlayer][j].set_IMPOSSIBLE();

      if (ourPlayer == worldSnap->my.number || 
	  worldSnap->ourTeam(ourPlayer).age > maxAge)
	continue;

      playerPos = worldSnap->ourTeam(ourPlayer).pos;
      ws->ball.pos = playerPos;

      if ( inOurPenaltyArea(playerPos) )
	{
	  // check for the player all other players as passpartners
	  for (int passPartner2=1; passPartner2 <= 11; passPartner2++)
	    {
	      if (passPartner2 == ourPlayer || 
		  passPartner2 == worldSnap->ourGoalieNo)
		continue;
	      grade = passEval->ourEval(ws, ourPlayer, 
					passPartner2, dummyPos);

	      playerGrades[ourPlayer][passPartner2] = grade;
//  	      DRAWTRACE( LINE(ws->ball.pos, ws->ourTeam(passPartner2).pos) );
//  #ifdef USEDRAWTRACE
//  	      {
//  		ostrstream text;
//  		text << grade << (char)0;
//      	char * str = text.str();
//  		DRAWTRACE( TEXT(str, ws->ourTeam(passPartner2).pos) );
//              delete str;
//  	      }
//  #endif
	    }
	}
    }

  int gradeCount[12][6] = {{0}};

  for (int i=1; i<=11; i++)
    for (int j = 1; j<=11; j++) {
      if (playerGrades[i][j].is_like(Grade::IMPOSSIBLE))
	gradeCount[i][0]++;

      if (playerGrades[i][j].is_like(Grade::BAD))
	gradeCount[i][1]++;

      if (playerGrades[i][j].is_like(Grade::POOR))
	gradeCount[i][2]++;

      if (playerGrades[i][j].is_like(Grade::AVERAGE))
	gradeCount[i][3]++;

      if (playerGrades[i][j].is_like(Grade::GOOD))
	gradeCount[i][4]++;

      if (playerGrades[i][j].is_like(Grade::PERFECT))
	gradeCount[i][5]++;
    }

  int count(0);
  for (int i = 1; i <= 11; i++) {
    if (gradeCount[i][5] > count) {
      count = gradeCount[i][5];
      ourPlayer = i;
      DECISIONTRACE("found passpartner " << i << " with " << count << " PERFECT passes");
    }
  }
  if (count == 0)
    for (int i = 1; i <= 11; i++) {
      if (gradeCount[i][4] > count) {
	count = gradeCount[i][4];
	ourPlayer = i;
	DECISIONTRACE("found passpartner" << i << " with " << count << " GOOD passes");
      }
    }
  if (count == 0)
    for (int i = 1; i <= 11; i++) {
      if (gradeCount[i][3] > count) {
	count = gradeCount[i][3];
	ourPlayer = i;
	DECISIONTRACE("found passpartner" << i << " with " << count << " AVERAGE passes");
      }
    }
  if (count == 0)
    for (int i = 1; i <= 11; i++) {
      if (gradeCount[i][2] > count) {
	count = gradeCount[i][2];
	ourPlayer = i;
	DECISIONTRACE("found passpartner" << i << " with " << count << " POOR passes");
      }
    }
  if (count == 0) {
    DECISIONTRACE("all passes are BAD or worse! using player 2");
    ourPlayer = 2;
  }

  playerPos = worldSnap->ourTeam(ourPlayer).pos;
  DRAWTRACE( POSITION(playerPos) );
      
  if (count != 0)
    {
      // check if it is necessary to turn to player
      AdvSkills::TurnCommand tcmd
	= skills->turn(worldSnap, worldSnap->my.bodyAngleTo(playerPos)+worldSnap->my.bodyDir);
      if (tcmd.state != AdvSkills::RFINISHED)
	{
	  skills->exec(worldSnap, tcmd);
	  DECISIONTRACE( "Goalie turning to passpartner (" << ourPlayer  
			 << ") @ " << playerPos 
			 << ", waitingTime=" << waitingTime );
	  thcmd = skills->turnHead(worldSnap, worldSnap->ourTeam(ourPlayer));
	  wantToTurnHead = true;
	  waitingTime = 1;
	  return true;
	}

      // check how long the pass to our player will take (place or not)
      int kickTime(1);
      AdvSkills::ExtKickCommand kcmd
	= skills->kick(worldSnap, .6*skills->stopableV, playerPos);
      kickTime = kcmd.time;
      OTHERTRACE("kickTime: " << kickTime);
      OTHERTRACE("time of EMRG: " << worldSnap->nextTimeOfChannel(AUMUL_EMRG));
      // if body-direction okay, wait until message-channel is good
      // and the player I wanna kick to has reached its position
      OTHERTRACE("dist2(" << ourPlayer << ") : " 
		 << worldSnap->ourTeam(ourPlayer).distTo(targetPos[ourPlayer-1]));
      if ((worldSnap->nextTimeOfChannel(AUMUL_EMRG)-worldSnap->time-kickTime+1 == 0) &&
	  (worldSnap->ourTeam(ourPlayer).distTo(targetPos[ourPlayer-1]) < 1.) )
	{
	  // emergency channel can be used, so kick to our player slowly
	  skills->exec(worldSnap, kcmd);
	  thcmd = skills->turnHead(worldSnap, worldSnap->ourTeam(ourPlayer));
	  wantToTurnHead = true;
	  DECISIONTRACE( "Goalie kicking, using emergency channel, waitingTime=" 
			 << waitingTime );
	  DECISIONTRACE( "reporting pass to tacticalmodule" );
	  // message should be sent 
	  tacticalSnap->reportOurTacticalAction(TA_PASS, worldSnap->my.number, ourPlayer);
	  if (kickTime == 1)
	    {
	      waitingTime = -1;
	      nextCycleDoNothing = true;
	      stdSitFinished = true;
	      DECISIONTRACE("stdsit finished");
	    }
	  else 
	    waitingTime = 1;
	  return true;
	}

      for (int i=1; i<=11; i++)
	DECISIONTRACE( "playerPos[" << i << "] = " << targetPos[i] );
      DECISIONTRACE( "ourPlayer.pos = " << worldSnap->ourTeam(ourPlayer).pos );

      skills->noAction();
      lookAround(skills, worldSnap, playerPos);
      DECISIONTRACE( "Goalie looking around, waiting for emergency channel, waitingTime=" 
		     << waitingTime << ", kickTime = " << kickTime );
      waitingTime = 1;
      return true;
    }

  if (timeLeft > 20)
    {
      skills->noAction();
      lookAround(skills, worldSnap, playerPos);
      DECISIONTRACE( "Goalie looking around,  waitingTime=" 
		     << waitingTime );
      if (waitingTime == 0)
	waitingTime ++; 
      return true;
    }   

  // ??? kickTarget bad (evtl. other players on the way)
  Vector2d kickTarget(-FIELD_LENGTH/2.0+5.0, -FIELD_WIDTH/2.0);
  DRAWTRACE( POSITION(kickTarget) );
  DECISIONTRACE( "EMERGENCY!! Goalie has no passpartners found" );
  DECISIONTRACE( "KICKING TO WALACHEI @ " << kickTarget);
  skills->exec(worldSnap, skills->kick(worldSnap, BALL_SPEED_MAX, kickTarget));
  waitingTime = -1;
  return false;
} 



bool 
StandardSituationModule::
doingGoalieFreeKickUs(AdvSkills *skills,
		      const Snapshot *worldSnap,
		      TacticalSnapshot *tacticalSnap)
{
  if (!iAmGoalie)
    {
      DECISIONTRACE( "I am not goalie, doing nothing" );
      if (mustAvoidBall(worldSnap) || worldSnap->my.inOurPenaltyArea())
	{
	  DECISIONTRACE( "I have to avoid the ball and have to keep out of our penalty area" );
	  Vector2d posEvalTarget =
	    positionEval->GetPlayerPosition(worldSnap->my.number);
	  WorldObject wO;
	  wO.pos = posEvalTarget;
	  if (wO.inOurPenaltyArea())
	    wO.pos.setX(-FIELD_LENGTH/2. + PENALTY_AREA_LENGTH + 8.);
	  
	  AdvSkills::TurtleCommand tcmd =
	    skills->goToPosition(worldSnap, wO.pos, 1);
	  if (tcmd.state != AdvSkills::TFINISHED)
	    skills->exec(worldSnap, tcmd);
	  else
	    skills->basicTurn(PI/4.);
	  lookAround(skills, worldSnap, worldSnap->ball.pos);
	}
      else
	skills->noAction();
      lookAround(skills, worldSnap, worldSnap->ball.pos);
      return false;
    }

  // I am Goalie: now do the freekick after a catch
  waitingTime--;

  // collecting info
  if (waitingTime > 0)
    {
      AdvSkills::TurnCommand tcmd =
	skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->otherGoal.pos)+
		     worldSnap->my.bodyDir);
      if (tcmd.state != AdvSkills::RFINISHED)
	skills->exec(worldSnap, tcmd);
      else
	skills->noAction();
      lookAround(skills, worldSnap, worldSnap->otherGoal.pos);
      DECISIONTRACE( "Goalie looking around, waitingTime = " << waitingTime );
      return true;
    }


  // check if goalie should (may) move or not, then kick
  DECISIONTRACE( "I am goalie and want to do " << worldSnap->playMode );
  // if not ballkickable => walk towards it
  Vector2d moveTarget, kickTarget;
  if ( goalieMayMove &&
       (goalieMovesInCurrentSituation < GOALIE_MAX_MOVES) &&
       (goalieMovesInCurrentSituation < 1 || goalieShouldMoveAgain) )
    {
      // goalie may move once again
      calcGoalieMoveAndKick(worldSnap, tacticalSnap,
			    moveTarget, kickTarget, true);
      // [C]
      if (worldSnap->my.distTo(moveTarget) > 1.0 ) 
	{
	  goalieMovesInCurrentSituation++;
	  goalieShouldMoveAgain = false;
	  DECISIONTRACE( "goalie moving to " << moveTarget
			 << " goalieMovesInCurrSit: " << goalieMovesInCurrentSituation);
	  skills->basicMove(moveTarget.getX(), moveTarget.getY());
	  DRAWTRACE(POSITION(moveTarget));
	  lookAround(skills, worldSnap, worldSnap->otherGoal.pos);
	  // [C]
	  waitingTime = waitingTimeAfterGoalieMove;
	  return true;
	}
    }
  else
    {
      // we have moved or may not, so do the kick
      calcGoalieMoveAndKick(worldSnap, tacticalSnap,
			    moveTarget, kickTarget, false);
      DECISIONTRACE( "goalie has already been moved or may not move, so waiting for"
		     << " the other modules" );
    }

  // next time another module should do some action so 
  waitingTime = waitingTimeAfterGoalieMove;
  nextCycleDoNothing = true;
  stdSitFinished = true;
  DECISIONTRACE( "doing nothing, plan to let other modules do their job" );
  DECISIONTRACE( "Reset waitingTime to " << waitingTime );
  skills->noAction();
  lookAround(skills, worldSnap, worldSnap->ball.pos);

  return true;
}
